/***********************************************************
 * FeC2.c      
 *
 *  Front-end Control and Configuration program           
 *
 * last updated: 24/05/2006
 * written by: Ervin Denes
 *                                                           
 * history:                                           
 *                                                   
 * v.4.5.1  24-May-2006: RORC revision and serial can identify the RORC
 *          23-May-2006: accept environment variables in the script
 * v.4.4.5  16-Feb-2006: check time-out option       
 * v.4.4.4  05-Jan_2006: new command: write_block_multiple
 *          21-Dec_2005: print sw version at the beginning of the run 
 * v.4.4.3  29-Aug-2005: add read_ddl-status command
 * v.4.4.1  04-Apr-2005: physmem offset depends on minor and channel number
 * v.4.3.14 18-Mar-2005: stop if protocol error occurs at block read commands
 * v.4.3.13 10-Mar-2005: set failed code if too much/less data arrive
 *                       for the READ_AND_CHECK_BLOCK command
 * v.4.3.13 03-Mar-2005: print warning if FEE command or address is too large
 * v.4.3.5  06-Aug-2004: time-out in usecs    
 * v.4.3.1  26-May-2004: add RESET ALL command
 * v.4.3.1  25-May-2004: fixing a bug at READ_AND_PRINT command
 * v.4.3.1  18-May-2004: add write_RDYRX and write_EOBTR commands
 *                       add program options -O and -U
 * v.4.3.1  29-Apr-2004: increase number of shared memory files         
 *                       add command WAIT
 * v.4.3.1  20-Apr-2004: add shared memory files         
 * v.4.2    22-Aug-2003: add DDL channels                
 * v.4.1    10-Jul-2003: add write_JTAG command                
 * v.4.1    07-Apr-2003: original version                      
 **********************************************************/
#define SOURCE_FILE "FeC2.c"
#define DESCRIPTION "Front-end Control and Configuration"

#define SCRIPT_FN "FeC2.scr"
#define LOG_FN "stdout stream"
#define MAX_FILES 32
#define MAX_DEFINES 100
#define MAX_TOKEN 16
#define MAX_PHYSMEM (1024*1024*8)
#define DDL_PARAM_MASK 0x7FFFF

#define ISM_KEY  0xFeC20000
#define ISM_SIZE 0x00400000 
#define MAX_SHM_FILE 15420
#define MAX_SHM_KEY 127
#define SIZE_SHM 0x400000 

#define FEC2LOG(message...) fprintf (logFile, message)

#include "rorc_ddl.h"

char *command_tokens[] = {
                          "DEFINE",
                          "RESET",
                          "READ_DDL_STATUS",
			  "WRITE_RDYRX",
			  "WRITE_EOBTR",
                          "WRITE_COMMAND",
                          "WRITE_BLOCK",
                          "WRITE_BLOCK_MULTIPLE",
                          "WRITE_JTAG",
                          "READ_AND_PRINT",
                          "READ_AND_CHECK",
                          "READ_UNTIL",
                          "READ_BLOCK",
			  "READ_AND_CHECK_BLOCK",
                          "WAIT",
                          "CALL_FILE",
                          "RETURN",
			  "STOP_IF_FAILED",
                          "STOP",
                          ""
                         };

enum command_index {DEFINE=1, RESET, READ_DDL_STATUS,
                              WRITE_RDYRX, WRITE_EOBTR,
                              WRITE_COMMAND, WRITE_BLOCK, WRITE_BLOCK_MULTIPLE,
                              WRITE_JTAG, READ_AND_PRINT, 
                              READ_AND_CHECK, READ_UNTIL, 
			      READ_BLOCK, READ_AND_CHECK_BLOCK, WAIT,
		              CALL_FILE, RETURN, STOP_IF_FAILED, STOP};
// minimum parameters:
int command_param[] = {-1, 2, 0,     0,
                              0,           0,
                              1,             2,           7,
                              2,          2,
                              3,              4,
                              2,          2,                    1, 
			      1,         0,      0,              0};

FILE *logFile;
char log_fn[80];
char script_fn[MAX_FILES][80];
int file_index, line_no[MAX_FILES];
int minor, channel, shm;
int revision, serial;

/*******************************************************/
void usage(char *name)
{
  printf("Front-end Control and Configuration program.\n");
  printf("usage: FeC2\n");
  printf("       [-{M|m} <minor> | -{R|r} <revision> -{N|n} <serial>][-{C|c} <channel>]\n");
  printf("       [-{F|f} <script file>] [-{L|l} <log file>] [-{T|t} <time-out>]\n");
  printf("       [-{O|o} <mem offset>] [-{U|u} <mem size>] [-S|-s] [-v] [-H|-h]\n");
  printf("   -M <minor>       RORC minor number, default: 0.\n");
  printf("   -R <revision>    RORC PCI revision number. Must be < %d.\n",
                                                RORC_MAX_REVISION + 1);
  printf("   -N <serial>      RORC hw serial number.\n");
  printf("                    If given RORC is identified by <revision> and <serial>,\n");
  printf("                    not by <minor> number.\n");
  printf("   -C <channel>     RORC channel number, default: 0.\n");
  printf("   -F <file name>   name of the scipt file. Default: %s.\n", SCRIPT_FN);
  printf("   -L <file name>   name of the log file. Default: %s.\n", LOG_FN);
  printf("   -T <time-out>    time-out value for DDL commands. Default: %d usec.\n",
          DDL_RESPONSE_TIME);
  printf("   -O <mem offset>  offset in physmem memory in MBs.\n");
  printf("                    Default: (<minor> * %d + <channel>) * %d.\n", 
                                        MAX_CHANNEL, MAX_PHYSMEM/1024/1024);
  printf("   -U <mem size>    utilizable physmem memory in MBs, default: %d.\n",
                                                     MAX_PHYSMEM/1024/1024);
  printf("   -S               use shared memory for file storage.\n");
  printf("   -v               verbose mode (capital V is not accepted).\n");
  printf("   -H               print usage.\n");
}

/******************************************************************************/
/* Create a Shared Memory Segment                                             */
/******************************************************************************/
int create_shm(key_t shm_key, int shm_size, unsigned long **shm_addr)
{
  int exit_code, shmid;

  exit_code = 0;

  shmid = shmget(shm_key, shm_size, IPC_CREAT | 0666);
  if (shmid < 0)
  {
    {
      FEC2LOG("%s %d: Can not create shared memory 0x%x. ",
               script_fn[file_index], line_no[file_index], shm_key);
      fflush(logFile);
      perror("");
      exit_code = EXIT_FAILURE;
      return (exit_code);
    }
  }
  debug("%s %d: 0x%x shared memory created.\n",
         script_fn[file_index], line_no[file_index], shm_key);

  // attach to data space
  if ((*shm_addr = shmat(shmid, NULL, 0)) == (unsigned long *)-1)
  {
    FEC2LOG("%s %d: Can not attach  shared memory 0x%x. ",
             script_fn[file_index], line_no[file_index], shm_key);
    fflush(logFile);
    perror("");
    exit_code = EXIT_FAILURE;
    return (exit_code);
  }

  return(exit_code);
}

/******************************************************************************/
/* Attach a Shared Memory Segment                                             */
/******************************************************************************/
int attach_shm(key_t key, int shm_size, unsigned long **shm_addr)
{
  int exit_code, shmid;

  exit_code = 0;

  shmid = shmget(key,  shm_size, 0666);
  if (shmid < 0)
  {
    {
      FEC2LOG("%s %d: Can not get shared memory 0x%x. ",
               script_fn[file_index], line_no[file_index], key);
      fflush(logFile);
      perror("");
      exit_code = EXIT_FAILURE;
      return (exit_code);
    }
  }
  debug("%s %d: Got shared memory 0x%x.\n", 
         script_fn[file_index], line_no[file_index], key);

  // attach to data space
  if ((*shm_addr = shmat(shmid, NULL, 0)) == (unsigned long *)-1)
  {
    FEC2LOG("%s %d: Can not attach  shared memory 0x%x. ",
             script_fn[file_index], line_no[file_index], key);
    fflush(logFile);
    perror("");
    exit_code = EXIT_FAILURE;
    return (exit_code);
  }

  return(exit_code);
}

/******************************************************************************/
/* Detach a Shared Memory Segment                                             */
/******************************************************************************/
int detach_shm(key_t key, unsigned long *shm_addr)
{
  int exit_code, ret;

  exit_code = 0;

  if ((ret = shmdt(shm_addr)) == -1)
  {
    FEC2LOG("%s %d:  Can not detach  shared memory 0x%x. ",
             script_fn[file_index], line_no[file_index], key);
    fflush(logFile);
    perror("");
    exit_code = EXIT_FAILURE;
    return (exit_code);
  }

  return(exit_code);
}

/********************************************************/
int read_ddl_param(__u32 *value, char *string_value) 
{
  __u32 temp;

  temp = strtoul(string_value, (char**)NULL, 16);
  *value = temp & DDL_PARAM_MASK;
  if (*value != temp)
  {
    FEC2LOG("%s %d: FEE command or address value (%s) is too large.\n",
          script_fn[file_index], line_no[file_index], string_value);
    FEC2LOG("                Maximum value can be: 0x%x.\n", DDL_PARAM_MASK);
    return (EXIT_FAILURE);

  }
 
  return (0);
}

/********************************************************/
int read_file(char* fn, char *form, int *len, unsigned long* buffer_addr)
{
  int i, index, he, ret, new_shm, exit_code, fl;
  int have_to_read, n_files, mem_copy, read_len;
//  struct shmid_ds shmid_ds;
  unsigned long *ism_addr;
  FILE *inpf;
  char format[80], real_fn[PATH_MAX];
  struct stat file_stat;
//  ism_stat_t ism_stat;
  ism_stat_t *shmaddr;
  key_t ism_key, file_key;
  time_t file_mtime;
  int file_offset;
  unsigned long* fshm_addr, *file_data_addr;
  int free_offset;
  int free_size;

  exit_code = 0; 
  if (!shm)
   have_to_read = 1;
  else
  {
    // file status
    if (realpath(fn, real_fn) == (char *)NULL)
    {
      FEC2LOG("%s %d: can not find real file name of %s. ",
               script_fn[file_index], line_no[file_index], fn);
      fflush(logFile);
      perror("");
      exit_code = EXIT_FAILURE;
      return (exit_code);
    }
    fl = strlen(real_fn);
    if (fl > MAX_FILE_NAME)
    {
      memmove(real_fn, &real_fn[fl - MAX_FILE_NAME], MAX_FILE_NAME);
      real_fn[MAX_FILE_NAME] = '\0';
      FEC2LOG("%s %d: real file name of %s is truncated to %d tail characters.",
               script_fn[file_index], line_no[file_index], fn, MAX_FILE_NAME);
    }

    ret = stat(fn, &file_stat);
    if (ret)
    {
      FEC2LOG("%s %d: can not open file: %s. ", 
               script_fn[file_index], line_no[file_index], fn);
      fflush(logFile);
      perror("");
      exit_code = EXIT_FAILURE;
      return (exit_code);
    }
    debug("%s %d: file: %s\n                size: %d, last mod: %s", 
             script_fn[file_index], line_no[file_index], real_fn,
             (int)file_stat.st_size, ctime(&file_stat.st_mtime));

    // does ISM exist?
    ism_key = ISM_KEY + 256 * (MAX_DEVICE * minor + channel);
    ret = attach_shm(ism_key, ISM_SIZE, &ism_addr);
    if (ret)
    {
      debug("%s %d: ISM does not exist. I will create it.\n", 
             script_fn[file_index], line_no[file_index]);
      // if not, create it
      exit_code = create_shm(ISM_KEY, ISM_SIZE, &ism_addr);
      if (exit_code)
        return (exit_code);
      *ism_addr = 0;
    }
    n_files = *ism_addr;
    debug("%s %d: number of files in ISM = %d.\n", 
           script_fn[file_index], line_no[file_index], n_files);

    // is this file in ISM?
    file_key = 0;
    shmaddr = (ism_stat_t *)(ism_addr +1);
    for (i = 0; i < n_files; i++)
    {
      debug("%d, %s, %s", i, shmaddr[i].fn, ctime(&shmaddr[i].mtime));
      debug("   key: 0x%x, offs: %d, size: %d\n", 
               (int)shmaddr[i].key,  shmaddr[i].offset, shmaddr[i].wordlen);
      if ((strcmp(real_fn, shmaddr[i].fn) == 0)    &&
          (shmaddr[i].mtime == file_stat.st_mtime))
      {
        file_key = shmaddr[i].key;
        file_offset = shmaddr[i].offset;
        read_len = shmaddr[i].wordlen;
        file_mtime = shmaddr[i].mtime;
        index = i;
        debug("%s %d: file found in ISM. index: %d, key: 0x%x, offset: %d, len: %d\n",
               script_fn[file_index], line_no[file_index],
                                        index, file_key, file_offset, read_len);
        break;
      }
    }
    if (file_key == 0)
    {
      have_to_read = 1;
      debug("%s %d: file %s with the given last mod time not found in ISM.\n",
             script_fn[file_index], line_no[file_index], fn);
    }
    else
    {
      // copy the file from the shared memory
      ret = attach_shm(file_key, SIZE_SHM, &fshm_addr);
      if (ret)
        have_to_read = 1;
      else
      { 
        *len = read_len;
        file_data_addr = fshm_addr + file_offset;
        for (i = 0; i < *len; i++);
          buffer_addr[i] = file_data_addr[i];

        debug("%s %d: %d words read from shared memory 0x%x\n",
               script_fn[file_index], line_no[file_index], *len, file_key);
        he = 16;
        if (*len < he)
        he = *len;
        debug("The first %d words:\n", he);
        for (i = 0; i < he; i++)
          debug("0x%08lx\n", buffer_addr[i]);

        ret = detach_shm(file_key, fshm_addr);
        ret = detach_shm(ISM_KEY, ism_addr);
        return (exit_code);
      }
    } // found in ISM
    file_mtime = file_stat.st_mtime;
  } //shm

  if (have_to_read)
  {
    inpf = fopen(fn, "r");
    if (inpf == NULL)
    {
      FEC2LOG("%s %d: Error opening file: %s. ", 
               script_fn[file_index], line_no[file_index], fn);
      fflush(logFile);
      perror("");
      exit_code = EXIT_FAILURE;
      return (exit_code);
    }
    // read the file
    if (strlen(form))
    {
      strcpy(format, form);
      read_len = getVal(buffer_addr, DDL_MAX_WORD, inpf, format);
      fclose(inpf);
      debug("%s %d: %d words read from file %s\n",
             script_fn[file_index], line_no[file_index], read_len, fn);
    }
    else
    {
      read_len = fread(buffer_addr, 4, DDL_MAX_WORD, inpf);
      fclose(inpf);
      debug("%s %d: %d words read from binary file %s\n",
             script_fn[file_index], line_no[file_index], read_len, fn);
    }
  }

  *len = read_len;
  he = 16;
  if (read_len < he)
    he = read_len;
  debug("The first %d words:\n", he);
  for (i = 0; i < he; i++)
    debug("0x%08lx\n", buffer_addr[i]);

  if (shm)
  {
    // copy the file into a shared memory
    mem_copy = 1;
    // find the last file shm
    new_shm = 0;
    if (!n_files)
    {
      file_key = ism_key + 1;
      new_shm = 1;
    }
    else
    {
      file_key = shmaddr[n_files-1].key;
      ret = attach_shm(file_key, SIZE_SHM, &fshm_addr); 
      if (ret)
        mem_copy = 0;
      else
      {
        debug("%s %d: 0x%x file shm attached.\n",
               script_fn[file_index], line_no[file_index], file_key);
        free_size = *(fshm_addr + 1);
        if (free_size < 4*read_len)
        {
          debug("%s %d: free space %d bytes is not enough.\n",
                 script_fn[file_index], line_no[file_index], free_size);
          ret = detach_shm(file_key, fshm_addr);
          file_key++;
          new_shm = 1;
        }
        else
        {
          free_offset = *fshm_addr;
          file_offset = free_offset;
          file_data_addr = fshm_addr + file_offset;
        }
      }
    }
    if (new_shm)
    {
      //create new file shm
      if ((file_key - ism_key) > MAX_SHM_KEY)
      {
        mem_copy = 0;
        debug("%s %d: too many file shm.\n",
               script_fn[file_index], line_no[file_index]);
      }
      else
      {
        ret = create_shm(file_key, SIZE_SHM, &fshm_addr);
        if (ret)
          mem_copy = 0;
        else
        {
          free_offset = 2; //offset of the free space
          free_size = SIZE_SHM - 2 * sizeof(unsigned long);
          file_offset = free_offset;
          file_data_addr = fshm_addr + file_offset;
          *fshm_addr = free_offset;
          *(fshm_addr + 1) = free_size;
        }
      }
    }

    if (mem_copy)
    {  
      // make the copy
      for (i = 0; i < read_len; i++);
        file_data_addr[i] = buffer_addr[i];
      debug("%s %d: %d words copied to file shm 0x%x at offset %d.\n",
             script_fn[file_index], line_no[file_index], 
             read_len, file_key, file_offset);

      //update free space
      free_size -= 4 * read_len;
      free_offset += read_len;
      *fshm_addr = free_offset;
      *(fshm_addr + 1) = free_size;
      ret = detach_shm(file_key, fshm_addr);

       //update ISM
      strcpy(shmaddr[n_files].fn, real_fn);
      if (n_files >= MAX_SHM_FILE)
      {
        debug("%s %d: too many file in ISM. Can not store shm index.\n",
               script_fn[file_index], line_no[file_index]);
      }
      else
      {
        shmaddr[n_files].key = file_key;
        shmaddr[n_files].offset = file_offset;
        shmaddr[n_files].wordlen = read_len;
        shmaddr[n_files].mtime = file_mtime;
        n_files++;
        *ism_addr = n_files;
      }
    }  
    ret = detach_shm(ism_key, ism_addr);
  }

  return (exit_code);
}

/********************************************************/
int send_command_and_wait_reply(rorc_pci_dev_t *dev,
                                __u32 fee_command,
                                __u32 fee_address, 
                                int timeout,
                                stword_t *stw,
                                int expected)


{
  int i, ret, n_reply;

  ret = ddlSendCommandAndWaitReply(dev, fee_command, fee_address, 
		                   timeout, stw, expected, &n_reply);
  if ((ret == RORC_LINK_NOT_ON) || (ret == RORC_TIMEOUT))
  {
    FEC2LOG("%s %d: Command 0x%x to address 0x%x can not be sent. DDL return code: %s\n",
          script_fn[file_index], line_no[file_index],
	  fee_command, fee_address, rorcInterpretReturnCode(ret));
    return (ret);
  }

  if (ret == RORC_STATUS_OK)
    return (0);

  if (ret == RORC_NOT_ENOUGH_REPLY)
  {
    FEC2LOG("%s %d: Not enough (%d) reply arrived for command 0x%x to address 0x%x in timeout %d\n",
       script_fn[file_index], line_no[file_index],
                          n_reply, fee_command, fee_address, timeout);
  }
  else
  {
    FEC2LOG("%s %d: Too many (%d) replies arrived for command 0x%x to address 0x%x\n",
             script_fn[file_index], line_no[file_index],
                          n_reply, fee_command, fee_address);
  }
  for (i = 0; i < n_reply; i++)
    FEC2LOG(" 0x%08lx\n", stw[i].stw);

  return (ret);
}

/********************************************************/
int read_data_block(rorc_pci_dev_t   *dev,
                    unsigned long    data_phys,
		    unsigned long    return_phys,
		    rorcReadyFifo_t  *rf,
	            __u32            fee_address,
		    int              timeout,
		    stword_t         *stw)
{
  int i, n_reply, step, exit_code = 0;

  exit_code = ddlReadDataBlock(dev, data_phys, return_phys, rf,
		               fee_address, timeout, stw, &n_reply, &step);
  
  if (exit_code == RORC_STATUS_OK)
    return (exit_code);

  if ((exit_code == RORC_LINK_NOT_ON) || 
     ((exit_code == RORC_TIMEOUT) && (step !=2)))
  {
    FEC2LOG("%s %d: Command %s to address 0x%x can not be sent. DDL return code: %s\n",
          script_fn[file_index], line_no[file_index],
          (step == 1) ? "STBRD" : "EOBTR",  fee_address, 
          rorcInterpretReturnCode(exit_code));
    return (exit_code);
  }

  if (exit_code == RORC_TIMEOUT)
  {
    FEC2LOG("%s %d: Data block does not arrive in time-out %d usec.\n",
           script_fn[file_index], line_no[file_index], 
	                                        timeout/dev->loop_per_usec);
    return (exit_code);
  }
  
  if ( exit_code == RORC_NOT_ENOUGH_REPLY)
  {
    FEC2LOG("%s %d: Not enough (%d) reply arrived for command %s to address 0x%x in timeout %d\n",
       script_fn[file_index], line_no[file_index],
               n_reply, (step == 1) ? "STBRD" : "EOBTR", fee_address, timeout);
  }
  else
  {
    FEC2LOG("%s %d: Too many (%d) replies arrived for command %s to address 0x%x\n",
             script_fn[file_index], line_no[file_index],
                          n_reply, (step == 1) ? "STBRD" : "EOBTR", fee_address);
  }
  for (i = 0; i < n_reply; i++)
    FEC2LOG(" 0x%08lx\n", stw[i].stw);

  return (exit_code);
}

/********************************************************/
int main(int argc, char **argv)
{
  int i, j, l, c, n_reply, step, exit_code = EXIT_SUCCESS;
  rorc_pci_dev_t prorc;
  int timeout, ustimeout, talalt, n_tok, max_token = MAX_TOKEN;
  int max_files = MAX_FILES, max_defines = MAX_DEFINES;
  int def_index, c_index, failed_code = 0;
  FILE *script[MAX_FILES];
  char *token[MAX_TOKEN]; 
  char *env_token;
  char tmp_token[80];
  char define[MAX_DEFINES][80];
  char def_val[MAX_DEFINES][80];
  char format[80];
  time_t time_val, time_now;
  struct timeval start_tv, now_tv, tv1, tv2;

  int physmem_fd, volt_offset, volt_serial;
  unsigned long phys_addr, phys_size_physmem;
  unsigned long buffer_phys, return_phys, data_phys;
  unsigned long block_phys;
  unsigned long max_physmem, physmem_offset;
  volatile unsigned long *addr_user;
  unsigned long *buffer_addr;
  volatile unsigned long *return_addr, *data_addr;
  rorcReadyFifo_t *rf;
  unsigned long data_offset;
  
  FILE *text_fp;
  FILE *outpf;
  __u32 fee_command, fee_address, poll_address;
  __u32 fee_expected;
  __u32 fee_mask;
  stword_t stw[DDL_MAX_REPLY];
  int ret, n_val, block_len, load_len;
  int dsec, dusec;
  double dtime;
  int read_timeout, wait_time;

  FILE *i_fp, *o_fp;
  int jtag_len, n_word, sum_word;

  int param;
  char *ddl_part_name[] = {"RORC", "DIU", "SIU", "ALL", ""};
  int ddl_part_reset[] = {RORC_RESET_RORC, 
                          RORC_RESET_DIU, 
                          RORC_RESET_SIU,
                          RORC_RESET_ALL};
#ifdef DEBUG
  unsigned long status;
#endif

  volt_serial = 0;
  revision = 0;
  serial = 0;
  minor = 0;
  channel = 0;
  shm = 0;
  ustimeout = DDL_RESPONSE_TIME;
  def_index = 0;
  file_index = 0;
  strcpy(log_fn, LOG_FN);
  strcpy(script_fn[file_index], SCRIPT_FN);
  volt_offset = 0;
  physmem_offset = 0;
  max_physmem = MAX_PHYSMEM;
 
/* Check parameters */

  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "R:r:N:n:M:m:C:c:T:t:F:f:L:l:O:o:U:u:SsvHh")) != EOF)
    {
      switch(toupper(c))
      {
        case 'R': revision = atoi(optarg); break;
        case 'N': serial = atoi(optarg); volt_serial = 1; break;
        case 'M': minor = atoi(optarg); break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'T': ustimeout = atoi(optarg); break;
        case 'F': strcpy(script_fn[file_index], optarg); break;
        case 'L': strcpy(log_fn, optarg); break;
        case 'O': physmem_offset = atoi(optarg)*1024*1024; volt_offset = 1; 
          break;
        case 'U': max_physmem = atoi(optarg)*1024*1024;
          if (max_physmem < MAX_PHYSMEM)
          {
            printf(" Utilizable physmem size should be at least %d MB.\n",
                                MAX_PHYSMEM/1024/1024);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'S': shm = 1; break;
        case 'V': break;
        default : usage (argv[0]); return -1; break;
      }
    }
  }

  if (strcmp(log_fn, LOG_FN))
  {
    logFile = fopen(log_fn, "a");
    if (logFile == NULL)
    {
      printf("Error opening log file: %s. ", log_fn);
      fflush(stdout);
      perror("");
      exit (EXIT_FAILURE);
    }
  }
  else
    logFile = stdout;

  script[file_index] = fopen(script_fn[file_index], "r");
  if (script[file_index] == NULL)
  {
    FEC2LOG("Error opening script file: %s. ", script_fn[file_index]);
    fflush(logFile);
    perror("");
    exit (EXIT_FAILURE);
  }
  line_no[file_index] = 0;

  /* sw version */
  FEC2LOG("Starting Front-end Control and Configuration (FeC2) program.\n");
  for (i = 0; i < argc; i++)
    FEC2LOG("%s ", argv[i]);
  FEC2LOG("\nCompiled RORC sw version: %s.\n", RORC_DRIVER_VERSION);

  /* find the RORC */
  if (volt_serial)
  {
    if ((revision < 1) || (revision > RORC_MAX_REVISION)) 
    {
      FEC2LOG("Error calling FeC2: RORC revision number %d is not permitted.\n", 
                                                               revision);
      exit(EXIT_FAILURE);
    }

    if (rorcFind(revision, serial, &minor) != RORC_STATUS_OK)
    {
      FEC2LOG("Error calling FeC2: RORC device with revision %d and serial %d\n"
              "                    not found or an other process uses it.\n",
                                revision, serial);
      exit (EXIT_FAILURE);
    }
    debug(" Rorc device with revision number %d"
                       " and serial number %d"
          " found. Minor = %d.\n", revision, serial, minor );
  }

  /* open RORC and map mailboxes */
  if (rorcOpenChannel(&prorc, minor, channel))
  {
    FEC2LOG("Error opening RORC device.\n");
    exit (EXIT_FAILURE);
  }
  timeout = ustimeout * prorc.loop_per_usec;
  if (timeout > DDL_MAX_WAIT_CYCLE )
  {
    FEC2LOG("Error: time-out value %d us is too large.\n", ustimeout);
    FEC2LOG("Please choose a value less than or equal to %d us.\n",
                                                       prorc.max_resp_time);
    exit (EXIT_FAILURE);
  }


  /* physmem */ 
  if (physmemOpen (&physmem_fd, &addr_user, &phys_addr,
			                       &phys_size_physmem))
  {
    FEC2LOG("Error opening physmem device.\n");
    exit (EXIT_FAILURE);
  }
        
  debug(" Allocation of the physmem memory.");
  debug(" Physical address: 0x%08lx.\n", phys_addr);
  debug(" Size of the physmem: %ld (0x%08lx = %ld MB) bytes,\n"
        "                      %ld (0x%08lx = %ld MW) words.\n",
             phys_size_physmem, phys_size_physmem, phys_size_physmem/1024/1024,
             phys_size_physmem >> 2, phys_size_physmem >> 2, 
                                    (phys_size_physmem >> 2)/1024/1024);

  if (!volt_offset)
    physmem_offset = (minor * MAX_CHANNEL + channel) *  MAX_PHYSMEM;

  if ((physmem_offset + max_physmem) > phys_size_physmem)
  {
    printf("Error: utilizable memory offset (%ld MB) + size (%ld MB) must be "
           "less than \n       the physmem area physiscal size (%ld MB).\n",
                     physmem_offset/1024/1024, max_physmem/1024/1024, 
		     phys_size_physmem/1024/1024);
    usage (argv[0]);
    physmemClose (physmem_fd, (unsigned long *)addr_user, phys_size_physmem);
    rorcClose(&prorc);
    exit (EXIT_FAILURE);
  }

  phys_addr += physmem_offset;
  addr_user += physmem_offset/4;

  debug(" Start address of utilizable memory: 0x%08lx (%p),\n" 
        "                             offset: 0x%08lx (%ld MB).\n",
             phys_addr, addr_user, physmem_offset, physmem_offset/1024/1024);
  debug(" Size of the utilizable memory: %ld (%ld MB) bytes," 
             " %ld (%ld MW) words.\n",
           max_physmem, max_physmem/1024/1024, 
           max_physmem >> 2, (max_physmem >> 2)/1024/1024);

  return_phys = phys_addr;
  data_phys   = return_phys + 0x40;
  buffer_phys = data_phys + DDL_MAX_BYTE + 4;

  return_addr = addr_user;
  data_addr   = return_addr + 0x10;
  buffer_addr = (unsigned long *)data_addr + DDL_MAX_WORD + 1;
  rf = (rorcReadyFifo_t *)return_addr;
  
  time_now = time(&time_val);
  gettimeofday(&tv1, NULL);
  FEC2LOG("\nStart processing script file %s at %s\n", script_fn[file_index],
              ctime(&time_val));

  while(1)
  {
    n_tok = getToken(token, max_token, script[file_index]);
    if (!n_tok)
    {
      FEC2LOG("%s %d: EOF encountered\n", 
               script_fn[file_index], line_no[file_index]);
      if (file_index == 0)
        goto vege;
      fclose(script[file_index]);
      file_index--;
      continue;
    }
    line_no[file_index]++;
    debug("\n%s %d: ", script_fn[file_index], line_no[file_index]);
    for (i = 0; i < n_tok; i++)
      debug("%s; ", token[i]);

    c_index = n_key_word(command_tokens, token[0]);
//    debug("- %d tokens, command index: %d", n_tok, c_index);
    debug("\n");
    if (!c_index)
    {
      FEC2LOG("%s %d: Invalid command: %s\n", 
               script_fn[file_index], line_no[file_index], token[0]);
      exit_code = EXIT_FAILURE;
      goto vege;
    }

    if ((n_tok - 1) < command_param[c_index])
    {
      FEC2LOG("%s %d: parameter missing for command %s\n",
               script_fn[file_index], line_no[file_index], token[0]);
      exit_code = EXIT_FAILURE;
      goto vege;
    }

    if (n_tok > 1)
    {
      /* check for environment variable substitution */
      for (i = 1; i < n_tok; i++)
      {
        if (token[i][0] == '$')
        {
          j = 1;
          if ((token[i][1] == '{') || (token[i][1] == '('))
            j = 2;
          l = strlen(token[i]) - 2 * j + 1;
          strncpy(tmp_token, token[i]+j, l);
          tmp_token[l] = '\0';
//          debug("tmp_token: %s;\n", tmp_token);
          env_token = getenv(tmp_token);
          if (env_token == NULL) 
          {
            FEC2LOG("%s %d: Undefined environment variable %s\n",
                     script_fn[file_index], line_no[file_index], token[i]);
            exit_code = EXIT_FAILURE;
            goto vege;
          } 
          token[i] = env_token;
          debug("%s %d: ", script_fn[file_index], line_no[file_index]);
          for (j = 0; j < n_tok; j++)
            debug("%s; ", token[j]);
          debug("\n");
        }
      }
    }

    if (def_index && (n_tok > 1) && (c_index != DEFINE))
    {
      /* check defines for substitution */

      talalt = 0;
      for (i = 1; i < n_tok; i++)
      {
        for (j = 0; j < def_index; j++)
        {
          if (!strcmp(define[j], token[i]))
          {
            token[i] = def_val[j]; 
	    talalt = 1;
            break;
          }
        }
      }
      if (talalt)
      {
        debug("%s %d: ", script_fn[file_index], line_no[file_index]);
        for (i = 0; i < n_tok; i++)
          debug("%s; ", token[i]);
        debug("\n");
      }
    }

    /* process commands */

    switch (c_index)
    {
      case DEFINE: 
        if (def_index >= max_defines)
        {
          FEC2LOG("%s %d: Too many DEFINEs\n",
                           script_fn[file_index], line_no[file_index]);
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        for (j = 0; j < def_index; j++)
        {
          if (!strcmp(define[j], token[1]))
          {
            FEC2LOG("%s %d: Redefinition of %s\n",
                     script_fn[file_index], line_no[file_index], token[1]);
            exit_code = EXIT_FAILURE;
            goto vege;
          }
        }

        strcpy(define[def_index], token[1]);
        strcpy(def_val[def_index], token[2]);
        def_index++;
        break;
      case RESET: 
        param = 0;
        if (n_tok > 1)
        {
          param = n_key_word(ddl_part_name, token[1]);
          if (param == 0)
          {
            FEC2LOG("%s %d: Invalid reset parameter: %s\n",
                     script_fn[file_index], line_no[file_index], token[1]);
            exit_code = EXIT_FAILURE;
            goto vege;
          }
          param--;
        }
	if (param < 3)
        {
          rorcReset(&prorc, ddl_part_reset[param]);
          debug("%s %d: Reset %s done\n", script_fn[file_index], 
			           line_no[file_index], ddl_part_name[param]);
	}
	else
        {
          rorcReset(&prorc, ddl_part_reset[0]);    // RORC
          debug("%s %d: Reset %s done\n",
                 script_fn[file_index], line_no[file_index], ddl_part_name[0]);
          rorcReset(&prorc, ddl_part_reset[2]);    // SIU
          debug("%s %d: Reset %s done\n",
                 script_fn[file_index], line_no[file_index], ddl_part_name[2]);
          rorcReset(&prorc, ddl_part_reset[1]);    // DIU
          debug("%s %d: Reset %s done\n",
                 script_fn[file_index], line_no[file_index], ddl_part_name[1]);
          rorcReset(&prorc, ddl_part_reset[0]);    // RORC
          debug("%s %d: Reset %s done\n",
                 script_fn[file_index], line_no[file_index], ddl_part_name[0]);
	}
        break;
      case READ_DDL_STATUS: 
#ifdef DEBUG
        /* DIU status */
        printf("\nRead DIU status:\n");
        status = ddlReadDiu(&prorc, 0, timeout);
        if (status != -1)
        {
          printf(" DIU status word: 0x%08lx, i.e.\n", status);
          ddlInterpretIFSTW(&prorc, status);
        }

        /* SIU status*/
        printf("\nRead SIU status:\n");
        status = ddlReadSiu(&prorc, 0, timeout);
        if (status != -1)
        {
          printf(" SIU status word: 0x%08lx, i.e.\n", status);
          ddlInterpretIFSTW(&prorc, status);
        }
#else
        FEC2LOG("%s %d: use switch -v for using READ_DDL_STATUS command\n",
                script_fn[file_index], line_no[file_index]);
#endif
        printf("\n");
        break;
      case WRITE_RDYRX: 
        exit_code = send_command_and_wait_reply(&prorc, RDYRX,
                                0, timeout, stw, 1);
        if (exit_code)
	{
	  exit_code = EXIT_FAILURE;
          goto vege;
	}
        break;
      case WRITE_EOBTR: 
        exit_code = send_command_and_wait_reply(&prorc, EOBTR,
                                0, timeout, stw, 1);
        if (exit_code)
        {
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        break;
      case WRITE_COMMAND: 
        exit_code = read_ddl_param(&fee_command, token[1]);
        if (exit_code)
         goto vege;
	/*
        for ( j = 18*33; --j != 0; ) {
	*/
         exit_code = send_command_and_wait_reply(&prorc, FECTRL,
                                fee_command, timeout, stw, 1);
         if (exit_code)
	 {
	  exit_code = EXIT_FAILURE;
          goto vege;
	 }
	 /*
        }
	 */
        break;
      case WRITE_BLOCK: 
        exit_code = read_ddl_param(&fee_address, token[1]);
        if (exit_code)
          goto vege;
	if (n_tok > 3)
          exit_code = read_file(token[2], token[3], &n_val, buffer_addr);
	else
          exit_code = read_file(token[2], "", &n_val, buffer_addr);
	if (exit_code)
	  goto vege;

	exit_code = ddlWriteDataBlock(&prorc, buffer_phys, n_val, 
        	return_phys, return_addr, fee_address, timeout, stw, 
		&n_reply, &step);
        if (exit_code)
	{
          FEC2LOG("%s %d: Data block can not be sent. DDL return code after step %d: %s\n",
                script_fn[file_index], line_no[file_index], step, 
         	 rorcInterpretReturnCode(exit_code));
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        break;
      case WRITE_BLOCK_MULTIPLE:
        exit_code = read_ddl_param(&poll_address, token[1]);
        if (exit_code)
          goto vege;
        exit_code = read_ddl_param(&fee_expected, token[2]);
        if (exit_code)
          goto vege;
        fee_mask = strtoul(token[3], (char**)NULL, 16);
        read_timeout = atoi(token[4]);
        exit_code = read_ddl_param(&fee_address, token[5]);
        if (exit_code)
          goto vege;
        block_len = atoi(token[6]);
	if (n_tok > 8)
          exit_code = read_file(token[7], token[8], &n_val, buffer_addr);
	else
          exit_code = read_file(token[7], "", &n_val, buffer_addr);
	if (exit_code)
	  goto vege;

        i = 0;
        block_phys = buffer_phys;
        while (n_val > 0)
        {
          i++;
          load_len = block_len > n_val ? n_val : block_len;
	  exit_code = ddlWriteDataBlock(&prorc, block_phys, load_len, 
          	return_phys, return_addr, fee_address, timeout, stw, 
	  	&n_reply, &step);
          if (exit_code)
	  {
            FEC2LOG("%s %d: Data block %d can not be sent. DDL return code after step %d: %s\n",
                    script_fn[file_index], line_no[file_index], i, step, 
           	    rorcInterpretReturnCode(exit_code));
            exit_code = EXIT_FAILURE;
            goto vege;
          }
          debug("%s %d: Data block %d of length %d words successfuly sent.\n",
                 script_fn[file_index], line_no[file_index], i, load_len); 
          debug("  source address: 0x%08lx, FEE address 0x%08x.\n", 
                                              block_phys, fee_address); 
          gettimeofday(&start_tv, NULL);
          failed_code = 0;
          dtime = 0;
          while (dtime < read_timeout)
          {
            exit_code = send_command_and_wait_reply(&prorc, FESTRD,
                                       poll_address, timeout, stw, 2);
            if (exit_code)
            {
              exit_code = EXIT_FAILURE;
              goto vege;
            }
            debug("%s %d: Value read: 0x%x, value expected: 0x%x, mask: 0x%x\n",
                       script_fn[file_index], line_no[file_index],
                       stw[0].part.param, fee_expected, fee_mask);
            if ((stw[0].part.param & fee_mask) == fee_expected)
              break;
            gettimeofday(&now_tv, NULL);
            elapsed(&now_tv, &start_tv, &dsec, &dusec);
            dtime = (double)dsec * MEGA + (double)dusec;
          }
          if ((stw[0].part.param & fee_mask) != fee_expected)
          {
            FEC2LOG("%s %d: Block %d not accepted in timeout %d microsec! ",
                   script_fn[file_index], line_no[file_index], i, read_timeout);
            FEC2LOG("Value read: 0x%x, value expected: 0x%x, mask: 0x%x.\n",
                     stw[0].part.param, fee_expected, fee_mask);
            exit_code = EXIT_FAILURE;
            failed_code = 1;
            goto vege;
          }
          fee_address += load_len;
          block_phys += 4 * load_len;
          n_val -= load_len;
        }
        break;
      case WRITE_JTAG:
        i_fp = fopen(token[1], "r");
        if (i_fp == NULL)
        {
          FEC2LOG("%s %d: Error opening input file: %s. ", 
                   script_fn[file_index], line_no[file_index], token[2]);
          fflush(logFile);
          perror("");
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        o_fp = fopen(token[2], "w");
        if (o_fp == NULL)
        {
          FEC2LOG("%s %d: Error opening output file: %s. ", 
                   script_fn[file_index], line_no[file_index], token[3]);
          fflush(logFile);
          perror("");
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        jtag_len = DDL_MAX_JTAG_LEN + 1;
	if (n_tok > 3)
          jtag_len = atoi(token[3]);
        if (jtag_len > DDL_MAX_JTAG_LEN)
          jtag_len = DDL_MAX_JTAG_LEN;

        data_offset = jtag_len/4 + 2;
        data_phys = phys_addr + (data_offset << 2);
        data_addr = addr_user + data_offset;
        return_phys = phys_addr;
        return_addr = addr_user;

        /* send start data receiver */
        rorcStartDataReceiver(&prorc, return_phys);

        sum_word = 0;
        while (1)
        {
          /* read JTAG data from file */
          n_word = JTAG2Memory(i_fp, data_addr, 8*jtag_len);
          if (n_word == 0)
            break;
      
          /* send JTAG */
          exit_code = ddlWriteJTAGBlock(&prorc, data_phys, n_word,
                                  return_phys, return_addr, param,
                                  timeout, stw, &n_reply, &step);
      
          debug("%s %d: ddlWriteJTAGBlock sent %d words, return code: %s, step = %d.\n", 
                        script_fn[file_index], line_no[file_index],
                        n_word, rorcInterpretReturnCode(exit_code), step);
          for (i = 0; i < n_reply; i++)
            debug(" received status %d: 0x%08lx\n", i+1, stw[i].stw);
          if (exit_code)
            break;

          /* write result to file */
          Memory2JTAG(o_fp, data_addr, return_addr, n_word);

          /* loop */
          sum_word += n_word;
          data_phys += (n_word << 2);
          data_addr += n_word;
        }

        /* send stop data receiver */
        rorcStopDataReceiver(&prorc);
	debug("%s %d: %d JTAG values sent and read back.\n", 
	    script_fn[file_index], line_no[file_index], 8*sum_word);
	fclose(i_fp);
	fclose(o_fp);
        if (exit_code)
          goto vege;
        break;
      case READ_AND_PRINT: 
        exit_code = read_ddl_param(&fee_address, token[1]);
        if (exit_code)
         goto vege;
        exit_code = send_command_and_wait_reply(&prorc, FESTRD,
                                fee_address, timeout, stw, 2);
        if (exit_code)
	{
	  exit_code = EXIT_FAILURE;
          goto vege;
	}
	if (n_tok > 3)
	{
	  text_fp = fopen(token[3], "a");
          if (text_fp == NULL)
          {
            FEC2LOG("%s %d: Error opening file: %s. ", 
		     script_fn[file_index], line_no[file_index], token[3]);
	    fflush(logFile);
	    perror("");
            exit_code = EXIT_FAILURE;
            goto vege;
          }
	  strcpy(format, token[2]);
	  strcat(format, "\n");
	  fprintf(text_fp, format, stw[0].part.param);
	  fclose(text_fp);
	}
	else
	{
          strcpy(format, "%s %d: ");
	  strcat(format, token[2]);
	  strcat(format, "\n");
          FEC2LOG(format, script_fn[file_index], line_no[file_index],
                       stw[0].part.param);
	}
        break;
      case READ_AND_CHECK:
        exit_code = read_ddl_param(&fee_address, token[1]);
        if (exit_code)
          goto vege;
        exit_code = read_ddl_param(&fee_expected, token[2]);
        if (exit_code)
         goto vege;
        fee_mask = strtoul(token[3], (char**)NULL, 16);
        exit_code = send_command_and_wait_reply(&prorc, FESTRD,
                                     fee_address, timeout, stw, 2);
        if (exit_code)
	{
	  exit_code = EXIT_FAILURE;
          goto vege;
	}
        debug("%s %d: Value read: 0x%x, value expected: 0x%x, mask: 0x%x\n",
                   script_fn[file_index], line_no[file_index], 
                   stw[0].part.param, fee_expected, fee_mask);
	failed_code = 0;
        if ((stw[0].part.param & fee_mask) != fee_expected)
        {
          FEC2LOG("%s %d: NO MATCH! ", 
                   script_fn[file_index], line_no[file_index]);
          FEC2LOG("Value read: 0x%x, value expected: 0x%x, mask: 0x%x.\n",
                   stw[0].part.param, fee_expected, fee_mask);
          exit_code = EXIT_FAILURE;
	  failed_code = 1;
        }
        break;
      case READ_UNTIL: 
        exit_code = read_ddl_param(&fee_address, token[1]);
        if (exit_code)
         goto vege;
        exit_code = read_ddl_param(&fee_expected, token[2]);
        if (exit_code)
         goto vege;
        fee_mask = strtoul(token[3], (char**)NULL, 16);
        read_timeout = atoi(token[4]);
        gettimeofday(&start_tv, NULL);
	failed_code = 0;
        dtime = 0;
        while (dtime < read_timeout)
        {
          exit_code = send_command_and_wait_reply(&prorc, FESTRD,
                                     fee_address, timeout, stw, 2);
          if (exit_code)
	  {
	    exit_code = EXIT_FAILURE;
            goto vege;
	  }
          debug("%s %d: Value read: 0x%x, value expected: 0x%x, mask: 0x%x\n",
                     script_fn[file_index], line_no[file_index], 
                     stw[0].part.param, fee_expected, fee_mask);
          if ((stw[0].part.param & fee_mask) == fee_expected)
            break;
          gettimeofday(&now_tv, NULL);
	  elapsed(&now_tv, &start_tv, &dsec, &dusec);
          dtime = (double)dsec * MEGA + (double)dusec;
        }
        if ((stw[0].part.param & fee_mask) == fee_expected)
          break;
        FEC2LOG("%s %d: NO MATCH in timeout %d microsec! ", 
                 script_fn[file_index], line_no[file_index], read_timeout);
        FEC2LOG("Value read: 0x%x, value expected: 0x%x, mask: 0x%x.\n",
                 stw[0].part.param, fee_expected, fee_mask);
        exit_code = EXIT_FAILURE;
	failed_code = 1;
	break;
      case READ_BLOCK: 
        exit_code = read_ddl_param(&fee_address, token[1]);
        if (exit_code)
         goto vege;
        exit_code = read_data_block(&prorc, data_phys, return_phys, rf,
                 	            fee_address, timeout, stw);
        if (exit_code)
          goto vege;
        if (rf->status && ((rf->status & 0xFF) != DTSW))
        {
          FEC2LOG("%s %d: Protocol error: DTSTW = 0x%08lx.\n",
          script_fn[file_index], line_no[file_index], rf->status);
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        if ((rf->length < 0) || (rf->length > DDL_MAX_TX_WORD))
        {
          FEC2LOG("%s %d: Protocol error: Data length = %ld (0x%08lx).\n",
          script_fn[file_index], line_no[file_index], rf->length, rf->length);
          exit_code = EXIT_FAILURE;
          goto vege;
        }

	// write to the file 
        outpf = fopen(token[2], "w");
        if (outpf == NULL)
        {
          FEC2LOG("%s %d: Error opening file: %s. ", 
	    script_fn[file_index], line_no[file_index], token[2]);
	  fflush(logFile);
	  perror("");
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        time_now = time(&time_val);
	n_val = rf->length;
        if (n_tok > 3)
	{
	  fprintf(outpf, "# %s %s", token[2], ctime(&time_now));
	  strcpy(format, token[3]);
	  strcat(format, " ");
	  for (i = 0; i < n_val; i++)
	  {
            fprintf(outpf, format, data_addr[i]);
	    if (!((i+1) % 8))
	      fprintf(outpf, "\n");
          } 
	  ret = i;
	}
        else
	  ret = fwrite((unsigned long *)data_addr, 4, n_val, outpf);
	fclose(outpf);
	debug("%s %d: %d words read, %d written to file %s.\n",
	    script_fn[file_index], line_no[file_index], n_val, ret, token[2]);
        if (ret != n_val)
        {
          FEC2LOG("%s %d: can not write %ld word, only %d ones.\n",
          script_fn[file_index], line_no[file_index], 
                                                (unsigned long)n_val, ret);
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        break;
      case READ_AND_CHECK_BLOCK: 
        exit_code = read_ddl_param(&fee_address, token[1]);
        if (exit_code)
         goto vege;
	if (n_tok > 3)
          exit_code = read_file(token[2], token[3], &n_val, buffer_addr);
	else
          exit_code = read_file(token[2], "", &n_val, buffer_addr);
	if (exit_code)
	  goto vege;

        exit_code = read_data_block(&prorc, data_phys, return_phys, rf,
                 	            fee_address, timeout, stw);
        if (exit_code)
          goto vege;
        if (rf->status && ((rf->status & 0xFF) != DTSW))
        {
          FEC2LOG("%s %d: Protocol error: DTSTW = 0x%08lx.\n",
          script_fn[file_index], line_no[file_index], rf->status);
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        if ((rf->length < 0) || (rf->length > DDL_MAX_TX_WORD))
        {
          FEC2LOG("%s %d: Protocol error: Data length = %ld (0x%08lx).\n",
          script_fn[file_index], line_no[file_index], rf->length, rf->length);
          exit_code = EXIT_FAILURE;
          goto vege;
        }

	// compare data
	failed_code = 0;
	if (rf->length > n_val)
	{
	  FEC2LOG("%s %d: More data (%ld) read than expected (%d)\n",
	      script_fn[file_index], line_no[file_index], rf->length, n_val);
	  exit_code = EXIT_FAILURE;
	  failed_code = 1;
	}
	if (rf->length < n_val)
	{
	  FEC2LOG("%s %d: Less data (%ld) read than expected (%d)\n",
	      script_fn[file_index], line_no[file_index], rf->length, n_val);
	  n_val = rf->length;
          exit_code = EXIT_FAILURE;
	  failed_code = 1;
	}
	for (i = 0; i < n_val; i++)
	{
	  if (data_addr[i] != buffer_addr[i])
	  {
	    FEC2LOG("%s %d: Compare error at word %d.\n",
                     script_fn[file_index], line_no[file_index], i);
			    
	    FEC2LOG("       Received: 0x%08lx, expected: 0x%08lx\n",
			    data_addr[i], buffer_addr[i]);
	    exit_code = EXIT_FAILURE;
	    failed_code = 1;
	  }
	}
        break;
      case WAIT:
        wait_time = atoi(token[1]);
        gettimeofday(&start_tv, NULL);
        failed_code = 0;
        usleep(wait_time);
        gettimeofday(&now_tv, NULL);
	elapsed(&now_tv, &start_tv, &dsec, &dusec);
        dtime = (double)dsec * MEGA + (double)dusec;
        debug("%s %d: The program slept %.0f useconds (insteed of %d).\n",
                     script_fn[file_index], line_no[file_index], 
                     dtime, wait_time);
        break;
      case CALL_FILE:
        for (i = 0; i <= file_index; i++)
        {
          if (!strcmp(script_fn[i], token[1]))
          {
            FEC2LOG("%s %d: Recursive call of script file %s\n",
                    script_fn[file_index], line_no[file_index], token[1]);
            exit_code = EXIT_FAILURE;
            goto vege;
          }
        }
        file_index++;
        if (file_index >= max_files)
        {
          file_index--;
          FEC2LOG("%s %d: Too many script files\n", 
                           script_fn[file_index], line_no[file_index]);
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        strcpy(script_fn[file_index], token[1]);
        script[file_index] = fopen(script_fn[file_index], "r");
        if (script[file_index] == NULL)
        {
          FEC2LOG("Error opening script file: %s. ", script_fn[file_index]);
          fflush(logFile);
          perror("");
          file_index--;
          exit_code = EXIT_FAILURE;
          goto vege;
        }
        line_no[file_index] = 0;
        break;
      case RETURN: 
        FEC2LOG("%s %d: RETURN encountered\n", 
                script_fn[file_index], line_no[file_index]);
        if (file_index == 0)
          goto vege;
        fclose(script[file_index]);
        file_index--;
        break;
      case STOP_IF_FAILED:
        debug("%s %d: STOP_IF_FAILED encountered. Failed code: %d\n", 
             script_fn[file_index], line_no[file_index], failed_code);
	if (failed_code)
	{
	  if (n_tok > 1)
	    exit_code = atoi(token[1]);
          goto vege;
	}
        break;
      case STOP:
	if (n_tok > 1)
	  exit_code = atoi(token[1]);
        debug("%s %d: STOP encountered. Exit code: %s\n", 
                script_fn[file_index], line_no[file_index], 
		 rorcInterpretReturnCode(exit_code));
        goto vege;
        break;
      default:
        FEC2LOG("%s %d: Invalid command: %s\n", 
                 script_fn[file_index], line_no[file_index], token[0]);
        exit_code = EXIT_FAILURE;
        goto vege;
    }
  }

vege:
  gettimeofday(&tv2, NULL);
  elapsed(&tv2, &tv1, &dsec, &dusec);

  for (i = file_index; i >= 0; i--)
  {
    fclose(script[i]);
    if (!line_no[i])
      FEC2LOG("%s: Script file empty\n", script_fn[i]);
  }

  /* unmap and close physmem memory */
  physmemClose (physmem_fd, (unsigned long *)(addr_user - physmem_offset/4), 
                                                            phys_size_physmem);

  /* unmap and close RORC */
  rorcClose(&prorc);

  FEC2LOG("\nProcessing of the Front-end Control and Configuration script ended.\n");
  FEC2LOG("Processing time = %d.%06d seconds\n", dsec, dusec);
  FEC2LOG("Exit code = %d\n", exit_code);
  exit (exit_code);
}
