/***********************************************************
 * clean_shm   
 *
 * Clean shared memory segments created by FeC2 program
 *
 * last updated: 30/04/2004
 * written by: Ervin Denes
 *                                                           
 * history:                                           
 *                                                   
 * v.4.3.1  30-Apr-2003: original version                      
 *                                                           
 **********************************************************/
#define SOURCE_FILE "clean_shm"
#define DESCRIPTION "Clean shared memories used by FeC2"

#define LOG_FN "stdout stream"

#define ISM_KEY  0xFeC20000
#define ISM_SIZE 0x00400000 
#define MAX_SHM_FILE 15420
#define MAX_FILE_NAME 255
#define MAX_SHM_KEY 127
#define SIZE_SHM 0x400000 

#include "rorc_ddl.h"

int minor, channel, check;

/*******************************************************/
void usage(char *name)
{
  printf("Clean shared memory segments used by FeC2.\n");
  printf("usage: %s [-{M|m} <minor>] [-{C|c} <channel>] [-X|-x]\n", name);
  printf("   -M <minor>       RORC minor number, default: 0.\n");
  printf("   -C <channel>     RORC channel number, default: 0.\n");
  printf("   -X               Do not delete only check shared memories.\n");
}

/******************************************************************************/
/* Get a Shared Memory Segment                                             */
/******************************************************************************/
int get_shm(key_t key, int shm_size, int *shmid)
{
  if( (*shmid = shmget(key, shm_size, 0666)) < 0)
  {
    printf(" Can not get shared memory 0x%x. ", key);
    fflush(stdout);
    perror("");
    return (EXIT_FAILURE);
  }

  printf(" Got shared memory 0x%x.\n", key);
  return(0);
}

/******************************************************************************/
/* Attach a Shared Memory Segment                                             */
/******************************************************************************/
int attach_shm(key_t key, int shmid, unsigned long **shm_addr)
{
  // attach to data space
  if ((*shm_addr = shmat(shmid, NULL, 0)) == (unsigned long *)-1)
  {
    printf(" Can not attach  shared memory 0x%x. ", key);
    fflush(stdout);
    perror("");
    return (EXIT_FAILURE);
  }

  printf(" Shared memory 0x%x attached.\n", key);
  return(0);
}

/******************************************************************************/
/* Detach a Shared Memory Segment                                             */
/******************************************************************************/
int detach_shm(key_t key, unsigned long *shm_addr)
{
  if (shmdt(shm_addr) == -1)
  {
    printf(" Can not detach  shared memory 0x%x. ", key);
    perror("");
    return (EXIT_FAILURE);
  }

  printf(" Shared memory 0x%x detached.\n", key);
  return(0);
}

/******************************************************************************/
/* Get information on Shareed Memory Segment                                  */
/******************************************************************************/
int stat_shm(key_t key, int shmid, struct shmid_ds *shmid_ds)
{
  if ((shmctl(shmid, IPC_STAT, shmid_ds)) == -1)
  {
    printf(" Can get SHM stat on shared memory 0x%x. ", key);
    perror("");
    return (EXIT_FAILURE);
  }

  return(0);
}

/******************************************************************************/
/* Remove a Shared Memory Segment                                             */
/******************************************************************************/
int remove_shm(key_t key, int shmid, struct shmid_ds *shmid_ds)
{

  if ((shmctl(shmid, IPC_RMID, shmid_ds)) == -1)
  {
    printf(" Can not remove shared memory 0x%x. ", key);
    perror("");
    return (EXIT_FAILURE);
  }

  printf(" Shared memory 0x%x removed\n.", key);
  return(0);
}

/********************************************************/
int main(int argc, char **argv)
{
  int i, j, c;
  int n_files, n_shm, deleted;
  int ism_id, shm_id;
  struct shmid_ds shmid_ds;
  ism_stat_t *shmaddr;
  unsigned long *ism_addr;
  key_t ism_key, file_key;
  int minor, channel;
  key_t fkey[MAX_SHM_KEY];

  minor = 0;
  channel = 0;
  check = 0;

  /* Check parameters */

  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:Xx")) != EOF)
    {
      switch(toupper(c))
      {
        case 'M': minor = atoi(optarg); break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'X': check = 1; break;
        default : usage (argv[0]); return -1; break;
      }
    }
  }

  for (i = 0; i < MAX_SHM_KEY; i++)
    fkey[i] = 0;

  // does ISM exist?
  ism_key = ISM_KEY + 256 * (MAX_DEVICE * minor + channel);
  if (get_shm(ism_key, ISM_SIZE, &ism_id))
    exit (EXIT_FAILURE);

  if (attach_shm(ism_key, ism_id, &ism_addr))
    exit (EXIT_FAILURE);

  if (stat_shm(ism_key, ism_id, &shmid_ds) == 0)
  {
    printf(" Number of current attaches of ISM 0x%x: %d.\n", 
                                           ism_key, (int)shmid_ds.shm_nattch);
    printf(" Segment size: %d bytes,", shmid_ds.shm_segsz);
    printf (" last change time: %s", ctime(&shmid_ds.shm_ctime)); 
  }

  n_files = *ism_addr;
  if (n_files < 1)
  {
    printf(" Number of files in ISM: %d, nothing to do.\n", n_files);
    exit (EXIT_FAILURE);
  }
  printf(" Number of files in ISM = %d.\n", n_files);

  file_key = 0;
  n_shm = 0;
  shmaddr = (ism_stat_t *)(ism_addr +1);
  for (i = 0; i < n_files; i++)
  {
    printf("%d, %s, %s", i, shmaddr[i].fn, ctime(&shmaddr[i].mtime));
    printf("   key: 0x%x, offs: %d, size: %d\n",
              (int)shmaddr[i].key,  shmaddr[i].offset, shmaddr[i].wordlen);
    file_key = shmaddr[i].key;
    debug(" File found in ISM. index: %d, key: 0x%x\n", index, file_key);

    deleted = 0;
    for (j = 0; j < n_shm; j++)
    {
      if (fkey[j] == file_key)
      {
        if (!check)
          printf(" SHM 0x%x already deleted\n", file_key);
        deleted = 1;
        break;
      }
    }
    if (!deleted)
    {
      fkey[n_shm] = file_key;
      if (n_shm < MAX_SHM_KEY)
        n_shm++;
      if (get_shm(file_key, SIZE_SHM, &shm_id))
        continue;

      if (stat_shm(file_key, shm_id, &shmid_ds) == 0)
      {
        printf(" Number of current attaches of SHM 0x%x: %d.\n",
                                           file_key, (int)shmid_ds.shm_nattch);
        printf(" Segment size: %d bytes,", shmid_ds.shm_segsz);
        printf (" last change time: %s", ctime(&shmid_ds.shm_ctime));
      }
      if (!check)
      {
        // shm torles
        remove_shm(file_key, shm_id, &shmid_ds);
      }
    } // if (!deleted)
  } //for

  detach_shm(ism_key, ism_addr);
  if (check)
    printf("Scan of shared memories finished.\n");
  else
  {
    remove_shm(ism_key, ism_id, &shmid_ds);
    printf("Cleaning of shared memories finished.\n");
  }

  exit (EXIT_SUCCESS);
}
