/**********************************************************
 *
 * ddg.c
 *
 * DDL Data Generator program                
 *
 * last updated: 02/05/2006
 * written by: Ervin Denes
 *
 * history:
 *
 * v.4.4.5    02-May-2006: correcting the printout of the stop time
 *            17-Mar-2006: accept RORC revision 4
 *            17-Feb-2006: check time-out option 
 * v.4.4.3    05-Sep-2005: pluggble D-RORC is allowed in case of sw trigger
 * v.4.3.14   20-Mar-2005: adding BUNCH_CROSSING_INCREMENT keyword
 *            16-Mar-2005: changes in bunch crossing algorithm
 * v.4.3.11   07-Dec-2004: original version
 ***********************************************************/
#define SOURCE_FILE "ddg.c"
#define DESCRIPTION "DDL Data Generator Program"

#define CONF_FN "ddg.conf"
#define LOG_FN "stdout stream"
#define DATA_FN "ddg.data"
#define MAX_TOKEN 3

#define SMI_OBJECT "DDG::DDG"

/* sizes are given in bytes */
#define MAX_NAME 81
#define MIN_PHYSMEM (1024*1024*8)
#define DEFAULT_DATA_LENGTH (1024*512 - 1)

#define DDL_NO_BLOCK_LENGTH 0xffffffff
#define BUNCH_PER_ORBIT 3564

#define TX_READY_FIFO_BASE 2048
#define TX_READY_FIFO_SIZE 1024
#define MAX_EV_IN_MEMORY (TX_READY_FIFO_SIZE/8) 

#define DDGLOG(message...) fprintf (logFile, message)

#include "rorc_ddl.h"

#include <smirtl.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

char *config_tokens[] = {
	                  "RORC_CHANNEL",
	                  "DATA_FILE",  
	                  "DATA_PATTERN",
                          "INIT_WORD",
	                  "DATA_LENGTH",   
	                  "RANDOM",     
			  "NORANDOM",
	                  "RESET",      
	                  "NORESET",      
	                  "HEADER",
	                  "NOHEADER",
			  "DDL_COMMANDS",
			  "NO_DDL_COMMANDS",
	                  "PHYSMEM_OFFSET",
	                  "PHYSMEM_LENGTH",
                          "BLOCK_LENGTH",       // header parameters
                          "NO_BLOCK_LENGTH",
                          "MINI_EVENT_ID",
                          "NO_MINI_EVENT_ID",
                          "FORMAT_VERSION",
                          "L1_TRIGGER",
                          "SUB_DETECTORS",
                          "ATTRIBUTES",
                          "STATUS_BITS",
                          "TRIGGER_CLASS_LOW",
                          "TRIGGER_CLASS_HIGH",
                          "ROI_LOW",
                          "ROI_HIGH",
                          "BUNCH_CROSSING_START",
                          "BUNCH_CROSSING_INCREMENT",
                          "BUNCH_CROSSING_SEED",
                          "MAX_EVENT",
	                  ""
	                };

enum config_index {RORC_CHANNEL = 1, 
	           DATA_FILE, DATA_PATTERN, INIT_WORD, DATA_LENGTH,
	           RANDOM, NORANDOM, RESET, NORESET, HEADER, NOHEADER,
	           DDL_COMMANDS, NO_DDL_COMMANDS, 
		   PHYSMEM_OFFSET, PHYSMEM_LENGTH,
                   BLOCK_LENGTH, NO_BLOCK_LENGTH,
                   MINI_EVENT_ID, NO_MINI_EVENT_ID,
                   FORMAT_VERSION, L1_TRIGGER, SUB_DETECTORS, ATTRIBUTES,
                   STATUS_BITS, TRIGGER_CLASS_LOW, TRIGGER_CLASS_HIGH,
                   ROI_LOW, ROI_HIGH,
                   BUNCH_CROSSING_START, BUNCH_CROSSING_INCREMENT, 
                   BUNCH_CROSSING_SEED,
                   MAX_EVENT};

// minimum parameters:
int command_param[] = {-1,
                   2,
                   1,         1,            1,         1,
                   0,      0,        0,     0,       0,      0,      
                   0,     0,
	    	   0,              0,
                   0,            0,
                   0,             0,
                   1,              1,          1,             1,
                   1,           1,                 1,
                   1,       1,
                   1,                    2,
                   1,
                   1 };

FILE *logFile;
FILE *confFile;
FILE *dataFile[MAX_DDG];
char log_fn[MAX_NAME];
char conf_fn[MAX_NAME];
int  scatter;
char smi_object[MAX_NAME];

typedef struct
{
  volatile unsigned long *user;
  unsigned long phys;
  unsigned long length;
  int index;
 } event_addr_t;
event_addr_t event[MAX_DDG][MAX_EV_IN_MEMORY];
event_addr_t tx_fifo[MAX_DDG];

int minor[MAX_DDG], channel[MAX_DDG];
rorc_pci_dev_t prorc[MAX_DDG];
int pattern[MAX_DDG]; //1 if pattern, 0 if file
int data_pattern[MAX_DDG], pattern_no[MAX_DDG];
unsigned long init_word[MAX_DDG];
char data_fn[MAX_DDG][MAX_NAME];
int random_length[MAX_DDG];
int events_in_memory;
int event_serial[MAX_DDG];
int reset[MAX_DDG], channel_ready[MAX_DDG];
int oldest_index[MAX_DDG];

unsigned long max_data_length[MAX_DDG];
unsigned long physmem_offset, physmem_length;

int max_event, kell_rdyrx;

/* header parameters */
int header;
int kell_block_length[MAX_DDG], kell_mini_event_id[MAX_DDG];
unsigned long format_version[MAX_DDG], l1_trigger[MAX_DDG];
unsigned long sub_detectors[MAX_DDG], status_bits[MAX_DDG];
unsigned long attributes[MAX_DDG];
unsigned long trigger_class_high[MAX_DDG];
unsigned long trigger_class_low[MAX_DDG];
unsigned long roi_high[MAX_DDG], roi_low[MAX_DDG];
int b_start, bunch_crossing_start, orbit_start;
int b_incr_min, b_incr_max, min_bc_incr, del_bc_incr;
int bunch_crossing[MAX_DDG], orbit_number[MAX_DDG];
int orig_bunch_crossing_seed, bunch_crossing_seed[MAX_DDG];
float delta_bc_incr;

int startReceived, stopReceived;

/*******************************************************/
void SMI_handle_command() 
{

   char action[64];
   int n_params;

   smi_get_action(action,&n_params);
   if (strcmp(action,"START") == 0) 
   {
     startReceived = 1;
     smi_terminate_action("RUNNING");
   } 
   else if (strcmp(action,"STOP") == 0)
   {
     stopReceived = 1;
     smi_terminate_action("IDLE");
   }
   else
   {
     DDGLOG(" Unrecognised message arrived: %s\n", action); 
   }
}

/*******************************************************/
void usage(char *name)
{
  printf("DDL Data Generator program.\n");
  printf("usage: %s [-{F|f} <conf file>] [-{L|l} <log file>]\n", name);
  printf("           [-{S|s} <SMI object>] [-{T|t} <time-out>] [-{N|n}] [-v]\n");
  printf("   -F <file name>   name of the configuration file. Default: %s.\n",
		  CONF_FN);
  printf("   -L <file name>   name of the log file. Default: %s.\n", LOG_FN);
  printf("   -S <SMI object>  the name of the associated SMI object to connect to\n");
  printf("                    in the form of <domain_name>::<object_name>.\n");
  printf("                    Default: %s\n", SMI_OBJECT); 
  printf("   -T <time-out>    time-out value for DDL commands. Default: %d usec.\n",
                              DDL_RESPONSE_TIME);
  printf("   -N               do not scatter data (use only 1 data buffer).\n");
  printf("   -v               verbose mode (capital V is not accepted).\n");
}

/*******************************************************/
/* Read configuration file                             */
/*******************************************************/
int read_config(char *config_fn)
{
  int i, channels, ddg_index, line_no;
  int n_tok, max_token = MAX_TOKEN;
  int c_index, n_val;
  char *token[MAX_TOKEN];


  /* set default values */
  for (ddg_index = 0; ddg_index < MAX_DDG; ddg_index++)
  {
    minor [ddg_index] = 0;
    channel[ddg_index] = 0;
    pattern[ddg_index] = 0;
    strcpy(data_fn[ddg_index], DATA_FN);
    data_pattern[ddg_index] = 'i'; 
    pattern_no[ddg_index] = 5;
    init_word[ddg_index] = 0;
    random_length[ddg_index] = 1; 
    reset[ddg_index] = 1;  
    max_data_length[ddg_index] = DEFAULT_DATA_LENGTH;
    header = 0;
    kell_rdyrx = 0;
    physmem_offset = 0; 
    physmem_length = MIN_PHYSMEM;

    kell_block_length[ddg_index] = 1;
    kell_mini_event_id[ddg_index] = 1;
    format_version[ddg_index] = 1; 
    l1_trigger[ddg_index] = 0;
    sub_detectors[ddg_index] = 0;
    attributes[ddg_index] = 0;
    l1_trigger[ddg_index] = 0;
    sub_detectors[ddg_index] = 0;
    attributes[ddg_index] = 0;
    status_bits[ddg_index] = 0;
    trigger_class_high[ddg_index] = 0;
    trigger_class_low[ddg_index] = 0;
    roi_high[ddg_index] = 0;
    roi_low[ddg_index] = 0;
    bunch_crossing_start = 1; 
    b_incr_min = 1;
    b_incr_max = 20;
    orbit_start = 0;
    orig_bunch_crossing_seed = 1;
    max_event = 0;
  }
  
  /* open configuration file */
  confFile = fopen(conf_fn, "r");
  if (confFile == NULL)
  {
    DDGLOG("Error opening configuration file: %s. ", conf_fn);
    fflush(logFile);
    perror("");
    exit (EXIT_FAILURE);
  }

  channels = -1;
  line_no = 0;
  debug ("Reading configuration file %s\n", conf_fn);

  while(1)
  {
    n_tok = getToken(token, max_token, confFile);

    if (!n_tok)
    {
      debug ("%s %d: EOF encountered\n", conf_fn, line_no);
      fclose(confFile);
      return (channels + 1);
    }
    line_no++;
//    debug("\n%s %d: ", conf_fn, line_no);
//    for (i = 0; i < n_tok; i++)
//      debug("%s; ", token[i]);

    c_index = n_key_word(config_tokens, token[0]);
//    debug("- %d token(s), command index: %d\n", n_tok, c_index);
    if (!c_index)
    {
      DDGLOG("%s %d: Invalid configuration command: %s\n", 
		      conf_fn, line_no, token[0]);
      exit (EXIT_FAILURE);
    }

    if ((n_tok - 1) < command_param[c_index])
    {
      DDGLOG("%s %d: parameter missing for configuration command '%s'\n",
               conf_fn, line_no, token[0]);
      exit (EXIT_FAILURE);
    }

    /* process configuration commands */

    if ((channels == -1) && ((c_index != RORC_CHANNEL)             &&
                             (c_index != PHYSMEM_OFFSET)           && 
                             (c_index != PHYSMEM_LENGTH)           &&
			     (c_index != DDL_COMMANDS)             &&
			     (c_index != NO_DDL_COMMANDS)          &&
                             (c_index != HEADER)                   &&
                             (c_index != NOHEADER)                 &&
                             (c_index != BUNCH_CROSSING_START)     &&
                             (c_index != BUNCH_CROSSING_INCREMENT) &&
                             (c_index != BUNCH_CROSSING_SEED)      &&
                             (c_index != MAX_EVENT)                )) 
    {
      DDGLOG("Error in %s %d: %s. No channel specified.\n", conf_fn, line_no, 
		                token[0]);
      exit (EXIT_FAILURE);
    }
    switch (c_index)
    {
      case RORC_CHANNEL:
	channels++;
        minor[channels] = atoi(token[1]);
	channel[channels] = atoi(token[2]);
	debug("\n%s %d: minor[%d]: %d, channel[%d]: %d\n", conf_fn, line_no, 
			                          channels, minor[channels], 
			                          channels, channel[channels]);
	for (i = 0; i < channels; i++)
	{
          if ((  minor[i] ==   minor[channels]) && 
              (channel[i] == channel[channels]))
          {
            DDGLOG("Error in %s %d: The same channel (%d, %d) can not be used twice.\n",
                     conf_fn, line_no, minor[channels], channel[channels]);
            exit (EXIT_FAILURE);
          }
	}

        /* open channel */
        if (rorcOpenChannel(&prorc[channels], 
				minor[channels], channel[channels]))
        {
          DDGLOG("Error opening RORC device %d %d.\n", 
		      minor[channels], channel[channels]);
          exit (EXIT_FAILURE);
        }
        debug("Channel %d %d open\n", minor[channels], channel[channels]); 
	if (pRorc(&prorc[channels]))
        {
          DDGLOG("Error: RORC device (%d %d) is a pRorc, can not be used as DDG.\n", 
		      minor[channels], channel[channels]);
          exit (EXIT_FAILURE);
        }
	if (!kell_rdyrx && (prorc[channels].rorc_revision < INTEG))
        {
          DDGLOG("Error: RORC device (%d %d) is not integrated D-RORC.\n" 
          " It can not be used as DDL DDG when NO_DDL_COMMANDS is set.\n",
		      minor[channels], channel[channels]);
          exit (EXIT_FAILURE);
        }
        break;
      case DATA_FILE:
	strncpy(data_fn[channels], token[1], MAX_NAME);
	pattern[channels] = 0;
	debug("%s %d: data file[%d]: %s\n", conf_fn, line_no, 
			                    channels, data_fn[channels]);
	data_fn[channels][MAX_NAME] = '\0';
	for (i = 0; i < channels; i++)
	{
          if (strcmp(data_fn[i], data_fn[channels]) == 0) 
          {
            DDGLOG("Error in %s %d: The same file (%s) can not be used twice.\n", 
                        conf_fn, line_no, data_fn[channels]);
            exit (EXIT_FAILURE);
          }
	}
        /* open data file */
        dataFile[channels] = fopen(data_fn[channels], "r");
        if (dataFile[channels] == NULL)
        {
          DDGLOG("Error in %s %d: Can not open file %s.\n", 
                        conf_fn, line_no, data_fn[channels]);
          fflush(logFile);
          perror("");
          exit (EXIT_FAILURE);
        }
        n_val = getVal((unsigned long *)&max_data_length[channels], 1, 
                                         dataFile[channels], "%ld");
        if (n_val != 1)
        {
          DDGLOG("Error in %s %d: Empty data file %s.\n",
                   conf_fn, line_no, data_fn[channels]);
          exit (EXIT_FAILURE);
        }
        if ((max_data_length[channels] < 1) || 
            (max_data_length[channels] > DDL_MAX_TX_WORD))
        {
          DDGLOG("Error in %s %d: Invalid max data length (%ld) in file %s.\n"
                     "Data length should be positive and less than %d\n",
                      conf_fn, line_no, max_data_length[channels], 
                      data_fn[channels], DDL_MAX_TX_WORD);
          exit (EXIT_FAILURE);
        }
	debug("%s %d: data file %s open. Max data length: %ld words (%4.2f kB).\n", 
                conf_fn, line_no, data_fn[channels], max_data_length[channels],
                   (double)max_data_length[channels]/256);
        break;
      case DATA_PATTERN: 
	data_pattern[channels] = tolower(*token[1]);
	pattern[channels] = 1;
	debug("%s %d: pattern[%d]: %c\n", conf_fn, line_no, 
			                  channels, data_pattern[channels]);
        switch (data_pattern[channels])
        {
          case 'c': pattern_no[channels] = 1; break;
          case 'a': pattern_no[channels] = 2; break;
          case '0': pattern_no[channels] = 3; init_word[channels] = -2; break;
          case '1': pattern_no[channels] = 4; init_word[channels] = 1 ; break;
          case 'i': pattern_no[channels] = 5; break;
          case 'd': pattern_no[channels] = 6; break;
          default : 
            DDGLOG("Error in %s %d: pattern (%c) must be 'c', 'a', '0', '1', 'i',  or 'd'\n",
			   conf_fn, line_no, data_pattern[channels]); 
	    exit (EXIT_FAILURE);
        }
        break;
      case INIT_WORD:
        init_word[channels] = strtoul(token[1], (char**)NULL, 16);
        debug("%s %d: init word[%d]: 0x%08lx\n", conf_fn, line_no, channels, 
                     init_word[channels]);
        break;
      case DATA_LENGTH:
        max_data_length[channels] = atoi(token[1]);
	debug("%s %d: max data length[%d]: %ld words (%4.2f kB)\n", 
                   conf_fn, line_no, channels, max_data_length[channels],
                   (double)max_data_length[channels]/256);
	if ((max_data_length[channels] < 1) ||
	   ((max_data_length[channels] + (header ? DDL_HEADER_SIZE_WORD : 0))
                                       > DDL_MAX_TX_WORD))
	{
          DDGLOG("Error in %s %d: data length (%ld) should be positive and less than %d\n", 
                conf_fn, line_no, max_data_length[channels], 
                (DDL_MAX_TX_WORD - (header ? DDL_HEADER_SIZE_WORD : 0)));
	  exit (EXIT_FAILURE);
	}
        break;
      case RANDOM:
	random_length[channels] = 1;
	debug("%s %d: channel %d: random length\n", 
                                            conf_fn, line_no, channels);
        break;
      case NORANDOM:
	random_length[channels] = 0;
	debug("%s %d: channel %d: no random length\n",
                                            conf_fn, line_no, channels);
        break;
      case RESET:
	reset[channels] = 1;
	debug("%s %d: channel %d: reset\n", conf_fn, line_no, channels);
        break;
      case NORESET:
	reset[channels] = 0;
	debug("%s %d: channel %d: no reset\n", conf_fn, line_no, channels);
        break;
      case HEADER:
	header = 1;
	debug("%s %d: header generation is switched on\n", conf_fn, line_no);
        break;
      case NOHEADER:
	header = 0;
	debug("%s %d: header generation is switched off\n", conf_fn, line_no);
        break;
      case DDL_COMMANDS:
	kell_rdyrx = 1;
	debug("%s %d: 'DDL command' mode is switched on\n", conf_fn, line_no);
        break;
      case NO_DDL_COMMANDS:
	kell_rdyrx = 0;
	debug("%s %d: 'DDL command' mode is switched off\n", conf_fn, line_no);
        break;
      case PHYSMEM_OFFSET:
        physmem_offset = atoi(token[1])*1024*1024;
	debug("%s %d: physmem offset: %ld MB\n", conf_fn, line_no, 
			                         physmem_offset/1024/1024);
	break;
      case PHYSMEM_LENGTH:
        physmem_length = atoi(token[1])*1024*1024;
	debug("%s %d: physmem size: %ld MB\n", conf_fn, line_no, 
			                       physmem_length/1024/1024);
        if (physmem_length < MIN_PHYSMEM)
        {
          DDGLOG("Error: utilizable physmem size (defined in %s as %ld MB) "
			  "should be at least %d MB.\n",
               config_fn, physmem_length/1024/1024,  MIN_PHYSMEM/1024/1024);
          exit (EXIT_FAILURE);
        }
	break;
      case BLOCK_LENGTH:
	kell_block_length[channels] = 1;
	debug("%s %d: channel %d: block length generation is switched on\n",
                                              conf_fn, line_no, channels);
	break;
      case NO_BLOCK_LENGTH:
	kell_block_length[channels] = 0;
	debug("%s %d: channel %d: block length generation is switched off\n",
                                              conf_fn, line_no, channels);
	break;
      case MINI_EVENT_ID:
	kell_mini_event_id[channels] = 1;
	debug("%s %d: channel %d: correct mini event id generation is switched on\n",
                                              conf_fn, line_no, channels);
	break;
      case NO_MINI_EVENT_ID:
	kell_mini_event_id[channels] = 0;
	debug("%s %d: channel %d: random mini event id generation is switched on\n",
                                              conf_fn, line_no, channels);
	break;
      case FORMAT_VERSION:
        format_version[channels] = strtoul(token[1], (char**)NULL, 16) & 0xFF;
        debug("%s %d: header format version[%d]: 0x%02lx\n", 
                     conf_fn, line_no, channels, format_version[channels]);
	break;
      case L1_TRIGGER:
        l1_trigger[channels] = strtoul(token[1], (char**)NULL, 16) & 0x3FF;
        debug("%s %d: L1 trigger message[%d]: 0x%03lx\n", 
                     conf_fn, line_no, channels, l1_trigger[channels]);
	break;
      case SUB_DETECTORS:
        sub_detectors[channels] = strtoul(token[1], (char**)NULL, 16) & 0xFFFFFF;
        debug("%s %d: participating sub detectors[%d]: 0x%06lx\n", 
                     conf_fn, line_no, channels, sub_detectors[channels]);
	break;
      case ATTRIBUTES:
        attributes[channels] = strtoul(token[1], (char**)NULL, 16) & 0xFF;
        debug("%s %d: block attributtes[%d]: 0x%02lx\n", 
                     conf_fn, line_no, channels, attributes[channels]);
	break;
      case STATUS_BITS:
        status_bits[channels] = strtoul(token[1], (char**)NULL, 16) & 0xFFFF;
        debug("%s %d: status and error bits[%d]: 0x%04lx\n", 
                     conf_fn, line_no, channels, status_bits[channels]);
	break;
      case TRIGGER_CLASS_LOW:
        trigger_class_low[channels] = strtoul(token[1], (char**)NULL, 16);
        debug("%s %d: trigger class low word[%d]: 0x%08lx\n", 
                     conf_fn, line_no, channels, trigger_class_low[channels]);
	break;
      case TRIGGER_CLASS_HIGH:
        trigger_class_high[channels] = strtoul(token[1], (char**)NULL, 16) & 
                                                                      0x3FFFF;
        debug("%s %d: trigger class hight part[%d]: 0x%05lx\n", 
                    conf_fn, line_no, channels, trigger_class_high[channels]);
	break;
      case ROI_LOW:
        roi_low[channels] = strtoul(token[1], (char**)NULL, 16) & 0xF;
        debug("%s %d: ROI low part[%d]: 0x%01lx\n",
                    conf_fn, line_no, channels, roi_low[channels]);
	break;
      case ROI_HIGH:
        roi_high[channels] = strtoul(token[1], (char**)NULL, 16);
        debug("%s %d: ROI high word[%d]: 0x%08lx\n",
                    conf_fn, line_no, channels, roi_high[channels]);
	break;
      case BUNCH_CROSSING_START:
        b_start = atoi(token[1]);
        orbit_start = b_start / BUNCH_PER_ORBIT;
        bunch_crossing_start = b_start % BUNCH_PER_ORBIT;
        debug("%s %d: orbit and bunch crossing start values: %d, %d\n",
                          conf_fn, line_no, orbit_start, bunch_crossing_start);
	break;
      case BUNCH_CROSSING_INCREMENT:
        b_incr_min = atoi(token[1]);
        b_incr_max = atoi(token[2]);
        debug("%s %d: bunch crossing increment min. and max. values: %d, %d\n",
                              conf_fn, line_no, b_incr_min, b_incr_max);
        if ((b_incr_min < 0) || (b_incr_min > b_incr_max) || (b_incr_max > 31))
        {
          DDGLOG("Error: bunch crossing increment min. (%d) or max. (%d)"
                                              " values are not correct.\n"
                 "(Sould be: min >= 0, min <= max, max < 32)\n", 
                                                    b_incr_min, b_incr_max);
          exit (EXIT_FAILURE);
        }
        break;
      case BUNCH_CROSSING_SEED:
        orig_bunch_crossing_seed = atoi(token[1]);
        debug("%s %d: bunch crossing seed value: %d\n",
                          conf_fn, line_no, orig_bunch_crossing_seed);
	break;
      case MAX_EVENT:
        max_event = atoi(token[1]);
        debug("%s %d: max event value: %d\n", conf_fn, line_no, max_event);
	break;
      default:
        DDGLOG("%s %d: Invalid command: %s\n", conf_fn, line_no, token[0]);
        exit (EXIT_FAILURE);
    }
  }
}

/*******************************************************/
unsigned long selectEventLength(unsigned long max_length)
{
  static int eventLengthSeed = 1;

  srandom(eventLengthSeed);
  eventLengthSeed = random();

  return ((unsigned long)((float)eventLengthSeed * (float) max_length / 
                                                           (RAND_MAX + 1.0)));
}

/*******************************************************/
int selectMiniEventId()
{
  static int miniEventSeed = 1;

  srandom(miniEventSeed);
  miniEventSeed = random();

  return ((int)((float)miniEventSeed * (float)0xFFF / (RAND_MAX + 1.0)));
}

/*******************************************************/
void selectBunchCrossing(int *bunchCrossingSeed, 
                         int *bunchCrossing,
                         int *orbitNumber)
{
  int bunchCrossingIncrement;

  if (b_incr_min == b_incr_max)
    bunchCrossingIncrement = min_bc_incr;
  else
  {
    srandom(*bunchCrossingSeed);
    *bunchCrossingSeed = random();
    bunchCrossingIncrement = min_bc_incr + (int)((float)*bunchCrossingSeed * 
                                                               delta_bc_incr);
  }
  *orbitNumber += bunchCrossingIncrement / BUNCH_PER_ORBIT;
  *bunchCrossing += bunchCrossingIncrement % BUNCH_PER_ORBIT;
  if (*bunchCrossing >= BUNCH_PER_ORBIT)
  {
    (*orbitNumber)++;
    *bunchCrossing %= BUNCH_PER_ORBIT;
  }
}

/*******************************************************/
void fill_header(volatile unsigned long *buffer_addr, 
		 int i_channel, int block_length) 
{
  int i, mini_event_id;

  for (i = 0; i < DDL_HEADER_SIZE_WORD; i++)
    *(buffer_addr + i) = 0x0;

  if (!kell_block_length[i_channel])
    *buffer_addr = DDL_NO_BLOCK_LENGTH;
  else
    *buffer_addr = block_length * 4;
  *(buffer_addr + 1) = (format_version[i_channel] << 24)          |
                       ((l1_trigger[i_channel] & 0x3FF) << 14)    |
                       (bunch_crossing[i_channel] & 0xFFF);
  *(buffer_addr + 2) = orbit_number[i_channel] & 0xFFFFFF;
//   debug(" ch: %d, BC: %d, 0x%08x; ON: %d, 0x%08x\n", i_channel, 
//   bunch_crossing[i_channel], bunch_crossing[i_channel],
//   orbit_number[i_channel], orbit_number[i_channel]);
  *(buffer_addr + 3) = (attributes[i_channel] << 24)              |
                       (sub_detectors[i_channel] & 0xFFFFFF);
  if (kell_mini_event_id[i_channel])
    mini_event_id = bunch_crossing[i_channel];
  else
    mini_event_id = selectMiniEventId();
  *(buffer_addr + 4) = ((status_bits[i_channel] & 0xFFFF) << 12)  |
                       (mini_event_id & 0xFFF);
  *(buffer_addr + 5) = trigger_class_low[i_channel];
  *(buffer_addr + 6) = (roi_low[i_channel] << 28)                 |
                       (trigger_class_high[i_channel] & 0x3FFFF);
  *(buffer_addr + 7) = roi_high[i_channel];
			  
}

/*******************************************************/
void read_data_file(int i_channel, int i, int serial)
{
  int i1, n_val;

  n_val = getVal(&event[i_channel][i].length, 1, dataFile[i_channel], "%lx");
  if (n_val != 1)
  {
    /* end of file */
    if (serial == 1)
    {
      DDGLOG("Error: No data for channel %d in file %s.\n", i_channel, 
                                                    data_fn[i_channel]);
      exit (EXIT_FAILURE);
    }
    rewind(dataFile[i_channel]);
    /* skip max_length */
    n_val = getVal((unsigned long *)&max_data_length[i_channel], 1,
                                            dataFile[i_channel], "%ld");

    n_val = getVal(&event[i_channel][i].length, 1, dataFile[i_channel], "%lx");
  }
//  debug(" Length of the %dth event read in channel %d: 0x%lx.\n",
//		                     i, i_channel, event[i_channel].length);

  i1 = 0;
  if (header)
    i1 += DDL_HEADER_SIZE_WORD;
  n_val = getVal((unsigned long *)event[i_channel][i].user + i1,
                   event[i_channel][i].length, dataFile[i_channel], "%lx");
  if (header)
  {
    if (serial == 1)
    {
      bunch_crossing_seed[i_channel] = orig_bunch_crossing_seed;
      bunch_crossing[i_channel] = bunch_crossing_start;
      orbit_number[i_channel] = orbit_start;
    }
    else
    {
      selectBunchCrossing(&bunch_crossing_seed[i_channel], 
                          &bunch_crossing[i_channel], &orbit_number[i_channel]);
    }
    fill_header(event[i_channel][i].user, i_channel, n_val + i1);
  }
  event[i_channel][i].length += i1;
}

/*******************************************************/
int main(int argc, char **argv) 
{
  int c, j, i, i1, i2, i_channel, n_channels;
  int exit_code = EXIT_SUCCESS;
  int timeout, ustimeout;
  int ret, n_status, ctstw_code;
  int channel_busy;

  stword_t stw;
  time_t time_val, time_now;
  struct timeval start_tv, now_tv;
  int dsec, dusec;
  double dtime;

  int physmem_fd;
  unsigned long req_mem, tx_fifo_mem;
  unsigned long phys_addr, phys_size_physmem;
  volatile unsigned long *addr_user;
  unsigned long data, event_size;
  volatile unsigned long *report_addr;

  ustimeout = DDL_RESPONSE_TIME;
  strcpy(log_fn, LOG_FN);
  strcpy(conf_fn, CONF_FN);
  strcpy(smi_object, SMI_OBJECT);
  scatter = 0;

  /* Check parameters */
  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "F:f:L:l:S:s:T:t:Nnvh")) != EOF)
    {
      switch(toupper(c))
      {
        case 'F': strcpy(conf_fn, optarg); break;
        case 'L': strcpy(log_fn, optarg); break;
        case 'S': strcpy(smi_object, optarg); break;
	case 'T': ustimeout = atoi(optarg); break;
        case 'N': scatter = 1; break;
        case 'V': break;
        default : usage (argv[0]); return -1; break;
      }
    }
  }

  /* open log file */
  if (strcmp(log_fn, LOG_FN))
  {
    logFile = fopen(log_fn, "a");
    if (logFile == NULL)
    {
      printf("Error opening log file: %s. ", log_fn);
      fflush(stdout);
      perror("");
      exit (EXIT_FAILURE);
    }
  }
  else
    logFile = stdout;
  
  n_channels = read_config(conf_fn);
  if (n_channels < 1) 
  {
    n_channels = 1;
    /* open default channel */
    if (rorcOpenChannel(&prorc[0], minor[0], channel[0]))
    {
      DDGLOG("Error opening RORC device %d %d.\n", minor[0], channel[0]);
      exit (EXIT_FAILURE);
    }
    debug("Channel %d %d open\n", minor[0], channel[0]); 
    if (pRorc(&prorc[0]))
    {
      DDGLOG("Error: RORC device (%d %d) is a pRORC, can not be used as DDG.\n", 
                                                      minor[0], channel[0]);
      exit (EXIT_FAILURE);
    }
    if (!kell_rdyrx && (prorc[0].rorc_revision < INTEG))
    {
      DDGLOG("Error: RORC device (%d %d) is not an integrated D-RORC.\n" 
      " It can not be used as DDL DDG when NO_DDL_COMMANDS is set.\n",
                                                      minor[0], channel[0]);
      exit (EXIT_FAILURE);
    }
  }
  timeout = ustimeout * prorc[0].loop_per_usec;
  if (timeout > DDL_MAX_WAIT_CYCLE )
  {
    DDGLOG("Error: time-out value %d us is too large.\n", ustimeout);
    DDGLOG("Please choose a value less than or equal to %d us.\n",
                                                       prorc[0].max_resp_time);
    exit (EXIT_FAILURE);
  }
  debug(" Timout values used: %d us (%d cycles),\n", ustimeout, timeout);
  min_bc_incr = (1 << b_incr_min) - 1;                // 2^min - 1
  del_bc_incr = (1 << (b_incr_max - b_incr_min)) - 1; // 2^(max-min) - 1
  del_bc_incr <<= b_incr_min;                         // * 2^min = 2^max-2^min
  delta_bc_incr = (float)del_bc_incr / (RAND_MAX + 1.0);

  /* altanative way:
  max_bc_incr = (1 << b_incr_max) - 1;                // 2^max - 1
  delta_bc_incr = (float)(max_bc_incr - min_bc_incr) / (RAND_MAX + 1.0);
  */
  debug(" Minimum and maximum Bunch Crossing increment values: 0x%08x, 0x%08x\n",
                                min_bc_incr, min_bc_incr + del_bc_incr);

  /* physmem */

  if (physmemOpen (&physmem_fd, &addr_user, &phys_addr,
                                              &phys_size_physmem))
  {
    DDGLOG("Error opening physmem device.\n");
    exit (EXIT_FAILURE);
  }

  debug(" Allocation of the physmem memory.");
  debug(" Physical address: 0x%08lx.\n", phys_addr);
  debug(" Size of the physmem: %ld (0x%08lx) B, %ld MB, %ld (0x%08lx) W.\n",
             phys_size_physmem, phys_size_physmem, phys_size_physmem/1024/1024,
             phys_size_physmem >> 2, phys_size_physmem >> 2);

  if ((physmem_offset + physmem_length) > phys_size_physmem)
  {
    DDGLOG("Error: utilizable memory offset (%ld MB) + size (%ld MB) must be "
           "less than \n       the physmem area physiscal size (%ld MB).\n",
    physmem_offset/1024/1024, physmem_length/1024/1024,
    phys_size_physmem/1024/1024);
    physmemClose (physmem_fd, (unsigned long *)addr_user, phys_size_physmem);
    exit (EXIT_FAILURE);
  }

  phys_addr += physmem_offset;
  addr_user += physmem_offset/4;
   
  debug(" Start address of utilizable memory: 0x%08lx (%p), offset: 0x%08lx.\n",
                      phys_addr, addr_user, physmem_offset);
  debug(" Size of the utilizable memory: %ld (0x%08lx) B, %ld MB, %ld (0x%08lx) W.\n",
      physmem_length, physmem_length, physmem_length/1024/1024,
      physmem_length >> 2, physmem_length >> 2);

  /* calculate required memory size */
  req_mem = 0;
  for (i_channel = 0; i_channel <  n_channels; i_channel++)
  {
    req_mem += 4 * max_data_length[i_channel];
    tx_fifo[i_channel].phys = phys_addr + i_channel * TX_READY_FIFO_BASE;
    tx_fifo[i_channel].user = addr_user + i_channel * TX_READY_FIFO_BASE/4; 
    for (j = 0; j < TX_READY_FIFO_SIZE/4; j++)
      *(tx_fifo[i_channel].user + j) = tx_fifo[i_channel].phys + 4*j; //!!!!!!!
  }
  if (header)
    req_mem += n_channels * DDL_HEADER_SIZE_BYTE;
  tx_fifo_mem = n_channels * TX_READY_FIFO_BASE;
  req_mem += tx_fifo_mem;
  debug(" Required minimum memory for %d channels: %ld bytes (%4.2f MB).\n", 
          n_channels, req_mem, (double)req_mem/1024/1024);
  debug(" For TxFIFO: %ld bytes (%4.2f kB).\n",
           tx_fifo_mem, (double)tx_fifo_mem/1024);
  if (req_mem > physmem_length)
  {
    DDGLOG("Error: physmem memory size (%ld MB) is not enough fort %d channels.\n",
            physmem_length/1024/1024, n_channels);
    DDGLOG(" The required memory is %4.2f MB. Use larger physmem or less channel.\n",
            (double)req_mem/1024/1024);
    physmemClose (physmem_fd, (unsigned long *)addr_user, phys_size_physmem);
    exit (EXIT_FAILURE);
  }

  /* calculates number of events stored in memory for each channels */
  if (scatter)
    events_in_memory = 1;
  else
  {
    events_in_memory = (physmem_length - n_channels * TX_READY_FIFO_BASE) /
                       (req_mem - n_channels * TX_READY_FIFO_BASE);
    if (events_in_memory > MAX_EV_IN_MEMORY)
      events_in_memory = MAX_EV_IN_MEMORY;
  }

  debug(" Number of events in memory for each channels: %d.\n",  
                                                     events_in_memory);

  event[0][0].phys = phys_addr + tx_fifo_mem;
  event[0][0].user = addr_user + tx_fifo_mem/4;
  for (i_channel = 0; i_channel <  n_channels; i_channel++)
  {
    if (i_channel > 0)
    {
      event[i_channel][0].phys = event[i_channel-1][0].phys + 
             events_in_memory * (4 * max_data_length[i_channel-1] + 
                                (header ? DDL_HEADER_SIZE_BYTE : 0));
      event[i_channel][0].user = event[i_channel-1][0].user + 
             events_in_memory * (max_data_length[i_channel-1] +
                                (header ? DDL_HEADER_SIZE_WORD : 0));
    }
    for (i = 1; i < events_in_memory; i++)
    {
      event[i_channel][i].phys = event[i_channel][i-1].phys + 
                                 4 * max_data_length[i_channel] + 
                                 (header ? DDL_HEADER_SIZE_BYTE : 0);
      event[i_channel][i].user = event[i_channel][i-1].user + 
                                     max_data_length[i_channel] +
                                 (header ? DDL_HEADER_SIZE_WORD : 0);
    }
    event_serial[i_channel] = 0; // event serial #, first ev = 1
    if (pattern[i_channel])
    {
      /* fill buffer (only for 1 event, since events do not change*/
      data = init_word[i_channel];
      i1 = 0;
      i2 = max_data_length[i_channel];
      if (header)
      {
        i1 += DDL_HEADER_SIZE_WORD;
        i2 += DDL_HEADER_SIZE_WORD;
        bunch_crossing_seed[i_channel] = orig_bunch_crossing_seed;
        bunch_crossing[i_channel] = bunch_crossing_start;
        orbit_number[i_channel] = orbit_start;
	fill_header(event[i_channel][0].user, i_channel, i2);
      }
      *(event[i_channel][0].user + i1) = event_serial[i_channel];
      i1++;
      for (i = i1; i < i2; i++)
      {
        *(event[i_channel][0].user + i) = data;
        switch (data_pattern[i_channel])
        {
          case 'c': break;
          case 'a': data = ~data; break;
          case '0':
            data = (data <<1) | 1;
            if (data == -1)
            data = init_word[i_channel];
            break;
          case '1':
            data <<= 1;
            if (data == 0)
            data = init_word[i_channel];
            break;
          case 'i': data++; break;
          case 'd': data--; break;
          default:  data++; break;
        }
      }
      event[i_channel][0].length = i2;
      /* copy the same event into the next (events_in_memory - 1) buffers */
      for (i = 1; i < events_in_memory; i++)
      {
        event[i_channel][i].length = event[i_channel][0].length;
        memcpy((void *)event[i_channel][i].user,
               (void *)event[i_channel][0].user, 4 * event[i_channel][i].length);
      }
    }
    else
    {
      /* fill buffers for events_in_memory events in each channel */
      for (i = 0; i < events_in_memory; i++)
      {
        read_data_file(i_channel, i, i+1);
      }
    }

    debug(" Start address for events of channel %d: 0x%08lx (offset: 0x%08lx).\n",
           i_channel, event[i_channel][0].phys,
                      event[i_channel][0].phys - phys_addr + physmem_offset);
    debug(" The corresponding TX FIFO address: 0x%08lx.\n\n", 
		        tx_fifo[i_channel].phys);
  }
  if (max_event)
    debug(" Maximal number of event to send: %d.\n", max_event);    

  /* init SMI */  
  smi_attach(smi_object, SMI_handle_command);
  smi_volatile();
  smi_set_state("IDLE");
  startReceived = 0;
  stopReceived = 0;

  for (i_channel = 0; i_channel <  n_channels; i_channel++)
  {
    channel_ready[i_channel] = 0;
    /* reset channels */
    if (reset[i_channel])
    {
      rorcReset(&prorc[i_channel], RORC_RESET_RORC);
      debug(" Channel %d reset\n", i_channel);
    }

    usleep(10);

    /* enable data transfer */
    dRorcWriteReg(&prorc[i_channel], C_TRBAR, tx_fifo[i_channel].phys);
    if (!(dRorcReadReg(&prorc[i_channel], C_CSR) & DRORC_CMD_DATA_TX_ON_OFF))
      dRorcWriteReg(&prorc[i_channel], C_CSR, DRORC_CMD_DATA_TX_ON_OFF);
  }
  
  /* wait for RDYRX commands and START message */  
  time_now = time(&time_val);
  gettimeofday(&start_tv, NULL);
  DDGLOG("\nWaiting for START message at %s\n", ctime(&time_val));

  n_status = 0;
  while(1)
  {
    sleep(1);
    if (kell_rdyrx)
    {
      for (i_channel = 0; i_channel <  n_channels; i_channel++)
      {
        if (ddlCheckStatus(&prorc[i_channel]))
        {
          stw = ddlReadStatus(&prorc[i_channel]);
          debug(" Status word arrived in channel %d: 0x%08lx\n", 
			                           i_channel, stw.stw);
	  if ((stw.part.code == RDYRX) && !channel_ready[i_channel])
	  {
	    n_status++;
	    channel_ready[i_channel] = 1;
	    ctstw_code = CTSTW;
	  }
	  else
	  {
	    ctstw_code = ILCMD;
	  }
	  ret = ddlSendCommand(&prorc[i_channel], stw.part.dest, ctstw_code, 
	                         stw.part.trid, stw.part.param, timeout);
	}
      }
    }
    if(startReceived) 
    {
      debug(" START message received\n");
      startReceived = 0;
      n_status++;
    }
    if(stopReceived) 
    {
      stopReceived = 0;
      DDGLOG(" STOP received, ddg program terminates\n");
      goto vege;
    }
    if (n_status >= kell_rdyrx * n_channels + 1)
    {
      gettimeofday(&start_tv, NULL);
      DDGLOG(" ddg program STARTED, sending data\n");
      break;
    }
  }

  /* push TX FIFOs */
  for (i = 0; i < events_in_memory; i++)
  {
    for (i_channel = 0; i_channel < n_channels ; i_channel++) 
    { 
      if (max_event)
      {
        if (i >= max_event)
        {
          DDGLOG(" Maximum event serial (%d) reached, ddg program terminates\n",
                                            max_event);
          exit(exit_code);
        }
      }

      if (i == 0)
        oldest_index[i_channel] = 0;

      *(tx_fifo[i_channel].user + 2*i) = -1;
      *(tx_fifo[i_channel].user + 2*i+1) = -1;
      event[i_channel][i].index = i;
      event_serial[i_channel]++;

      if (pattern[i_channel])
      {
        *(event[i_channel][i].user + (header ? DDL_HEADER_SIZE_WORD : 0)) = 
		                                    event_serial[i_channel];
        if (header)
        {
          if (i > 0)
            selectBunchCrossing(&bunch_crossing_seed[i_channel], 
                         &bunch_crossing[i_channel], &orbit_number[i_channel]);
          fill_header(event[i_channel][i].user, i_channel, 
                                                  event[i_channel][i].length);
        }
      }
			
      if (random_length[i_channel])
        event_size = selectEventLength(event[i_channel][i].length - 
                                   (header ? DDL_HEADER_SIZE_WORD : 0)) +
                                   (header ? DDL_HEADER_SIZE_WORD : 0);
      else
        event_size = event[i_channel][i].length;
      if (header && kell_block_length[i_channel])
        *event[i_channel][i].user = event_size * 4;
      dRorcPushTxFreeFifo(&prorc[i_channel], event[i_channel][i].phys,
		                             event_size,
		                             event[i_channel][i].index);
      debug(" Data block %d of length %8ld pushed in channel %d\n", i,
                                                      event_size, i_channel);
    }
  }

  /* data cycle */
  while(1)
  {
    if(stopReceived) 
    {
      stopReceived = 0;
      DDGLOG(" STOP received, ddg program terminates\n");
      goto vege;
    }

    if(max_event)
    {
      if (event_serial[0] >= max_event)
      {
        DDGLOG(" Maximum event serial (%d) reached, ddg program terminates\n",
                                            max_event);
        exit(exit_code);
      }
    }

//    usleep(10);

    for (i_channel = 0; i_channel <  n_channels; i_channel++)
    {

      if (kell_rdyrx)
      {
        /* check status */
        if (ddlCheckStatus(&prorc[i_channel]))
        {
          stw = ddlReadStatus(&prorc[i_channel]);
          debug(" Status word arrived in channel %d: 0x%08lx\n", 
			                           i_channel, stw.stw);
          ret = ddlSendCommand(&prorc[i_channel], stw.part.dest, ILCMD, 
	                         stw.part.trid, stw.part.param, timeout);
        }
        if (rorcCheckLink(&prorc[i_channel]) != RORC_STATUS_OK)
        {
          DDGLOG(" Error: link is down at channel %d. Program stops.\n",
                                                              i_channel);
          goto vege;
        }
      }

      /* loop for a DMA ending */
      channel_busy = 0;
      j = 0;
      report_addr = tx_fifo[i_channel].user + 2 * oldest_index[i_channel];
      while (1)
      {
	if (*report_addr == -1)
	{
//	  if (dRorcReadReg(&prorc[i_channel], C_TXDC) & 0xFFFFFF)
//	  {
            channel_busy = 1;
            break;
//	  }
//          j++;
//          if (j > timeout)
//            break;
	}
	else
          break;
      }
      if (channel_busy)
      {
//        debug(" Channel %d busy\n", i_channel);
        continue; // go to next channel
      }
//      if (*report_addr == -1)
//      {
//        DDGLOG(" In channel %d Tx report address # %d is -1 thought Tx DMA count is 0.\n",
//                                    i_channel, oldest_index[i_channel]);
//        debug (" C_TXDC for channel %d: 0x%08lx\n", i_channel,
//                                    dRorcReadReg(&prorc[i_channel], C_TXDC));  
//	exit_code = EXIT_FAILURE;
//	goto vege;
//      }
//      debug(" TX FIFO index %d is freed in channel %d.\n",
//                                        oldest_index[i_channel], i_channel); 

      /* send data */
      i = oldest_index[i_channel];
      *(tx_fifo[i_channel].user + 2*i) = -1;
      *(tx_fifo[i_channel].user + 2*i+1) = -1;
      event[i_channel][i].index = i;
      event_serial[i_channel]++;
      if (pattern[i_channel])
      {
        *(event[i_channel][i].user + (header ? DDL_HEADER_SIZE_WORD : 0)) = 
		                                    event_serial[i_channel];
        if (header)
        {
          selectBunchCrossing(&bunch_crossing_seed[i_channel], 
                         &bunch_crossing[i_channel], &orbit_number[i_channel]);
          fill_header(event[i_channel][i].user, i_channel, 
                                                  event[i_channel][i].length);
        }
      }
      else
      {
        /* read next event */
        read_data_file(i_channel, i, event_serial[i_channel]);
      }
			
      if (random_length[i_channel])
        event_size = selectEventLength(event[i_channel][i].length - 
                                   (header ? DDL_HEADER_SIZE_WORD : 0)) +
                                   (header ? DDL_HEADER_SIZE_WORD : 0);
      else
        event_size = event[i_channel][i].length;
      if (header && kell_block_length[i_channel])
        *event[i_channel][i].user = event_size * 4;

      dRorcPushTxFreeFifo(&prorc[i_channel], event[i_channel][i].phys,
		                             event_size,
		                             event[i_channel][i].index);

      debug(" Data block %d of length %8ld pushed in channel %d\n", 
                          event_serial[i_channel], event_size, i_channel);
      oldest_index[i_channel]++;
//      if (oldest_index[i_channel] >= (TX_READY_FIFO_SIZE/8))
      if (oldest_index[i_channel] >= events_in_memory)
        oldest_index[i_channel] = 0;
    }

  }

vege:
  /* stop data sending */
//  for (i_channel = 0; i_channel <  n_channels; i_channel++)
//  {
//    if ((dRorcReadReg(&prorc[i_channel], C_CSR) & DRORC_CMD_DATA_TX_ON_OFF))
//      dRorcWriteReg(&prorc[i_channel], C_CSR, DRORC_CMD_DATA_TX_ON_OFF);
//  }

  /* calculate and print the elapsed time */
  gettimeofday(&now_tv, NULL);
  DDGLOG("\n-------------------------------------------\n");
  elapsed (&now_tv, &start_tv, &dsec, &dusec);
  dtime = (double)dsec + (double)dusec/MEGA;
  DDGLOG("The program pushed the following number of events in %d.%06d seconds:\n",
                                                   dsec, dusec);
  for (i_channel =0; i_channel < n_channels; i_channel++)
    DDGLOG("  Channel %d: %d events\n", i_channel, event_serial[i_channel]);

  time_now = time(&time_val);
  DDGLOG("The program ends at %s\n", ctime(&time_val));

  exit (exit_code);
}
