/***********************************************************
 * diu_id.c       
 *
 *  Read and write diu hw id           
 *
 * last updated: 13/08/2004
 * written by: Peter Csato and Ervin Denes
 ***********************************************************/
#define SOURCE_FILE "diu_id.c"
#define DESCRIPTION "read and write DIU HW ID"

#include <ctype.h>
#include "rorc_ddl.h"

#define MAX_OPTION 5
#define PLD_LEN 7

/***********************************************************/
void usage (char* argv)
{
  printf("usage: %s [-{M|m} <minor>] [-{C|c} <channel>]\n", argv);
  printf("              [-V <major version> -v <minor version>\n");
  printf("               -P <PLD version> -B <speed version> -S <serial#>]\n");
  printf(" <minor>    RORC's card number. Default = 0.\n");
  printf(" <channel>  RORC's channel number. Default = 0.\n");
  printf("  If no HW version parameters are given: only read DIU id\n" );
}

/***********************************************************/
int main( int argc, char *argv[])
{
  __u32 param;
  __u8 data[DDL_MAX_HW_ID];
  __u8 c_pld[PLD_LEN+1];
  int write_id, length, version_major, version_minor, serial;
  int speed;
  int i, n, c, ret, minor, channel, time, ustime = DDL_RESPONSE_TIME;
  rorc_pci_dev_t prorc;
  stword_t stw;
  rorcHwSerial_t hw;

  /* Default values */

  minor = 0;
  channel = 0;
  write_id = 0;

  /* Check parameters */

  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:V:v:P:p:B:b:S:s:")) != EOF)
    {
      switch(c)
      {
        case 'm':
        case 'M': minor = atoi(optarg); if (minor < 0) minor = 0; break;
        case 'c':
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'v': version_minor = atoi(optarg); write_id++; break; 
        case 'V': version_major = atoi(optarg); write_id++; break;
        case 'p':
        case 'P':
          if (strlen(optarg) > PLD_LEN)
            optarg[PLD_LEN] = '\0';
          strcpy(c_pld, optarg);
          while (strlen(c_pld) < PLD_LEN)
          {
            strcat(c_pld, " ");
          }
          write_id++;
          break;
        case 'b':
        case 'B': speed = atoi(optarg); write_id++; break;
        case 's':
        case 'S': serial = atoi(optarg); write_id++; break;
        default : usage (argv[0]); return -1; break;
      }
    }
  }

  if (write_id)
  {
    if (write_id != MAX_OPTION)
    {
      printf("Give all or none of the parameters.\n");
      usage(argv[0]);
      exit (EXIT_FAILURE);
    }

    if ((version_major > 9) || (version_minor > 9))
    {
      printf("The version number must be less or equal to 9.\n");
      usage(argv[0]);
      exit (EXIT_FAILURE);
    }

    if (serial > 99999)
    { 
      printf("The serial number must be less or equal to 99999.\n");
      usage(argv[0]);
      exit (EXIT_FAILURE);
    }

    if ((speed < 0) || (speed > 9999))
    {
      printf("The speed version must be less than 10000 Mbps.\n");
      usage(argv[0]);
      exit (EXIT_FAILURE);
    }

    sprintf(data, "DDL card %dv%d LD: %s SP: %4d Mbps S/N: %05d",
             version_major, version_minor, c_pld, speed, serial);

    printf("DIU HW id to write: '%s'\n", data);
  }

  /* open RORC and map mailboxes */
  if (rorcOpenChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);
  time = ustime * prorc.loop_per_usec;

  if (prorc.diu_version != NEW)
  {
    printf("DIU version = %s. Can not read or write HW id.\n",
            prorc.diu_version == OLD ? "OLD" : 
	    prorc.diu_version == EMBEDDED ? "EMBEDDED" : "No DIU plugged");
    rorcClose(&prorc);
    exit (EXIT_FAILURE);
  }

  if (write_id)
  {
    length = strlen(data) + 1;
    for (i = 0; i < length; i++)
    {
      param = 0x40000 | (data[i] << 8) | i ;
      ddlSendCommand(&prorc, DIU, RHWVER, 0, param, time);

      n = 0; 
      while (1)
      {
        ret = ddlWaitStatus(&prorc, time);
        if (ret >= time)
          break;
        n++;
        stw = ddlReadStatus(&prorc);
      }

      if (!n)
      {
        printf("No reply from DIU. HW id can not be written.\n");
        rorcClose(&prorc);
        exit (EXIT_FAILURE);
      }
    }
  }

  /* read HW id */

  hw = ddlSerial(&prorc, DIU, time);
  ddlInterpretHW(hw, DIU);

  /* unmap and close RORC */
  rorcClose(&prorc);

  return 0;
}

