/***********************************************************
 * diu_monitor.c
 *
 *  Monitor continiously the DIU status via RORC         
 *
 * last updated: 28/10/2003
 * written by: Ervin Denes
 ***********************************************************/
#define SOURCE_FILE "diu_monitor.c"
#define DESCRIPTION "Monitor the DIU status"

#define LINE_UP "[1A[80D[0J"

#include <ctype.h>
#include <signal.h>
#include "rorc_ddl.h"

static struct sigaction sigac;
// int interrupt_arrived = 0;
rorc_pci_dev_t prorc;
/*******************************************************/

void usage(char *name)
{
  printf("Read the DIU Status until a ^C` command via pRORC.\n");
  printf("usage: %s [-{M|m} <minor>] [-{C|c} <channel>] [-{V|v} <diu_version>]\n", name);
  printf("                  [-{T|t} <time-out>]\n");
  printf("   -M <minor>        RORC number, default; 0.\n");
  printf("   -C <channel>      RORC's channel number, default; 0.\n");
  printf("   -V <diu_version>  %d: old DIU, %d: new DIU. Default =  new DIU.\n",
                               OLD, NEW);
  printf("   -T <time-out>     waiting time for responses, default: %d usec.\n",
                               DDL_RESPONSE_TIME);
}

/********************************************************
*          signal handler                               *
********************************************************/
void sig_hand(int signum)
{
/* signum: 1(SIGHUP)  2(SIGINT)  3(SIGQUIT) */
  if(signum < 4)
  {
    interrupt_arrived = 1;
  } 
} 

/********************************************************
*          signal handling                              *
********************************************************/
void set_sig_hand(void)
{
  sigac.sa_flags = 0;
  sigac.sa_handler =  sig_hand;
  if (sigaction(SIGHUP, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGHUP ");
     exit(1);
  }
  if (sigaction(SIGINT, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGINT ");
     exit(1);
  }
  if (sigaction(SIGQUIT, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGQUIT ");
     exit(1);
  }
}


/********************************************************/

int main(int argc, char **argv)
{
  int c, minor, channel, n, diu_version, transid;
  int ustime, time;
  unsigned long status, last_status;
  unsigned int parameter;
  struct timeval tv1, tv2;
  int dsec, dusec;

  /* Check parameters */

  minor = 0;
  channel = 0;
  ustime = DDL_RESPONSE_TIME;
  diu_version = NEW;
  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:T:t:V:v:")) != EOF)
    {
      switch(toupper(c))
      {
        


case 'M': minor = atoi(optarg); break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'T': ustime = atoi(optarg); break;
        case 'V': 
          diu_version = atoi(optarg); 
          if ((diu_version != OLD) && (diu_version != NEW))
          {
            printf("Version number should be %d or %d\n", OLD, NEW);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        default : usage (argv[0]); exit (EXIT_FAILURE); break;
      }
    }
  }

  /* open RORC and map mailboxes */
  if (rorcMapChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);
  time = ustime * prorc.loop_per_usec;
  prorc.diu_version = diu_version;

  printf ("Supposed DIU version = %s\n", prorc.diu_version == OLD ? "old" : "new");

  /* clear RORC */

  if (pRorc(&prorc))
  {
    n = pRorcEmptyMb(&prorc, 1, 0);
    if (n)
      printf("Before sending DDL command IMB1 contained %d status.\n", n);
  }
  else
  {
    n = dRorcEmptyRxStatus(&prorc, 0);
    if (n) 
      printf("Before sending DDL command, DDL status register contained %d status.\n", n);
  }

  /* signal handling */
  interrupt_arrived = 0;
  set_sig_hand();

  printf("Press ^C to stop\n\n");
  fflush(stdout);
  transid = 0;
  gettimeofday(&tv1, NULL);
  last_status = 0;
  while(1)
  {
    if (interrupt_arrived)
      break;

    /* send DDL command for DIU status*/
    gettimeofday(&tv2, NULL);
    elapsed (&tv2, &tv1, &dsec, &dusec);
    status = ddlReadDiu(&prorc, transid, time);
    if (status != -1)
    {
      if (status != last_status)
      {
        parameter = status >> 12; 
        printf("%s Time: %d.%06d s, DIU stw: 0x%08lx, param.: 0x%05x (%6d)\n",
                   LINE_UP, dsec, dusec, status, parameter, parameter);
        fflush(stdout);
      }
      last_status = status;
      usleep(10000);
    }
    // transid = incr15(transid);
  }

  /* calculate and print the elapsed time */
  gettimeofday(&tv2, NULL);
  elapsed (&tv2, &tv1, &dsec, &dusec);
  printf("The program stoped in %d.%06d seconds\n", dsec, dusec);
  

  /* unmap and close RORC */
  rorcClose(&prorc);

  exit (0);
}
