/***********************************************************
 * diu_status.c
 *
 *  Read DIU status via RORC         
 *
 * last updated: 17/03/2006
 * written by: Peter Csato and Ervin Denes
 ***********************************************************/
#define SOURCE_FILE "diu_status.c"
#define DESCRIPTION "Read DIU status"

#include <ctype.h>
#include "rorc_ddl.h"

#define MAX_ADDRESS 0x20

/*******************************************************/

void usage(char *name)
{
  printf("Send a Read and Clear DIU Status` command via pRORC.\n");
  printf("usage: %s [-{M|m} <minor>] [-{C|c} <channel>] [-{V|v} <diu_version>]\n", name);
  printf("                  [-{T|t} <time-out>]\n");
  printf("   -M <minor>        RORC number, default; 0.\n");
  printf("   -C <channel>      RORC's channel number, default; 0.\n");
  printf("   -V <diu_version>  %d: old DIU, %d: new DIU. Default =  new DIU.\n",
                               OLD, NEW);
  printf("   -T <time-out>     time to wait for responses, default: %d usec.\n",
                               DDL_RESPONSE_TIME);
}

/********************************************************/

int main(int argc, char **argv)
{
  int c, minor, channel, n, diu_version, transid;
  int time, ustime;
  rorc_pci_dev_t prorc;
  unsigned long status;
  rorcHwSerial_t hw;

  /* Check parameters */

  minor = 0;
  channel = 0;
  ustime = DDL_RESPONSE_TIME;
  diu_version = NEW;
  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:T:t:V:v:")) != EOF)
    {
      switch(toupper(c))
      {
        case 'M': minor = atoi(optarg); break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'T': ustime = atoi(optarg); break;
        case 'V': 
          diu_version = atoi(optarg); 
          if ((diu_version != OLD) && (diu_version != NEW))
          {
            printf("Version number should be %d or %d\n", OLD, NEW);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        default : usage (argv[0]); exit (EXIT_FAILURE); break;
      }
    }
  }

  /* open RORC and map mailboxes */
  if (rorcMapChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);
  time = ustime * prorc.loop_per_usec;
  prorc.diu_version = diu_version;
  if (prorc.rorc_revision >= INTEG)
    prorc.diu_version = EMBEDDED;

  printf ("Supposed DIU version = %s\n",
		  prorc.diu_version == OLD ?      "old" : 
                  prorc.diu_version == EMBEDDED ? "embedded" :
                                                  "new");

  /* clear RORC */

  if (pRorc(&prorc))
  {
    n = pRorcEmptyMb(&prorc, 1, 0);
    if (n)
      printf("Before sending DDL command IMB1 contained %d status.\n", n);
  }
  else
  {
    n = dRorcEmptyRxStatus(&prorc, 0);
    if (n) 
      printf("Before sending DDL command, DDL status register contained %d status.\n", n);
  }

  transid = 0;

  if (prorc.diu_version == NEW)
  {
    /* read hw version */
    hw = ddlSerial(&prorc, DIU, time);  
    ddlInterpretHW(hw, DIU);
  } 

  if (prorc.diu_version != OLD)
  {
    /* read fw version */
    status = ddlReadFW(&prorc, DIU, time);  
    if (status != -1)
      ddlInterpretFW(status);
  }

  /* send DDL command for DIU status*/
  status = ddlReadDiu(&prorc, transid, time);
  if (status != -1)
  {
    printf(" DIU status word: 0x%08lx, i.e.\n", status);
    ddlInterpretIFSTW(&prorc, status);
  }

  /* unmap and close RORC */
  rorcClose(&prorc);

  exit (0);
}
