/***********************************************************
 * drorc_flash_programmer.c                                *
 * Programming the D-RORC flash memory                     *
 * last updated: 11/06/2006                                *
 * written by: Csaba Soos and Ervin Denes                  *
 *                                                         *
 * history:                                                *
 *                                                         *
 * v.4.5.1                                                 *
 *          10-Jun-2006: Erase, read and write functions   *
 *          10-May-2006: Data from IntelHEX file           *
 *          11-Apr-2006: Original version: rd/wr one byte  *
 *                                                         *
 **********************************************************/
#define SOURCE_FILE "drorc_flash_programmer.c"
#define DESCRIPTION "Programming the D-RORC flash memory"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/io.h>

#include "rorc_aux.h"
#include "pci.h"

#define TRUE 1
#define FALSE 0

#define LINE_UP "[1A[80D[0J"

#define DRORC_VENDOR_ID 0x10dc
#define DRORC_DEVICE_ID 0x0033
#define DRORC_REVISION 4

/* Flash interface registers' definitions */
#define FLASH_DATA_REG  0x80        /* Command/Data register */
#define DREG_DATA_BITS  0x000000ff  /* Data bits */
#define DREG_WRITE_BIT  0x00000100  /* Write bit */
#define DREG_ERASE_BIT  0x00000200  /* Erase bit */
#define DREG_RESET_BIT  0x00000400  /* Reset bit */

#define FLASH_ADDR_REG  0x84        /* Flash memory address register */
#define USER_FW_ADDR    0x00400000  /* Memory offset of the user firmware */
#define FACTORY_FW_ADDR 0x00600000  /* Memory offset of the factory firmware */

#define FW_IMAGE_SIZE   0x00200000  /* Max firmware image size */
#define FLASH_MEM_SIZE  0x00800000  /* Flash-memory size */
#define SECTOR_SIZE     0x00010000  /* Sector size : 64 KB*/
#define ERASE_MASK      0xffff0000  /* Mask bits for sector erase */

#define FLASH_STS_REG   0x80
#define STS_DATA_BITS   0x000000ff
#define STS_DONE_BIT    0x00000100
#define STS_ERROR_BIT   0x00000200
#define STS_BUSY_BIT    0x00000400

#define FLASH_BUSY_TO   2
#define FLASH_RESET_TO  20
#define FLASH_READ_TO   10
#define FLASH_WRITE_TO  100
#define FLASH_ERASE_TO  1000000

#define terminate(x...) (fprintf(stderr, "%s: ", myname), \
                         fprintf (stderr, x), \
                         pci_cleanup(all_devices), \
                         exit(EXIT_FAILURE))

#define printdb(x...) (printf (x), fflush (stdout))
#define printdbX(x...)

static char *myname = NULL;
struct pci_access *all_devices;
struct pci_dev *device;

int list = FALSE, rd = FALSE, wr = FALSE, ver = FALSE, er = FALSE;
int bus, dev, func, rev;
char *id = NULL;
char *dat_file = NULL, dat_file_name[80];

/* time-outs */
double loop_per_usec;      /* loop/us for the given machine */
int reset_to, read_to, write_to, erase_to, busy_to;

extern char *optarg;
extern int optind, opterr;

/**********************************************************/
static void usage(char *program_name)
{
  fprintf(stderr, " usage:\t%s -l\n"
                  "\t%s -e ID\n"
                  "\t%s -r ID -f fn [-v]\n"
                  "\t%s -w ID -f fn [-v]\n"
                  "\n"
                  "\t\t-l      -- list RORC devices on the PCI\n"
                  "\t\t-e      -- erase flash memory\n"
		  "\t\t-r      -- read the flash memory and write data to file\n"
		  "\t\t-r -v   -- read the flash memory and compare with file\n"
		  "\t\t-w [-v] -- write [and verify] the flash memory\n"
		  "\t\tID      -- PCI ID string in form bus:dev:func\n"
                  "\t\tfn      -- data file in IntelHEX format\n",
                  program_name, program_name, program_name, program_name);
        
   exit(EXIT_FAILURE);
}

/**********************************************************/
int parsePciBusString(const char *busID, int *bus, int *device, int *func)
{
  /*
   * The format is assumed to be "bus:device:func", where bus, device
   * and func are hexadecimal integers. func may be omitted and assumed to
   * be zero, although it doing this isn't encouraged.
   */

  char *p, *s, *end;

  s = strdup(busID);

  p = strtok(s, ":");
  if (p == NULL || *p == 0)
    return FALSE;
  *bus = strtoul(p, &end, 16);
  if (*end != '\0')
    return FALSE;

  p = strtok(NULL, ":");
  if (p == NULL || *p == 0)
    return FALSE;
  *device = strtoul(p, &end, 16);
  if (*end != '\0')
    return FALSE;

  *func = 0;
  p = strtok(NULL, ":");
  if (p == NULL || *p == 0)
    return TRUE;
  *func = strtoul(p, &end, 16);
  if (*end != '\0')
    return FALSE;

  return TRUE;
}

/**********************************************************/
void treat_param(int argc, char **argv)
{
  int i, id = FALSE, c;

  myname = argv[0];
  for (i = 0; i < argc; i++)
    printf("%s ", argv[i]);
  printf("\n");

  while ((c = getopt(argc, argv, "le:r:w:vf:")) != -1)
  {
    switch (tolower(c))
    {
      case 'l':
        list = TRUE;
        break;
      case 'e':
        er = TRUE;
        id = parsePciBusString(optarg, &bus, &dev, &func);
        break;
      case 'r':
        rd = TRUE;
        id = parsePciBusString(optarg, &bus, &dev, &func);
        break;
      case 'w':
        wr = TRUE;
        id = parsePciBusString(optarg, &bus, &dev, &func);
        break;
      case 'v':
        ver = TRUE;
        break;
      case 'f':
        strcpy(dat_file_name, optarg);
        dat_file = dat_file_name;
        break;
      default:
        usage(myname);
    }
  }

  if (!(list || er || rd || wr))
    usage(myname);
  if ((list && er) || (rd && wr))
    usage(myname);
  if ((list || er) && (rd || wr))
    usage(myname);
  if ((list || er) && (dat_file != NULL))
    usage(myname);
  if ((er || rd || wr) && !id) 
  {
    fprintf(stderr, "%s: Bad PCI ID string\n", myname);
    usage(myname);
  }
  if ((rd || wr) && (dat_file == NULL))
  {
    fprintf(stderr, "%s: Data file name not given\n", myname);
    usage(myname);
  }
  if (ver && (er || list))
    usage(myname);
}

/**********************************************************/
int flashReset (struct pci_dev *device)
{
  int i;

  /* Reset the flash memory */
  pci_write_long(device, FLASH_DATA_REG, DREG_RESET_BIT);

  /* Wait for the flash memory to become ready after the reset */
  for (i = 0; i < reset_to; i++)
  {
    if(!(pci_read_long(device, FLASH_STS_REG) & STS_BUSY_BIT))
      return 0;
  }

  return -1;
}

/**********************************************************/
int flashEraseSector (struct pci_dev *device, int address)
{
  int i, status;

  /* Put the sector address into the address register */
  pci_write_long(device, FLASH_ADDR_REG, address & ERASE_MASK);

  /* Wait a bit ... */
  usleep(1);

  /* Put the erase command into the data register */
  pci_write_long(device, FLASH_DATA_REG, DREG_ERASE_BIT);

  /* Wait until STS_BUSY_BIT is set */
  i=0;
  while(i < busy_to)
  {
    i++;
    status = pci_read_long(device, FLASH_STS_REG);
    if(status & STS_DONE_BIT)
      return 0;
    if(status & STS_BUSY_BIT)
      break;
  }
  if (i == busy_to)
    return -1;

  /* Wait for the flash to become ready after the erase */
  for(i = 0; i < erase_to; i++)
  {
    if(pci_read_long(device, FLASH_STS_REG) & STS_DONE_BIT)
      return 0;
  }

  return -1;
}

/**********************************************************/
int flashReadByte (struct pci_dev *device, int address)
{
  int i, status;

  /* Put the address into the address register */
  pci_write_long(device, FLASH_ADDR_REG, address);
		
  /* Wait until STS_BUSY_BIT is set */
  i=0;
  while(i < busy_to)
  {
    i++;
    status = pci_read_long(device, FLASH_STS_REG);
    if(status & STS_DONE_BIT)
      return status & STS_DATA_BITS; 
    if(status & STS_BUSY_BIT)
      break;
  }
  if (i == busy_to)
    return -1;

  /* Read until the request is done or the timeout expires */
  i = 0;
  while(i < read_to)
  {
    i++;
    status = pci_read_long(device, FLASH_STS_REG) & STS_DONE_BIT;
    if (status & STS_DONE_BIT)
      return status & STS_DATA_BITS; 
  }

  return -1;
}

/**********************************************************/
int flashWriteByte (struct pci_dev *device, int data)
{
  int i, addr, status;

  /* read address register */
  if (data == 0xFF)
  {
    addr = pci_read_long(device, FLASH_ADDR_REG);
    addr++;
    pci_write_long(device, FLASH_ADDR_REG, addr);
    return 0;
  }

  /* Put the data into the data register */
  pci_write_long(device, FLASH_DATA_REG, data|DREG_WRITE_BIT);

  /* Wait until STS_BUSY_BIT is set */
  i=0;
  while(i < busy_to)
  {
    i++;
    status = pci_read_long(device, FLASH_STS_REG);
    if(status & STS_DONE_BIT)
      return 0;
    if(status & STS_BUSY_BIT)
      break;
  }
  if (i == busy_to)
    return -1;

  /* Read until the request is done or the timeout expires */
  i = 0;
  while(i < write_to)
  {
    i++;
    if(pci_read_long(device, FLASH_STS_REG) & STS_DONE_BIT)
      return 0;
  }
  return -1;

}

/**********************************************************/
int main(int argc, char *argv[])
{
  int i, ilen, value = 0, dat_val[255], ulba_rec[2];
  int rec_num, rec_len, rec_type, addr, err_code;
  int identical, empty, isect, nsect;
  int sector_empty[FW_IMAGE_SIZE / SECTOR_SIZE];
  unsigned int ulba, rba, istart, iend, offset;
  struct pci_filter dev_filter;
  FILE * fp;
  unsigned char* data_pt;
  char err_txt[80];
  struct timeval tv1, tv2;
  int max_loop, dsec, dusec;
  double dtime;
	
  myname = argv[0];
  treat_param(argc, argv);

  all_devices = pci_alloc();
  if (all_devices == NULL)
  {
    fprintf(stderr, " PCI memory allocation failed\n");
    exit(EXIT_FAILURE);
  }
  pci_init(all_devices);
  pci_scan_bus(all_devices);

  if (all_devices->devices == NULL)
  {
    fprintf(stderr, " No device has been found\n");
    exit(EXIT_FAILURE);
  }

  if (list)
  {
    /* Listing all the RORC devices found on the PCI*/
    i = 0;
    device = all_devices->devices;
    while(device)
    {
      if ((device->vendor_id == DRORC_VENDOR_ID) &&
	  (device->device_id == DRORC_DEVICE_ID))
      {
        i++;
        rev = pci_read_byte(device, 8);
        printf(" (%03d:%03d:%03d)  VENDOR: 0x%04x, DEVICE: 0x%04x,"
                                 " REVISION: 0x%02x\n",
                  device->bus, device->dev, device->func,
                  device->vendor_id, device->device_id, rev);
      }
      device = device->next;
    }
    if (!i)
      terminate(" No RORC device has been found\n");

    pci_cleanup(all_devices);
    exit(EXIT_SUCCESS);
  }

  // Am I super user? */
  if(iopl(3))
    terminate("%s (being root helps)\n", strerror(errno));

  printdb (" Device selected: (%03d:%03d:%03d)\n", bus, dev, func);

  /* Let's start from the beginning of the chain */
  device = all_devices->devices;
  while (device != NULL)
  {
    /* Find the device (ID) that was requested by the user */
    if((device->bus == bus) && (device->dev == dev) && (device->func == func))
      break;
    device = device->next;
  }
  if (device == NULL)
  {
    /* We reached the end of the list, but we have not found
     * the device. Let's clean up and leave the program. */
    terminate("Device %03d:%03d:%03d has not been found\n", bus, dev, func);
  }

  /* Initialize device filter*/
  pci_filter_init(all_devices, &dev_filter);
  /* Set up filter */
  dev_filter.vendor = DRORC_VENDOR_ID;
  dev_filter.device = DRORC_DEVICE_ID;
  if (!pci_filter_match(&dev_filter, device))
  {
    /* The VENDOR and DEVICE ID of this device does not match
     * with the IDs of the D-RORC. Clean up and exit. */
    terminate("Uknown device\n");
  }

  /* Device found */
  rev = pci_read_byte(device, 8);
  printf (" (%03d:%03d:%03d)  VENDOR: 0x%04x, DEVICE: 0x%04x,"
                                            " REVISION: 0x%02x\n",
            device->bus, device->dev, device->func,
            device->vendor_id, device->device_id, rev);

  // Check revision number 
  if (rev != DRORC_REVISION)
    terminate("Revision number error. Expected: 0x%02x, found: 0x%02x\n",
                            DRORC_REVISION, rev);

  /* calibrate loop time for time-outs */
  max_loop = 1000000;
  gettimeofday(&tv1, NULL);

  for (i = 0; i < max_loop; i++)
    pci_read_long(device, FLASH_STS_REG);

  gettimeofday(&tv2, NULL);
  elapsed(&tv2, &tv1, &dsec, &dusec);
  dtime = (double)dsec * MEGA + (double)dusec;
  loop_per_usec = (double)max_loop/dtime;
  printdb(" Calculated loop_per_usec: %.2f\n", loop_per_usec);
  if (loop_per_usec < 2.)
    loop_per_usec += 2.;
  busy_to  = loop_per_usec * FLASH_BUSY_TO;
  reset_to = loop_per_usec * FLASH_RESET_TO;
  read_to  = loop_per_usec * FLASH_READ_TO;
  write_to = loop_per_usec * FLASH_WRITE_TO;
  erase_to = loop_per_usec * FLASH_ERASE_TO;

  // Is the flash memory there?
  /* Put an address into the address register and read it back*/
  pci_write_long(device, FLASH_ADDR_REG, USER_FW_ADDR);
  usleep(1);
  value = pci_read_long(device, FLASH_ADDR_REG);
  if (value != USER_FW_ADDR)
    terminate("Can not read flash memory.\n"
              " Data read: 0x%08x, expected: 0x%08x\n", value, USER_FW_ADDR);

  // Reset the flash memory
  if(flashReset(device))
    terminate("Reset timeout\n");
  printf(" Flash memory reset\n");

  /* is the flash memory empty? */
  nsect = FW_IMAGE_SIZE / SECTOR_SIZE;
  empty = TRUE;
  for (i = 0; i < nsect; i++)
    sector_empty[i] = TRUE;
  isect = 0;
  for (offset = USER_FW_ADDR; offset < FACTORY_FW_ADDR; offset++)
  {
    value = flashReadByte(device, offset);
    if (value < 0)
      terminate("Read timeout at address 0x%08x\n", offset);
    printdbX (" 0x%08x: 0x%02x\n", offset, value);
    if (empty && (value != 0xFF))
    {
      printdb(" Flash memory is not empty. Address 0x%08x contains 0x%02x\n", 
                                                     offset, value);
      empty = FALSE;
    }
    if (sector_empty[isect] && (value != 0xFF))
    {
      printdbX(" Sector %d is not empty. Address 0x%08x contains 0x%02x\n", 
                                                isect + 1, offset, value);
      sector_empty[isect] = FALSE;
    }
    if (!((offset + 1) % SECTOR_SIZE))
      isect++;
  }
  printdb(" Empty (O) and not empty (X) sectors:\n");
  for (i = 0; i < nsect; i++)
    printdb(" %c", sector_empty[i] ? 'O' : 'X');
  printdb("\n");

  if (empty)
  {
    if (er)
    {
      printdb(" Flash memory is empty\n");
      pci_cleanup(all_devices);
      exit(EXIT_SUCCESS);
    }
    if (rd)
    {
      if (ver)
        terminate(" Flash memory is empty. Verification not done\n");
      else
        terminate(" Flash memory is empty. File not written\n");
    }
    printdb(" Flash memory is empty\n");
  }

  if (er)
  {
    /* erase flash memory */
    printf("\n");
    for (offset = USER_FW_ADDR, isect = 0; offset < FACTORY_FW_ADDR; 
                                               offset += SECTOR_SIZE, isect++)
    {
      if (sector_empty[isect])
        continue;
      if (flashEraseSector(device, offset))
        terminate("Erase timeout at sector 0x%08x\n", offset);
      printdb ("%s Sector 0x%08x erased\n", LINE_UP, offset);
    }
    printf("%s Flash memory erased\n", LINE_UP);
    pci_cleanup(all_devices);
    exit(EXIT_SUCCESS);
  }

  /* allocate memory */
  data_pt = malloc(FW_IMAGE_SIZE);
  if (data_pt == NULL)
    terminate("Can not allocate memory for flash data.\n");
  for (i = 0; i < FW_IMAGE_SIZE; i++)
    *(data_pt + i) = 0xFF;

  /* Open data file */
  fp = fopen(dat_file_name, "r");
  if (rd && !ver)
  {
    if (fp != NULL)
      terminate("File %s exists.\n"
                "Remove it or choose another file for writing\n", dat_file_name);
    fp = fopen(dat_file_name, "w");
  }
  if (fp == NULL)
   terminate ("Can not open file %s. %s\n", dat_file_name, strerror(errno));

  if (wr || ver)
  {
    /* Read data file */
    rec_num = 1;
    ulba = 0;
    istart = FW_IMAGE_SIZE;
    iend = 0;
    while(1)
    {
      /* Read a record */
      err_code = getIntelHexRecord(fp, &rec_len, &addr, &rec_type, dat_val,
                                                                     err_txt);
      if (err_code)
        terminate("IntelHEX format error in file %s, record %d:\n" 
                  "          %s\n\n", dat_file_name, rec_num, err_txt);
  
      rba = ulba + addr;        // Record Base Address;
      printdbX("rec %d: len=%d, addr=0x%04x, type=%d, rba=0x%08x\n",
                   rec_num, rec_len, addr, rec_type, rba);
      if (rec_len)
      {
        printdbX("data:\n");
        for (i = 0; i < rec_len; i++)
          printdbX(" %02x", dat_val[i]);
        printdbX("\n");
      }
  
      if ((rec_type != 0) && (rec_type != 1) && (rec_type != 4))
       terminate ("IntelHEX format error in file %s, record %d:\n" 
                  "          not allowed record type (%d)\n\n", 
                                          dat_file_name, rec_num, rec_type);
  
      if (rec_type == 1)          // EOF
        break;
  
      rec_num++;
      if (rec_type == 4)          // Upper Linear Base Address
      {
        ulba = ((dat_val[0] << 8) + dat_val[1]) << 16;
        continue;
      }
  
      if ((rba + rec_len) > FW_IMAGE_SIZE)
        terminate("IntelHEX format error in file %s, record %d:\n" 
                  "          addr (0x%08x) + len (%d) is larger then"
                  " fw image size (%dMB)\n\n",
            dat_file_name, rec_num, rba, rec_len, FW_IMAGE_SIZE/1024/1024);
      if (istart > rba)
        istart = rba;
      if (iend < (rba + rec_len))
        iend = rba + rec_len;
      for (i = 0; i < rec_len; i++)
        *(data_pt + rba + i) = dat_val[i];
    } 
    fclose(fp);
    ilen = iend - istart;
    printf(" %d IntelHEX records read from file %s\n", rec_num, dat_file_name);
    printf(" Image length: %d bytes (%d.%d KB) from address 0x%x to address 0x%x\n", 
                                ilen, ilen/1024, ilen%1024, istart, iend - 1);
  }

  if (rd && !ver)
  {
    istart = 0;
    iend = FW_IMAGE_SIZE;
    ilen = FW_IMAGE_SIZE;
  }

  if (wr)
  {
    /* Can we write to these locations? */
    if ((USER_FW_ADDR + iend) > FACTORY_FW_ADDR)
      terminate("Attempt to write to read-only address (to address >= 0x%x).\n"
                      " Flash memory not changed\n", FACTORY_FW_ADDR);

    /* check flash memory */
    if (!empty)
    {
      identical = TRUE;
      offset = USER_FW_ADDR + istart;
      for (i = 0; i < ilen; i++)
      {
        value = flashReadByte(device, offset);
        printdbX (" 0x%08x: 0x%02x\n", offset, value);
        if (value < 0)
          terminate("Read timeout at address 0x%08x\n", offset);

        /* compare data */
        if (*(data_pt + i) != value)
        {
          identical = FALSE;
          printdb (" Flash memory and data file %s differ at address 0x%08x.\n"
                    " Data read: 0x%02x, expected: 0x%02x\n",
                               dat_file_name, offset, value, *(data_pt + i));
          break;
        }
        offset++;
      }

      if (identical)
      {
        printdb(" %d bytes (%d.%d KB) read from flash memory addresses 0x%x to 0x%x\n",
                 ilen, ilen/1024, ilen%1024,
                            USER_FW_ADDR + istart, USER_FW_ADDR + iend-1); 
        printf(" Flash memory and data file %s are identical\n", dat_file_name);
        pci_cleanup(all_devices);
        exit(EXIT_SUCCESS);
      }

      /* erase flash memory if necessary */
      printf("\n");
      for (offset = USER_FW_ADDR, isect = 0; offset < FACTORY_FW_ADDR;
                                        offset += SECTOR_SIZE, isect++)
      {
        if (sector_empty[isect])
          continue;
        if (flashEraseSector(device, offset))
          terminate("Erase timeout at sector 0x%08x\n", offset);
        printdb ("%s Sector 0x%08x erased\n", LINE_UP, offset);
      }
      printf("%s Flash memory erased\n", LINE_UP);
    }

    /* Put the start address into the address register */
    offset = USER_FW_ADDR + istart;
    pci_write_long(device, FLASH_ADDR_REG, offset);
    usleep(1);

    /* write data bytes */

    printf(" \n");
    for (i = 0; i < ilen; i++)
    {
      if (flashWriteByte(device, *(data_pt + istart + i)))
        terminate("Write timeout at address 0x%08x\n", offset);
      offset++;
      if (!(offset % SECTOR_SIZE))
      {
        printdb ("%s Flash memory sector 0x%08x written\n", LINE_UP,
                                                   offset - SECTOR_SIZE);
      }
    }

    printf ("%s %d bytes (%d.%d KB) written to flash memory addresses 0x%x to 0x%x\n",
               LINE_UP, ilen, ilen/1024, ilen%1024,
                            USER_FW_ADDR + istart, USER_FW_ADDR + iend-1); 
  }

  if (rd || ver)
  {
    offset = USER_FW_ADDR + istart;
//  offset = FACTORY_FW_ADDR + istart;
    for (i = 0; i < ilen; i++)
    {
      if (offset > FLASH_MEM_SIZE-1)
        terminate("Wrong address 0x%08x\n", offset);

      value = flashReadByte(device, offset);
      if (value < 0)
        terminate("Read timeout at address 0x%08x\n", offset);
      printdbX (" 0x%08x: 0x%02x\n", offset, value);

      if (!ver)
      {
        /* memorise data for file */
        *(data_pt + i) = value;
      }
      else
      {
        /* compare data */
        if (*(data_pt + i) != value)
          terminate("Verify error at address 0x%08x.\n"
                    " Data read: 0x%02x, expected: 0x%02x\n",
                                 offset, value, *(data_pt + i));
      }
      offset++;
    }
    printf (" %d bytes (%d.%d KB) read from flash memory addresses 0x%x to 0x%x\n",
                 ilen, ilen/1024, ilen%1024,
                            USER_FW_ADDR + istart, USER_FW_ADDR + iend-1); 
    if (ver)
      printf(" Data verification is OK\n");
  }

  if (rd && !ver)
  {
    /* write dat to file */
    rec_len = 128;
    rec_num = 1;
    ulba = -1;
    for (addr = istart, rec_num = 1; addr < iend; addr += rec_len)
    {
      rba = addr >> 16;
      if (rba != ulba)
      {
        ulba = rba;
        ulba_rec[0] = ulba >> 8;
        ulba_rec[1] = ulba & 0xFF; 
        err_code = putIntelHexRecord(fp, 2, 0, 4, ulba_rec, err_txt);
        if (err_code)
          terminate("Write error in file %s, record %d:\n" 
                "          %s\n\n", dat_file_name, rec_num, err_txt);
        rec_num++;
      }

      if ((addr + rec_len) >= iend)
        rec_len = iend - addr; 
      for (i = 0; i < rec_len; i++)
        dat_val[i] =  *(data_pt+addr+i);

      rec_type = 0;
      err_code = putIntelHexRecord(fp, rec_len, addr & 0xFFFF, rec_type, 
                                                          dat_val, err_txt);
      if (err_code)
        terminate("Write error in file %s, record %d:\n" 
                  "          %s\n\n", dat_file_name, rec_num, err_txt);
      rec_num++;
    }

    /* write EOF */ 
    err_code = putIntelHexRecord(fp, 0, 0, 1, dat_val, err_txt);
    if (err_code)
      terminate("Write error in file %s, record %d:\n" 
                "          %s\n\n", dat_file_name, rec_num, err_txt);
    fclose(fp);
    printf(" %d IntelHEX records written to file %s\n", rec_num, dat_file_name);
    printf(" Image length: %d bytes (%d.%d KB) from address 0x%x to address 0x%x\n", 
                             ilen, ilen/1024, ilen%1024, istart, iend - 1);
  }
	
  /* Clean up and exit */
  pci_cleanup(all_devices);
  exit(EXIT_SUCCESS);
}
