/***********************************************************
 * prbs_test.c
 *
 *  Monitor PRBS (Pseudo Random Bit Sequence) test      
 *
 * last updated: 21/06/2005
 * written by: Ervin Denes
 ***********************************************************/
#define SOURCE_FILE "prbs_test.c"
#define DESCRIPTION "PRBS (Pseudo Random Bit Sequence) test program"

#define LINE_UP "[1A[80D[0J"
#define LOG_FILE_NAME "prbs.log"
#define SLEEP_TIME 10000
#define NEW_LINE_TIME 3600

#include <ctype.h>
#include <signal.h>
#include "rorc_ddl.h"

static struct sigaction sigac;
// int interrupt_arrived = 0;
rorc_pci_dev_t prorc;
/*******************************************************/

void usage(char *name)
{
  printf("Read PRBS status until a ^C.\n");
  printf("usage: %s [-{M|m} <minor>] [-{C|c} <channel>] [-{L|l} <log_file>]\n",
                                                                      name);
  printf("                 [-{S|s} <sleep_time>] [-{N|n} <NL_time>] [-{T|t} <time-out>]\n");
  printf("   -M <minor>        RORC number, default; 0.\n");
  printf("   -C <channel>      RORC's channel number, default; 0.\n");
  printf("   -L <log_file>     log file name. Default: %s.\n", LOG_FILE_NAME);
  printf("   -S <sleep_time>   time between PRBS status reads. Default: %d usec.\n",
                               SLEEP_TIME);
  printf("   -S <NL_time>      time when a new line is written in the log file.\n");
  printf("                     Default: %d s.\n", NEW_LINE_TIME);
  printf("   -T <time-out>     waiting time for responses, default: %d usec.\n",
                               DDL_RESPONSE_TIME);
}

/********************************************************
*          signal handler                               *
********************************************************/
void sig_hand(int signum)
{
/* signum: 1(SIGHUP)  2(SIGINT)  3(SIGQUIT) */
  if(signum < 4)
  {
    interrupt_arrived = 1;
  } 
} 

/********************************************************
*          signal handling                              *
********************************************************/
void set_sig_hand(void)
{
  sigac.sa_flags = 0;
  sigac.sa_handler =  sig_hand;
  if (sigaction(SIGHUP, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGHUP ");
     exit(1);
  }
  if (sigaction(SIGINT, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGINT ");
     exit(1);
  }
  if (sigaction(SIGQUIT, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGQUIT ");
     exit(1);
  }
}


/********************************************************/

int main(int argc, char **argv)
{
  int c, minor, channel, ret, transid;
  int ustime, timeout, nl_time, sleep_time;
  unsigned long status, last_status;
  unsigned int parameter;
  struct timeval tv1, tv2;
  time_t time_val, time_now;
  int dsec, dusec;
  int printed, prdsec;
  int hour, min, sec;
  char log_file_name[80];
  FILE *lf;
  stword_t stw;

  /* Check parameters */

  minor = 0;
  channel = 0;
  ustime = DDL_RESPONSE_TIME;
  sleep_time = SLEEP_TIME;
  nl_time = NEW_LINE_TIME;
  strcpy(log_file_name, LOG_FILE_NAME);
  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:S:s:N:n:T:t:L:l:")) != EOF)
    {
      switch(toupper(c))
      {
        case 'M': minor = atoi(optarg); break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'S': sleep_time = atoi(optarg); break;
        case 'N': nl_time = atoi(optarg); break;
        case 'T': ustime = atoi(optarg); break;
        case 'L': strcpy(log_file_name, optarg); break;
        default : usage (argv[0]); exit (EXIT_FAILURE); break;
      }
    }
  }

  /* signal handling */
  interrupt_arrived = 0;
  set_sig_hand();

  /* open log file */
  if ((lf = fopen(log_file_name, "a")) == NULL)
  {
    printf ("Can not open file %s. ", log_file_name);
    fflush (stdout);
    perror ("");
    exit (EXIT_FAILURE);
  }

  /* open RORC and map mailboxes */
  if (rorcMapChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);
  timeout = ustime * prorc.loop_per_usec;

  /* check SIU fw */
  ret = ddlSendCommand(&prorc, SIU, IFLOOP, 0, 0, timeout);
  if (ret == RORC_LINK_NOT_ON)
  {
    printf("Can not check SIU fw, since link is not on.\n");
    exit (EXIT_FAILURE);
  }
  if (ret == RORC_TIMEOUT)
  {
    printf("Can not send command for cheking SIU fw.\n");
    exit (EXIT_FAILURE);
  }

  ret = ddlWaitStatus(&prorc, timeout);
  if (ret >= timeout)
  {
    printf("No reply arrived from SIU.\n");
    exit (EXIT_FAILURE);
  }

  stw = ddlReadStatus(&prorc);
  if (stw.part.code == ILCMD)
  {
    printf("This SIU fw does not support PRBS test mode.\n");
    exit (EXIT_FAILURE);
  }

  if (stw.part.code != CTSTW)
  {
    printf("Unexpected reply from SIU: 0x%08lx.\n", stw.stw);
    exit (EXIT_FAILURE);
  }

  /* reset DIU */
  rorcArmDDL(&prorc, RORC_RESET_DIU);

  /* wait a little before SIU reset */
  usleep(1000);

  /* reset SIU */
  if (rorcArmDDL(&prorc, RORC_RESET_SIU) != RORC_STATUS_OK)
  {
    printf("Error: Could not reset SIU. I stop.\n");
    rorcClose(&prorc);
    exit (EXIT_FAILURE);
  }

  /* wait a little before SIU Test Mode */
  usleep(1000);

  /* set SIU to PRBS test mode */
  ret = ddlSetSiuTestMode(&prorc, timeout, &stw);
  if (ret == RORC_LINK_NOT_ON)
  {
    printf("Can not set SIU to Test Mode, since link is not on.\n");
    exit (EXIT_FAILURE);
  }
  if (ret != RORC_STATUS_OK)
  {
    printf("Can not sent SIU to Test Mode.\n");
    exit (EXIT_FAILURE);
  }

  /* set DIU to PRBS test mode and clear error counter*/
  ret = ddlStartTestMode(&prorc, timeout, &stw);
  if (ret != RORC_STATUS_OK)
  {
    printf("Can not send DIU command.\n");
    exit (EXIT_FAILURE);
  }

  /* start log file */
  gettimeofday(&tv1, NULL);
  time_now = time(&time_val);
  fprintf(lf, "PRBS test. Starting time: %s", ctime(&time_val));
  fflush(lf);

  /* start loop */
  printf("Press ^C to stop\n\n");
  fflush(stdout);
  transid = 0;
  last_status = 0;
  printed = 0;

  while(1)
  {
    if (interrupt_arrived)
      break;

    /* read PRBS error counter */
    gettimeofday(&tv2, NULL);
    elapsed (&tv2, &tv1, &dsec, &dusec);
    ret = ddlReadPRBS(&prorc, transid, timeout, &stw);
    if (ret == RORC_STATUS_OK)
    {
      status = stw.stw;
      parameter = status >> 12; 
      printf("%s Time: %d.%02d s, PRSTW status: 0x%08lx, errors: %6d\n",
                 LINE_UP, dsec, dusec/10000, status, parameter);
      fflush(stdout);
      prdsec = !(dsec % nl_time); // dsec is multiple of nl_time
      if ((status != last_status) || (prdsec && !printed))
      {
        fprintf(lf, "Time: %d.%06d s, PRSTW status: 0x%08lx, errors: %6d\n",
                     dsec, dusec, status, parameter);
        fflush(lf);
      }
      last_status = status;
      printed = prdsec;
      usleep(sleep_time);
    }
    else
      printf("\nDid not succeed to read PRBS error count\n\n");
    // transid = incr15(transid);
  }

  /* calculate elapsed time */
  gettimeofday(&tv2, NULL);
  elapsed (&tv2, &tv1, &dsec, &dusec);

  /* close log file */
  time_now = time(&time_val);
  fprintf(lf, "PRBS test ended after %d.%06d seconds of run at %s.\n", 
                                             dsec, dusec, ctime(&time_val));
  fclose (lf);

  /* stop PRBS test mode */
  ret = ddlStopTestMode(&prorc, timeout, &stw);
  if (ret == RORC_STATUS_OK)
    printf("\nPRBS test stopped.\n");
  else
    printf("\nCan not stop PRBS test mode.\n");

  /* unmap and close RORC */
  rorcClose(&prorc);

  hour = dsec / 3600;
  min = ((dsec % 3600) /60);
  sec = dsec - hour * 3600 - min * 60;
  printf("The program stoped in %d.%02d seconds (%dh %02dm %02d.%02ds).\n", 
                    dsec, dusec/10000, hour, min, sec, dusec/10000);

  exit (0);
}
