/*
 ***************************************************
 * rorc_create_rf.c
 *
 *  Allocate physmem memory for Ready FIFO 
 *  and give its address to RORC
 *
 * last updated: 22/08/2003
 * written by: Peter Csato and Ervin Denes
 ***************************************************
*/
#define SOURCE_FILE "rorc_create_rf.c"
#define DESCRIPTION "Allocate Ready FIFO/Receive Report Base Address"

#include <ctype.h>
#include "rorc_ddl.h"

#define RF_SIZE 128

/*******************************************************/

void usage(char *name)
{
  printf("usage: %s [-{M|m} <RORC minor>|0] [-{C|c} <RORC channel>|0]\n", name);
  printf("                      [-{L|l} <RF size>|%d] [-{O|o} <offset in kilobytes>|0]\n",
                     RF_SIZE);
}

/********************************************************/

int main(int argc, char **argv)
{
  int c, minor, channel, rf_size;
  rorc_pci_dev_t prorc;
  int physmem_fd;
  unsigned long phys_size_physmem, rf_offset;
  volatile unsigned long *addr_user;
  unsigned long *min_rf_word, *max_rf_word, *rf_addr;
  unsigned long phys_addr, min_rf_phys, max_rf_phys;
  unsigned long rf_offset_byte, rf_size_byte;

  /* open physmem */
  if (physmemOpen (&physmem_fd, &addr_user, &phys_addr,
                   &phys_size_physmem))
    exit (EXIT_FAILURE);

  printf(" Allocation of the physmem memory.\n");
  printf(" Physical address: 0x%08lx.\n", phys_addr);
  printf(" Size of the physmem memory: %ld (0x%08lx) bytes, %ld (0x%08lx) words.\n",
           phys_size_physmem, phys_size_physmem,
           phys_size_physmem >> 2, phys_size_physmem >> 2);


  /* default parameters */
  minor = 0;
  channel = 0;
  rf_size = RF_SIZE;
  rf_offset = 0;

  /* Check parameters */
  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:L:l:O:o:")) != EOF)
    {
      switch(toupper(c))
      {
        case 'M': minor = atoi(optarg); break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'L': rf_size = atoi (optarg); break;
        case 'O': rf_offset = atoi(optarg); break;
        case '?': usage (argv[0]); exit (EXIT_FAILURE); break;
        case -1:  usage (argv[0]); exit (EXIT_FAILURE); break;
      }
    }
  }

  rf_offset_byte = rf_offset * 1024;
  rf_size_byte = rf_size * 8;
  if ((rf_offset_byte + rf_size_byte) >= phys_size_physmem)
  {
    printf(" Offset (%ld bytes) + RF size (%ld bytes) is more than the physical size (%ld bytes).\n",
                rf_offset_byte, rf_size_byte, phys_size_physmem);
    exit (EXIT_FAILURE);
  } 

  /* fill ready FIFO with -1 */
  min_rf_word = (unsigned long *)addr_user + rf_offset_byte/4;
  max_rf_word = min_rf_word + rf_size_byte/4;

  for (rf_addr = min_rf_word; rf_addr < max_rf_word; rf_addr++)
    *rf_addr = -1;

  min_rf_phys = phys_addr + rf_offset_byte;
  max_rf_phys = min_rf_phys + rf_size_byte -4;
  printf(" RF start: 0x%08lx, end: 0x%08lx\n", min_rf_phys, max_rf_phys);

  /* open RORC and map mailboxes */
  if (rorcMapChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);

  if (pRorc(&prorc))
  {
    /* set physical address in OMB2 and send "Put Ready FIFO base" command */
    pRorcInitCmdProc (&prorc);
    pRorcWriteMb(&prorc, 2, min_rf_phys);
    pRorcWriteMb(&prorc, 1, PRORC_CMD_PUT_READY_BASE);
    pRorcWaitCmdProc (&prorc);
  }
  else
  {
    dRorcWriteReg(&prorc, C_RRBAR, min_rf_phys);
  }

  printf(" Ready FIFO created in physmem area, length = %ld bytes.\n",
           rf_size_byte);
  printf(" Physical address 0x%08lx sent to RORC %d channel %d as Ready FIFO base address\n", 
           min_rf_phys, minor, channel);

  /* unmap and close RORC */
  rorcClose(&prorc);

  /* unmap and close physmem memory */
  if (physmemClose (physmem_fd, (unsigned long *)addr_user, phys_size_physmem))
    exit (EXIT_FAILURE);

  exit (EXIT_SUCCESS);

}
