/************************************************************
* Device driver methods for pRORC and D-RORC cards          *
*                                                           *
* last updated: 07/08/2006                                  * 
* made for LINUX kernel 2.4 and 2.6                         *
* written by: Peter Csato and Ervin Denes                   *
*                                                           *
* history:                                                  *
*                                                           *
* v.5.1.2 07-Aug-2006: adding licence infos                 *
*                                                           *
* v.5.1.1 27-Jul-2006: modifications for SLC4 (kernel 2.6)  *
*                                                           *
* v.4.2   04-Nov-2003: add version release number           *
*                                                           *
* v.4.2   25-Jul-2003: include 2 DDL channel for a RORC     *
*                                                           *
* v.4.1   02-Jul-2003: fix the signature change of          *
*                      remap_range_page for kernel          *
*                      version 2.4.20 onwards               *
*                                                           *
*************************************************************/
#define SOURCE_FILE "rorc_driver"
#define DESCRIPTION "Device driver methods for pRORC and D-RORC cards"

#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/kernel.h>
#include <linux/init.h>

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "S593X.h"
#include "drorc_regs.h"
#include "rorc_driver.h"
#include "rorc_version.h"

MODULE_DESCRIPTION("RORC DRIVER");
MODULE_AUTHOR("Ervin Denes and Peter Csato");
MODULE_LICENSE("GPL");

/* PLD vendor/id */
/*
#define RORC2_VENDOR   0x1556
#define RORC2_ID       0x5555
*/

/* ALTERA vendor/id */
/*
#define RORC2_VENDOR	0x1172
#define RORC2_ID	0x0004
*/

/* CERN vendor/id */
#define RORC2_VENDOR    0x10dc
#define RORC2_ID        0x0033

#define DEVICE_MAJOR	125
#define DEVICE_NAME	"prorc"

struct pci_device_id rorc_pci_tbl[] __devinitdata = {
         { RORC2_VENDOR, RORC2_ID, PCI_ANY_ID, PCI_ANY_ID },
         { 0, 0, 0, 0 },
 };

/* Method prototypes */
int rorc_init_one(struct pci_dev *dev, const struct pci_device_id *ent);
void rorc_remove_one(struct pci_dev *dev);
int configure_device(rorc_t *dev);


/* File operations supported on /dev/prorc devices */
static int rorc_open(struct inode* inode, struct file* file);
static int rorc_close(struct inode* inode, struct file* file);
static int rorc_mmap(struct file *filp, struct vm_area_struct *vma);
static int rorc_ioctl(struct inode* inode, struct file* file, unsigned int cmd, unsigned long arg);

struct pci_driver rorc_driver = {
         name :          DEVICE_NAME, 
         id_table :      rorc_pci_tbl,
         probe :         rorc_init_one,
         remove :        rorc_remove_one,
 };

struct file_operations dev_fops = {
   ioctl:   rorc_ioctl,           /* ioctl */
   mmap:    rorc_mmap,            /* mmap */
   open:    rorc_open,            /* open */
   release: rorc_close,           /* release */
};

struct pci_dev *dev = NULL;
rorc_t *dv;
static rorc_t prorc[MAX_DEVICE];
static int device_count;
char *kernel_version [] = {"2.4", "2.6"};

/********************************************/
int __init rorc_init_one(struct pci_dev *dev, const struct pci_device_id *ent)
{
  int err,i;

  printk("<1>"DEVICE_NAME": RORC pci card, device count: %d\n",
                  device_count); 

  i = pci_enable_device (dev);
  if (i)
     return i;

  dv = &prorc[device_count];
  memset(dv, 0, sizeof(*dv));
  dv->pcidev = dev;
  dv->id = device_count;
  pci_read_config_byte(dev, PCI_REVISION_ID, &dv->revision);
  sprintf(dv->name, DEVICE_NAME"%d", dv->id);
  err = configure_device(dv);
  if ( err < 0 )
     return err;
  device_count++;


  return 0;
}

/********************************************/
void __exit rorc_remove_one(struct pci_dev *dev)
{
  printk("<1>"DEVICE_NAME": Bye ...\n"); 
}

/********************************************/
int configure_device(rorc_t *dev)
{
  unsigned long phys, offset;
  caddr_t opreg_page;
  size_t size;

  printk("<1>"DEVICE_NAME": pci vendor: %0x irq: %d rev: %d bar0: %lx bar1: %lx\n",
                            dev->pcidev->vendor, dev->pcidev->irq, 
			    dev->revision, pci_resource_start (dev->pcidev, 0),
			    pci_resource_start (dev->pcidev, 1));

  /* Memory-map the prorc operation registers */
  phys = pci_resource_start (dev->pcidev, 0);

  if ((phys & PCI_BASE_ADDRESS_SPACE))
  {
    printk(KERN_WARNING "Not a memory mapped device.\n");
    return -EIO;
  }
  phys &= PCI_BASE_ADDRESS_MEM_MASK;
  offset = phys & ~PAGE_MASK;
  size = PAGE_ALIGN(phys + sizeof(rorc_regs)) - (phys & PAGE_MASK);
  opreg_page = (caddr_t)ioremap(phys & PAGE_MASK, size);
  if(!opreg_page)
  {
    printk(KERN_WARNING "ioremap failed for BAR0.\n");
    return -EIO;
  }

  if (dev->revision == 1)
  {
    dev->amccreg = (s593x_regs*)((caddr_t)opreg_page + offset);
    printk("<1>"DEVICE_NAME": mapped prorc BAR0 register address: %p\n", 
                                                                dev->amccreg);

    dev->amccreg->mcsr.bit.rte = 0;  // disable bus mastering read
    dev->amccreg->mcsr.bit.wte = 0;  // disable bus mastering write
    dev->pid[1] = -1;                // There is only one channel on this card 
  }
  else
  {
    dev->opreg = (rorc_regs*)((caddr_t)opreg_page + offset);
    printk("<1>"DEVICE_NAME": mapped prorc BAR0 register address: %p\n", 
                                                                dev->opreg);
    if (pci_resource_start (dev->pcidev, 1))
      dev->pid[1] = 0;               // There is channel 1 on this card 
    else
      dev->pid[1] = -1;              // There is NO channel 1 on this card 
  }
  dev->pid[0] = 0;
  pci_set_master(dev->pcidev);       // enable bus mastering */

  return 0;
}


/********************************************/
static int
rorc_open(struct inode* inode, struct file* file)
{
   unsigned int minor = MINOR(inode->i_rdev);

   if (minor >= device_count)
     return -ENXIO;

/*  MOD_INC_USE_COUNT;  Dont allow the module to be removed while in use */
   return 0;
}


/********************************************/
static int
rorc_close(struct inode* inode, struct file* file)
{

/* MOD_DEC_USE_COUNT; */
   return 0;
}


/********************************************/
static int
rorc_mmap(struct file *filp, struct vm_area_struct *vma)
{
 int err;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20))
  err = remap_page_range(vma->vm_start, vma->vm_pgoff*PAGE_SIZE, 
                         vma->vm_end-vma->vm_start, vma->vm_page_prot);
#else
  err = remap_page_range(vma, vma->vm_start, vma->vm_pgoff*PAGE_SIZE, 
                         vma->vm_end-vma->vm_start, vma->vm_page_prot);
#endif
  if (err)
    return -EAGAIN;

  return 0;
}

/********************************************/
static int
rorc_ioctl(struct inode* inode, struct file* file,
             unsigned int cmd, unsigned long arg)
{

   unsigned int minor = MINOR(inode->i_rdev);
   // int size = _IOC_SIZE(cmd);

   int value;
   u32 regaddr;

   rorc_t *dv;

   if (minor >= device_count)
     return -ENXIO;

   dv = &prorc[minor];

/*
   printk("<1>ioctl: minor = %d\n", minor);
   if (dv->revision == 1) return -EINVAL;
     printk("<1>amccreg: %p\n", dv->amccreg);
   else
     printk("<1>opreg: %p\n", dv->opreg);
   printk("<1>command: %x\n", cmd);
*/

   switch (cmd) {
     case RORC_COUNT:
       put_user(device_count, (int*)arg);
       break;
     case RORC_R_VENDOR:              // read Vendor ID
       put_user(dv->pcidev->vendor, (short*)arg);
       break;
     case RORC_R_DEVICE:              // read Device ID
       put_user(dv->pcidev->device, (short*)arg);
       break;
     case RORC_R_IRQ:                 // read IRQ #
       put_user(dv->pcidev->irq, (int*)arg);
       break;
     case RORC_R_BAR0:                // read BAR0
       put_user(pci_resource_start (dv->pcidev, 0) , (u32*)arg);
       break;
     case RORC_R_BAR1:                // read BAR1
       put_user(pci_resource_start (dv->pcidev, 1) , (u32*)arg);
       break;
     case RORC_R_ROM:                 // read ROM
       put_user(pci_resource_start (dv->pcidev, 6) , (u32*)arg);
       break;
    case RORC_R_REVISION:
       put_user(dv->revision , (int*)arg);
       break;
    case RORC_R_DR_MAJ:               // read driver version major number
       put_user(RORC_DRIVER_MAJOR , (int*)arg);
       break;
    case RORC_R_DR_MIN:               // read driver version minor number
       put_user(RORC_DRIVER_MINOR , (int*)arg);
       break;
    case RORC_R_DR_REL:               // read driver version release number
       put_user(RORC_DRIVER_RELEASE , (int*)arg);
       break;
     case RORC_R_PCI_CFG:
       get_user(regaddr, (u32*)arg);
       put_user(pci_resource_start (dv->pcidev, regaddr) , (u32*)arg);
       break;
     case RORC_R_CHAN:                // read channel status
       get_user(value, (unsigned int*)arg);
       put_user(dv->pid[value] , (unsigned int*)arg);
       break;
     case RORC_SET_CHAN:              // set channel status to used
       get_user(value, (unsigned int*)arg);
       dv->pid[value] = current->pid;
       break;
     case RORC_RES_CHAN:              // set channel status to free
       get_user(value, (unsigned int*)arg);
       dv->pid[value] = 0;
       break;
/* pRORC ------------------------------------------------------------*/
     case PRORC_RESET:
       if (dv->revision != 1) return -EINVAL;
       dv->amccreg->mcsr.bit.aor = 1;     /* assert add-on reset */
       udelay (2);
       dv->amccreg->mcsr.bit.a2pfr = 1 ;  /* add-on read FIFO reset */
       dv->amccreg->mcsr.bit.p2afr = 1 ;  /* add-on write FIFO reset */
       dv->amccreg->mcsr.bit.mfr = 1 ;    /* mailbox flags reset */
       dv->amccreg->mcsr.bit.aor = 0;     /* release add-on reset */
       break;
     case PRORC_FF_RESET:
       if (dv->revision != 1) return -EINVAL;
       dv->amccreg->mcsr.bit.a2pfr = 1 ;
       dv->amccreg->mcsr.bit.p2afr = 1 ;
       break;
     case PRORC_MB_RESET:
       if (dv->revision != 1) return -EINVAL;
       dv->amccreg->mcsr.bit.mfr = 1 ;
       break;
     case PRORC_SET_WTE:
       if (dv->revision != 1) return -EINVAL;
       get_user(value, (int*)arg);
       dv->amccreg->mcsr.bit.wte = value & 0x1;
       break;
     case PRORC_SET_WMS:
       if (dv->revision != 1) return -EINVAL;
       get_user(value, (int*)arg);
       dv->amccreg->mcsr.bit.wms = value & 0x1;
       break;
     case PRORC_SET_WVR:
       if (dv->revision != 1) return -EINVAL;
       get_user(value, (int*)arg);
       dv->amccreg->mcsr.bit.wvr = value & 0x1;
       break;
     case PRORC_SET_RTE:
       if (dv->revision != 1) return -EINVAL;
       get_user(value, (int*)arg);
       dv->amccreg->mcsr.bit.rte = value & 0x1;
       break;
     case PRORC_SET_RMS:
       if (dv->revision != 1) return -EINVAL;
       get_user(value, (int*)arg);
       dv->amccreg->mcsr.bit.rms = value & 0x1;
       break;
     case PRORC_SET_RVW:
       if (dv->revision != 1) return -EINVAL;
       get_user(value, (int*)arg);
       dv->amccreg->mcsr.bit.rvw = value & 0x1;
       break;
     default: return -EINVAL;
   }
  return 0;
}


/********************************************/
static int __init rorc_pci_init(void)
{
  int err;
  static int device_major = DEVICE_MAJOR;

  printk(DEVICE_NAME": RORC device driver version %s\n", RORC_DRIVER_VERSION);
  /* Register device major number */
  err = register_chrdev(device_major, DEVICE_NAME, &dev_fops);
  if ( err < 0 )
  {
    printk(KERN_ERR DEVICE_NAME": Unable to get major %d\n", device_major);
    return -ENODEV;
  }
  if (device_major == 0)
  {
    device_major = err;
  }
  printk(DEVICE_NAME": Device major number is %d\n", device_major);

  device_count = 0;
  return pci_module_init (&rorc_driver);
}


/********************************************/
static void __exit rorc_pci_cleanup(void)
{
  static int device_major = DEVICE_MAJOR;
  pci_unregister_driver (&rorc_driver);
  unregister_chrdev(device_major, DEVICE_NAME);
}

/********************************************/
module_init(rorc_pci_init);
/********************************************/
module_exit(rorc_pci_cleanup);
/********************************************/

