/***********************************************************
 * rorc_id.c
 *
 *  Display the RORC's firmware and hardware ids         
 *
 * last updated: 17/03/2006
 * written by: Peter Csato and Ervin Denes
 ***********************************************************/
#define SOURCE_FILE "rorc_id.c"
#define DESCRIPTION "Show RORC's hw and fw ids"

#include <ctype.h>
#include "rorc_ddl.h"

#define MAX_OPTION 4
#define PLD_LEN 7

/*******************************************************/
void usage(char *name)
{
  printf("Display the RORC's firmware and hardware id words.\n");
  printf("usage: %s [-{M|m} <minor>] [-{C|c} <cahnnel>]\n", name);
  printf("               [-V <major version> -v <minor version>\n");
  printf("                -{P|p} <PLD version> -{S|s} <serial#>\n");
  printf("                -{N|n} <channels>]\n");
  printf("               [-{D|d}] [-{T|t} <time-out>]\n");
  printf("   -M <minor>         RORC minor number, default: 0.\n");
  printf("   -C <channel>       RORC channel number, default: 0.\n");
  printf("                      default: 1 for PCI rev 1 or 2, 2 for PCI rev 3.\n");
  printf("   -D                 display DIU and SIU hw and fw id as well.\n");
  printf("   -T <time-out>      time-out value for DIU and SIU id. Default: %d usec.\n",
          DDL_RESPONSE_TIME);
  printf(" The following opitions can be used only for writing hardwer id.\n");
  printf("   -V <major version> RORC cards's major version number.\n");
  printf("   -v <minor version> RORC cards's minor version number.\n");
  printf("   -P <PLD version>   version of PLD on the RORC card.\n");
  printf("   -S <Serial#>       RORC card's serial number.\n");
  printf("   -N <channels>      number of DDL channels,\n");
}

/********************************************************/
int main(int argc, char **argv)
{
  int i, len, ret, c, minor, channel, diu, siu;
  int ustime , time;
  rorc_pci_dev_t prorc;
  rorcHwSerial_t hw;
  unsigned long fw;
  int n1, n2, n3;
  __u8 address;
  __u8 data[DDL_MAX_HW_ID];
  __u8 c_pld[PLD_LEN+1];
  int write_id, version_major, version_minor, numb_chan, serial;
  int was_chan;

  minor = 0;
  channel = 0;
  write_id = 0;
  was_chan = 0;
  numb_chan = 1;
  diu = 0;
  siu = 0;
  ustime = DDL_RESPONSE_TIME;
 
/* Check parameters */

  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:V:v:P:p:N:n:DdS:s:T:t:")) != EOF)
    {
      switch(c)
      {
        case 'm':
        case 'M': minor = atoi(optarg); break;
        case 'c':
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'v': version_minor = atoi(optarg); write_id++; break;
        case 'V': version_major = atoi(optarg); write_id++; break;
        case 'p':
        case 'P':
          if (strlen(optarg) > PLD_LEN)
            optarg[PLD_LEN] = '\0';
          strcpy(c_pld, optarg);
          while (strlen(c_pld) < PLD_LEN)
          {
            strcat(c_pld, " ");
          }
          write_id++;
          break;
        case 's':
        case 'S': serial = atoi(optarg); write_id++; break;

        case 'd':
        case 'D': diu = 1; siu = 1; break;
        case 'n':
        case 'N': numb_chan = atoi(optarg); write_id++; was_chan = 1 ; break;
        case 'T': ustime = atoi(optarg); break;
        case 't':
        default: usage (argv[0]); exit (EXIT_FAILURE); break;
      }
    }
  }

  if (write_id)
  {
    if (write_id < MAX_OPTION)
    {
      printf("Give all or none of the parameters.\n");
      usage(argv[0]);
      exit (EXIT_FAILURE);
    }

    if ((version_major > 9) || (version_minor > 9))
    {
      printf("The version number must be less or equal to 9.\n");
      usage(argv[0]);
      exit (EXIT_FAILURE);
    }

    if ((numb_chan < 1) || (numb_chan > 2))
    {
      printf("The number of channels should be 1 or 2.\n");
      usage(argv[0]);
      exit (EXIT_FAILURE);
    }

    if (serial > 99999)
    {
      printf("The serial number must be less or equal to 99999.\n");
      usage(argv[0]);
      exit (EXIT_FAILURE);
    }
  }

  /* sw version */
  printf("Compiled RORC sw version: %s\n", RORC_DRIVER_VERSION);
  
  /* open RORC and map mailboxes */
  if (rorcOpenChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);
  time = ustime * prorc.loop_per_usec;

  /* driver version */
  printf("RORC driver version: %d.%d.%d\n", 
        prorc.driver_major, prorc.driver_minor, prorc.driver_release);
  printf("RORC minor = %d\n", minor);

  /* pRorc revision id */
  if (prorc.rorc_revision > 0)
    printf("RORC revision id: %d\n", prorc.rorc_revision);

  if (pRorc(&prorc))
  {
    if (write_id)
      printf("RORC version: pRORC. I can write HW id only for a D-RORC.\n\n");

    /* read non volatile ram */
    hw = rorcSerial(&prorc);
    rorcInterpretSerial(hw);

    /* empty incoming mb1, 2, and 3 if necessary */
    n1 = pRorcEmptyMb(&prorc, 1, 0);
    n2 = pRorcEmptyMb(&prorc, 2, 0);
    n3 = pRorcEmptyMb(&prorc, 3, 0);
    if (n1 + n2 + n3)
      printf ("Before sending pRORC command\n");
    if (n1)
      printf ("     IMB1 contained %d status.\n", n1);
    if (n2)
      printf ("     IMB2 contained %d status.\n", n2);
    if (n3)
      printf ("     IMB3 contained %d status.\n", n3);
  }
  else
  {
    if (write_id)
    {
      if ((prorc.rorc_revision >= INTEG) && !was_chan)
        numb_chan = 2;

      /* build HW id */
      rorcBuildHwSerial(data, prorc.rorc_revision, version_major, version_minor,
                              c_pld, numb_chan, serial);
      printf("RORC HW id to write: '%s'\n", data);
    
      /* write hw id using the i2c bus */
      len = strlen(data);
      if (len > DDL_MAX_HW_ID - 1)
        len = DDL_MAX_HW_ID - 1;

      address = 0x0;
      for (i = 0; i < len+1; i++, address++)
      ret = i2c_write_a_byte(&prorc, address, data[i], MAX_WAIT);
      if (ret == RORC_TIMEOUT)
      {
        printf("No reply from D-RORC. HW id can not be written.\n");
        rorcClose(&prorc);
        exit (EXIT_FAILURE);
      }
      printf("D-RORC HW id witten.\n");
    }

    /* read hw id from I2C bus */
    hw = rorcSerial(&prorc);
    rorcInterpretSerial(hw);
  }

  /* read fw id */
  fw = rorcReadFw (&prorc);
  rorcInterpretFw(fw);

  if (rorcCheckVersion(&prorc))
    printf("RORC fw version is NOT compatible with the present sw.\n");
  else
    printf("RORC fw version is compatible with the present sw.\n");

  if ((diu || siu) && (prorc.diu_version == NO_DIU))
  {
    printf("No DIU plugged into the RORC channel %d\n", prorc.ddl_channel);
    rorcClose(&prorc);
    exit (0);
  }

  if (diu || siu)
    printf("RORC channel %d:\n", channel);

  if (diu)
  {
    hw = ddlSerial(&prorc, DIU, time);
    if (strlen(hw.data))
      printf("DIU HW id read: '%s'\n", hw.data);

    fw = ddlReadFW(&prorc, DIU, time);
    if (ST_CODE(fw) == FWSTW)
      ddlInterpretFW(fw);
  }

  if (siu)
  {
    hw = ddlSerial(&prorc, SIU, time);
    if (strlen(hw.data))
      printf("SIU HW id read: '%s'\n", hw.data);

    fw = ddlReadFW(&prorc, SIU, time);
    if (ST_CODE(fw) == FWSTW)
      ddlInterpretFW(fw);
  }

  /* unmap and close RORC */
  rorcClose(&prorc);

  exit (0);
}
