/****************************************************
 * rorc_pop_fifo.c
 *
 *  Testing the Free FIFO         
 *
 * last updated: 22/08/2003
 * written by: Peter Csato and Ervin Denes
 ****************************************************/
#define SOURCE_FILE "rorc_pop_fifo.c"
#define DESCRIPTION "Test of Free FIFO"

#include <ctype.h>
#include "rorc_lib.h"

#define MAX_LOOP  128
#define MAX_INDEX 255
#define MAX_BLOCK 1024

/*******************************************************/

void usage(char *name)
{
  printf("Pop block address, length and Ready FIFO index from the Free FIFO.\n");
  printf("usage: %s [-{M|m} <minor>] [-{C|c} <channel>]\n", name);
  printf("                [-{A|a} <address>] [-{L|l} <length>] [-{I|i} <index>]\n");
  printf("                [-{P|p} <pattern>] [-{R|r} <repeat>] [-{T|t}] [-{N|n}]\n");
  printf(" <minor>	card number >= 0,	default = 0\n");
  printf(" <channel>	channel number >= 0,	default = 0\n");
  printf(" <address>	block address,		default = physmem address\n");
  printf(" <length>	block size in words,	default = %d\n", MAX_BLOCK);
  printf(" <index>	Ready FIFO index<=%d,	deafult = 0\n", MAX_INDEX);
  printf(" <pattern>	i=inc, c=const,		deafult = i\n");
  printf(" <repeat>	number of pops<=%d,	deafult = 0\n", MAX_LOOP);
  printf("		0 means pop until Ready FIFO is empty\n");
  printf(" -T or -t	pop Tx Address FIFO,    default = Rx Address FIFO\n");

  printf(" -N		no print, check data    default = print, no check\n");
}

/********************************************************/

int main(int argc, char **argv)
{
  int pattern, max_loop, no_check, comp_err;
  int tx;
  int c, loop, p_minor, channel, def_addr;
  rorc_pci_dev_t prorc;
  __u32 address, blockAddress[MAX_LOOP];
  int length, blockLength[MAX_LOOP];
  int index, readyFifoIndex[MAX_LOOP];
  int status;
  unsigned long phys_addr;
  volatile unsigned long *addr_user;

  int physmem_fd;
  unsigned long phys_size_physmem;

  /* default parameters */
  p_minor = 0;
  channel = 0;
  address = 0x0;
  tx = 0;
  def_addr = 0;
  length = MAX_BLOCK;
  index = 0;
  pattern = 'i';
  max_loop = 0;
  no_check = 1;

  /* Check parameters */
  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:A:a:L:l:I:i:P:p:R:r:T:t:N:n")) != EOF)
    {
      switch(toupper(c))
      {
        case 'M': p_minor = atoi(optarg); if (p_minor < 0) p_minor = 0; break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'A': address = strtoul(optarg, (char**)NULL, 16); def_addr = 1; break;
        case 'L': length = atoi(optarg) & 0x00ffffff; break;
        case 'I':
          index = atoi(optarg);
          if (index < 0) index = 0;
          if (index > MAX_INDEX) index = MAX_INDEX;
          break;
        case 'P': pattern = tolower(*optarg); break;
        case 'R':
          max_loop = atoi(optarg); 
          if (max_loop < 0) max_loop = 0;
          if (max_loop > MAX_LOOP) max_loop = MAX_LOOP;
          break;
	case 'T': tx = 1; break;
        case 'N': no_check = 0; break;
        default : usage (argv[0]); return -1; break;
      }
    }
  }

  if (max_loop == 0)
  {
    max_loop = MAX_LOOP;
  }

  if ((pattern != 'i') && (pattern != 'c'))
    pattern = 'i';

  if (physmemOpen (&physmem_fd, &addr_user, &phys_addr,
                   &phys_size_physmem))
    exit (EXIT_FAILURE);

  printf(" Allocation of %ld words of physmem memory.\n",
           phys_size_physmem >> 2);
  printf(" Physical address: 0x%08lx.\n", phys_addr);

  if (!def_addr)
    address = phys_addr;

  if (!tx)
    printf(" Poping maximum %d data from the Rx Free FIFO of pRORC%d.\n",
		               max_loop, p_minor);
  else
    printf(" Poping maximum %d data form the Tx Free FIFO of pRORC%d.\n",
		               max_loop, p_minor);
  
  if (no_check)
  {
    printf("\nPopped address, length and Free FIFO index:\n\n");
  }
  else
  {
    printf(" Block address: 0x%08x, block length: %d, Free FIFO index: %d.\n",
             address, length, index);
    printf(" Checking address and index according to %c pattern.\n", pattern);
    blockAddress[0]   = address;
    blockLength[0]    = length;
    readyFifoIndex[0] = index;
    for (loop = 1; loop < max_loop; loop++)
    {
      switch(pattern)
      {
        case 'i':
          blockAddress[loop]   = blockAddress[loop-1] + (blockLength[loop-1] << 2);
          blockLength[loop]    = blockLength[loop-1];
          readyFifoIndex[loop] = (readyFifoIndex[loop-1] + 1) & MAX_INDEX;
          break;
        case 'c':
          blockAddress[loop]   = blockAddress[0];
          blockLength[loop]    = blockLength[0];
          readyFifoIndex[loop] = readyFifoIndex[0];
          break;
       }
     }
   }

  /* open RORC and map mailboxes */
  if (rorcMapChannel(&prorc, p_minor, channel))
    exit (EXIT_FAILURE);

  /* start cycles */
  comp_err = 0;
  for (loop = 0; loop < max_loop; loop++) 
  {
    status = rorcPopFreeFifo (&prorc, tx, &address, &length, &index);
    if (no_check)
    {
      if (status == RORC_FF_EMPTY)
      {
        printf(" Free FIFO is empty.\n");
        break;
      }
      printf("%3d: 0x%08x %d %d\n", loop+1, address, length, index);
    }
    else
    {
      if (status == RORC_FF_EMPTY)
      {
        printf("Free FIFO is empty after %d loop(s).\n", loop);
        break;
      }
      if (blockAddress[loop]   != address ||
          blockLength[loop]    != length ||
          readyFifoIndex[loop] != index)
      {
        printf("\nDATA COMPARE ERROR! ");
        printf("%3d: expected = 0x%08x, %d, %d; received = 0x%08x, %d, %d.\n", 
          loop+1, blockAddress[loop],blockLength[loop],readyFifoIndex[loop],
                  address, length, index);
        comp_err = 1;
        break;
      }
    }
  }
  printf("%d record(s) read.\n", loop);
  if (!no_check && !comp_err)
    printf("\nData check OK\n");

  /* unmap and close RORC */
  rorcClose(&prorc);

  /* unmap and close physmem memory */
  if (physmemClose (physmem_fd, (unsigned long *)addr_user, phys_size_physmem))
    exit (EXIT_FAILURE);

  return 0;
}

