/****************************************************
 * rorc_push_fifo.c
 *
 *  Testing the Free FIFO         
 *
 * last updated: 22/08/2003
 * written by: Peter Csato and Ervin Denes
 ****************************************************/
#define SOURCE_FILE "rorc_push_fifo.c"
#define DESCRIPTION "Test of Free FIFO"

#include <ctype.h>
#include "rorc_ddl.h"

#define MAX_LOOP  128
#define MAX_BLOCK 1024
#define MAX_INDEX 255

/*******************************************************/

void usage(char *name)
{
  printf("Push block address, length and Ready FIFO index into the Free FIFO.\n");
  printf("usage: %s [-{M|m} <RORC minor number>] [-{C|c} <channel>]\n", name);
  printf("                 [-{A|a} <address>] [-{L|l} <length>] [-{I|i} <index>]\n");
  printf("                 [-{P|p} <pattern>] [-{R|r} <repeat>] [-{T|t}]\n");
  printf(" <RORC minor number>	card number >= 0,	default = 0\n");
  printf(" <channel>          	channel number >= 0,	default = 0\n");
  printf(" <address>		block address,		default = physmem address\n");
  printf(" <length>		block size in words,	default = %d\n", MAX_BLOCK);
  printf(" <index>		Ready FIFO index<=%d,	deafult = 0\n", MAX_INDEX);
  printf(" <pattern>		i=inc, c=const,		deafult = i\n");
  printf(" <repeat>		number of pushes<=%d,	deafult = 1\n", MAX_LOOP);
  printf(" -T or -t		push Tx Address FIFO,   default = Rx Address FIFO\n");  
}

/********************************************************/

int main(int argc, char **argv)
{
  int pattern, max_loop, tx;
  int c, loop, p_minor, channel, def_addr;
  rorc_pci_dev_t prorc;
  __u32 address, blockAddress[MAX_LOOP];
  int length, blockLength[MAX_LOOP];
  int index, readyFifoIndex[MAX_LOOP];

  unsigned long phys_addr;
  volatile unsigned long *addr_user;

  int physmem_fd;
  unsigned long phys_size_physmem;

  /* default parameters */
  p_minor = 0;
  channel = 0;
  address = 0x0;
  def_addr = 0;
  length = MAX_BLOCK;
  index = 0;
  pattern = 'i';
  max_loop = 1;
  tx = 0;

  /* Check parameters */
  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:A:a:L:l:I:i:P:p:R:r:Tt")) != EOF)
    {
      switch(toupper(c))
      {
        case 'M': p_minor = atoi(optarg); if (p_minor < 0) p_minor = 0; break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'A': 
          address = strtoul(optarg, (char**)NULL, 16); 
          def_addr = 1; 
          break;
        case 'L': length = atoi(optarg) & 0x00ffffff; break;
        case 'I': 
          index = atoi(optarg);
          if (index < 0) index = 0;
          if (index > MAX_INDEX) index = MAX_INDEX;
          break;
        case 'P': pattern = tolower(*optarg); break;
        case 'R': 
          max_loop = atoi(optarg); 
          if (max_loop > MAX_LOOP) max_loop = MAX_LOOP;
          break;
	case 'T': tx = 1; break;
        default : usage (argv[0]); exit (EXIT_FAILURE); break;
      }
    }
  }
  if ((pattern != 'i') && (pattern != 'c'))
    pattern = 'c';

  if (physmemOpen (&physmem_fd, &addr_user, &phys_addr,
                   &phys_size_physmem))
    exit (EXIT_FAILURE);

  printf(" Allocation of %ld words of physmem memory.\n",
           phys_size_physmem >> 2);
  printf(" Physical address: 0x%08lx.\n", phys_addr);

  if (!def_addr)
    address = phys_addr;

  blockAddress[0]   = address;
  blockLength[0]    = length;
  readyFifoIndex[0] = index;
  for (loop = 1; loop < max_loop; loop++)
  {
    switch(pattern)
    {
      case 'i':
        blockAddress[loop]   = blockAddress[loop-1] + (blockLength[loop-1] << 2);
        blockLength[loop]    = blockLength[loop-1];
        readyFifoIndex[loop] = (readyFifoIndex[loop-1] + 1) & MAX_INDEX;
        break;
      case 'c':
        blockAddress[loop]   = address;
        blockLength[loop]    = length;
        readyFifoIndex[loop] = index;
        break;
     }
   }

  /* open RORC and map mailboxes */
  if (rorcMapChannel(&prorc, p_minor, channel))
    exit (EXIT_FAILURE);

  if (!tx)
    printf(" Pushing %d data into the Rx Free FIFO of D-RORC%d.\n", 
             max_loop, p_minor);
  else
  {
    if (pRorc(&prorc))
    {
      printf("Tx Free FIFO can not be pushed for RORC card revision 1.\n");
      usage (argv[0]);
      rorcClose(&prorc);
      exit (EXIT_FAILURE);
    }
    printf(" Pushing %d data into the Tx Free FIFO of D-RORC%d.\n", 
             max_loop, p_minor);
  }
  printf(" Block address: 0x%08x, block length: %d, Ready FIFO index: %d.\n",
           address, length, index);
  if (max_loop > 1)
    printf(" Changing address and index according to %c pattern.\n", pattern);

  /* start cycles */
  for (loop = 0; loop < max_loop; loop++)
  {
/*
   printf("%3d: 0x%08x %d %d\n", loop+1, blockAddress[loop],blockLength[loop],
                                         readyFifoIndex[loop]);
*/
    if (!tx)
    {
      rorcPushFreeFifo (&prorc,
                         blockAddress[loop],
                         blockLength[loop],
                         readyFifoIndex[loop]);
    }
    else
    { 
      dRorcPushTxFreeFifo (&prorc,
                           blockAddress[loop],
                           blockLength[loop],
                           readyFifoIndex[loop]);
    }
  }

  /* unmap and close RORC */
  rorcClose(&prorc);

  /* unmap and close physmem memory */
  if (physmemClose (physmem_fd, (unsigned long *)addr_user, phys_size_physmem))
    exit (EXIT_FAILURE);

  return 0;
}

