/***********************************************************
 * rorc_receive.c                                          *
 * Testing RORC's Data Receiver                            *
 * last updated: 20/06/2006                                *
 * written by: Peter Csato and Ervin Denes                 *
 *                                                         *
 * history:                                                *
 *                                                         *
 * v.5.1.4  20-Aug-2006: flushing stdout before looping    *
 * v.4.4.7  11-Jul-2006: fixing error in reset sequence    *
 * v.4.4.6  26-Jun-2006: print block number on output file *
 * v.4.4.5  09-Mar-2006: sleep after loop-back settings    *
 *          16-Feb-2006: Check resp_time option            *
 * v.4.4.4  21-Dec-2005: Clear SIU/DIU status after reset  *
 *                       Print sw version at the benning   *
 *                       Call ddlSetSiuLoopback for -g2    *
 *                       Check DTSTW error bit             *
 * v.4.4.3  10-May-2005: add "ERROR" text to error messages*
 * v.4.3.11 24-Jan-2005: add option -w for response time   *
 * v.4.3.11 20-Jan-2005: accepts blocks with DDL header    *
 * v.4.3.11 14-Dec-2004: use usec for working indicator    *
 *                       and print page or GB moved        * 
 *                       together with the indicator       *
 * v.4.3.5  07-Aug-2004: use usec time-outs                *
 * v.4.3.1  29-Jun-2004: event length can me more than 2MB *
 * v.4.2    20-Apr-2004: fixing bug of writing to file     *
 * v.4.2    20-Oct-2003: add pattern file in mif format    *
 * v.4.2    17-Oct-2003: add option -r, seed became -j     *
 * v.4.2    01-Sep-2003: page size must be multiple of 8   *
 * v.4.2    27-Aug-2003: add loopbacks                     *
 * v.4.2    23-Aug-2003: add channel variable              *
 * v.4.2    11-Aug-2003: add usleep after reset            *
 *                                                         *
 **********************************************************/
#define SOURCE_FILE "rorc_receiver.c"
#define DESCRIPTION "Test of RORC's Data Generator and Data Receiver"

// #define LINE_UP "[1A[0J"
#define LINE_UP "[1A[80D[0J"


#include <signal.h>
#include <getopt.h>
#include "rorc_ddl.h"

#define printf(message...) (fprintf (stdout, message), fflush(stdout))

#define MAX_MAPPING   (1024*1024*100)

#define MAX_LEN       (1024*512-1) /* 2^19-1 */
#define PAGE_LEN      (1024*90*12)
#define MIN_BLOCK_LEN (64*4)
#define MAX_BLOCK     (MAX_MAPPING/MIN_BLOCK_LEN + 1)

#define RF_GAP          1    // every entry used in RF
#define MAX_RF_INDEX    (128 * RF_GAP - 1) 
#define MAX_MAPPING_RF  ((MAX_RF_INDEX + 1) * 8)

#define MAX_FIFO 80
#define MIN_FIFO MAX_FIFO - 1
#define SLEEP_TIME 0
#define GB_PRINT 1
#define CHECK_LEVEL 3

typedef struct
{  int                      rf;
   volatile unsigned long   *addr;
} memoryBlock_t;

/* global variables */ 

static struct sigaction sigac;
// int interrupt_arrived = 0;

int block_len, block_len_word, max_block;
int rf_free_index, next_free_index, free_block;
memoryBlock_t block[MAX_BLOCK];
rorcReadyFifo_t *rf;
volatile unsigned long *addr_user, *next_free_block;   
unsigned long phys_addr, next_free_block_phys;
unsigned long max_buffer_size, max_mapping, mem_offset;
unsigned long fee_address;
int read_back;

int max_len, ustimeout, timeout, random_seed;
int generator, check_level, data_check, counter_check, ddl_header; 
int scatter, loopback, send_trig, reset_level;
int kell_diu, kell_siu;
int pattern, pattern_no, max_event;
int max_fifo, min_fifo, sleep_time;
int p_minor, channel;
int init_count, user_init_count, gb_print, page_print;
unsigned long resp_wait, resp_cycle;
unsigned long init_word;
unsigned long *expected, *full_event;
char *stat_file, *output_file, *pattern_file;
char stat_file_name[80], output_file_name[80], pattern_file_name[80];
char indicator[4] = { '/', '-', '\\', '|' };
int file_counter=0;
char filename[100];
/*******************************************************/

void usage(char *name)
{
  printf("usage: %s\n", name);
  printf("       [{-M|-m|--minor}       <RORC's minor number>|0]\n");
  printf("       [{-C|-c|--channel}     <RORC's channel number>|0]\n");
  printf("       [-v|--verbose]\n");
  printf("       [-G|-g|--generator     <loopback mode>|0]\n");
  printf("       [-D|-d|--no_scatter]\n");
  printf("       [{-R|-r|--reset_lev}   <reset level>|3]\n");
  printf("       [{-X|-x|--check}       <check level>|%d]\n", CHECK_LEVEL);
  printf("       [-Y|-y|--DDL_header]\n");
  printf("       [-Z|-z|--no_RDYRX]\n");
  printf("       [{-B|-b|--page}        <block (page) length in bytes>|%d]\n", 
                              PAGE_LEN);
  printf("       [{-U|-u|--physmem}     <utilisable physmem memory in MBs>|%d]\n",
                              MAX_MAPPING/1024/1024);
  printf("       [{-O|-o|--offset}      <offset of utilisable memory in MBs>|0]\n");
  printf("       [{-E|-e|--events}      <events>|0]\n");
  printf("       [{-I|-i|--init_word}   <init word>|0]\n");
  printf("       [{-P|-p|--pattern}     {c|a|0|1|I|d|<mif file name>}]\n");
  printf("       [{-S|-s|--stat_file}   <stat_file>]\n");
  printf("       [{-L|-l|--length}      <data length in words>|%d]\n", MAX_LEN);
  printf("       [{-N|-n|--init_count}  <initial event count>|1]\n");
  printf("       [{-J|-j|--rand_len]    <random seed>|0]\n");
  printf("       [{-F|--max_fifo}       <max FIFO>|%d]\n", MAX_FIFO);
  printf("       [{-f|--min_fifo}       <min FIFO>|%d]\n", MIN_FIFO);
  printf("       [{-T|-t|--sleep_time}  <sleep time between events in ms>|%d]\n",
		                                          SLEEP_TIME);
  printf("       [{-W|-w|--resp_wait}   <wait time for command responses in us>|%d]\n", 
		                                          DDL_RESPONSE_TIME);
  printf("       [{-Q|--byte_print}     <# of GBs to print>|%d]\n", GB_PRINT);
  printf("       [{-q|--page_print}     <# of pages to print>|0]\n");
  printf("       [{-K|-k|--output_file} <output file name>]\n");
  printf("       [{-A|-a|--front_end}   <FEE address>]\n");
  printf(" -A means: send STBRD before, EOBTR after each event\n");
  printf(" -G means: use Rorc's data generator and set loopback\n");
  printf("           0 = no loopback, 1 = DIU loopback, 2 = SIU loopback\n");
  printf("           any other value = RORC loopback\n");
  printf(" -D means: do not scatter the received data\n");
  printf(" -R 0 means: do not reset RORC, neither DIU nor SIU\n");
  printf(" -R 1 means: reset RORC only\n");
  printf(" -R 2 means: reset RORC and DIU, do not reset SIU\n");
  printf(" -R 3 means: reset RORC, DIU and SIU before collecting data\n");
  printf(" -X 0 means: do not check the event\n");
  printf(" -X 1 means: check only the first word of events\n");
  printf(" -X 2 means: do not check the first word of events\n");
  printf(" -X 3 means: check the whole event\n");
  printf(" -Y means: the received blocks have DDL header (not checked)\n");
  printf(" -Z means: do not send RDYRX/EOBTR commands (default in case of -G)\n");
}

/********************************************************
*          signal handler                               *
********************************************************/
void sig_hand(int signum)
{

/* signum: 1(SIGHUP)  2(SIGINT)  3(SIGQUIT) */
  if(signum < 4)
  {
    interrupt_arrived = 1;
  }
}

/********************************************************
*          signal handling                              *
********************************************************/
void set_sig_hand(void)
{
  sigac.sa_flags = 0;
  sigac.sa_handler =  sig_hand;
  if (sigaction(SIGHUP, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGHUP ");
     exit(1);
  }
  if (sigaction(SIGINT, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGINT ");
     exit(1);
  }
  if (sigaction(SIGQUIT, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGQUIT ");
     exit(1);
  }
}

/********************************************************
*         fill block                                    *
********************************************************/
void fill_block(volatile unsigned long *addr, int len)
{
  int i;

  debug("FILL: user_addr=%8p, len=%d\n", addr, len);
  for (i = 0; i < len; i++)
    *(addr + i) = (unsigned long) (addr + i);
}

/********************************************************
*         next_push                                    *
********************************************************/
void next_push(rorcDescriptor_t *prorc_dev, int data_check)
{
  rf[rf_free_index].length = -1;
  rf[rf_free_index].status = -1;
//  if (data_check)
//    fill_block(next_free_block, block_len_word);
  rorcPushFreeFifo (prorc_dev, next_free_block_phys, 
                             block_len_word, rf_free_index);
  debug("PUSH: phys=0x%08lx, user=%8p, index=%d\n", next_free_block_phys,
                               next_free_block, rf_free_index);

  free_block--;
  next_free_index++;
  if (next_free_index >= max_block)
    next_free_index = 0;
  rf_free_index += RF_GAP;
  if (rf_free_index > MAX_RF_INDEX)
    rf_free_index = 0;
  next_free_block_phys += block_len;
  next_free_block += block_len_word;
  if (!scatter || 
      ((next_free_block_phys + block_len - phys_addr - 1) > max_buffer_size))
  {
    next_free_block_phys = phys_addr;
    next_free_block = addr_user;
  }

  block[next_free_index].rf = rf_free_index;
  block[next_free_index].addr = next_free_block;

  debug("PUSH: next phys=0x%08lx, user=%8p, index=%d into block[%d]\n", 
      next_free_block_phys, next_free_block, rf_free_index, next_free_index);
}

/********************************************************
*         treat program parameters                      *
********************************************************/
void treat_param(int argc, char **argv)
{
  int c;
  int option_index = 0;
  char *short_options =
    "vM:m:C:c:B:b:U:u:O:o:P:p:E:e:G:g:DdR:r:X:x:ZzI:i:N:n:L:l:J:j:S:s:F:f:T:t:W:w:Q:q:K:k:A:a:Yy";

   const struct option long_options[]=
   {
     {"verbose",   0, 0, 'v'}, {"minor",     1, 0, 'm'}, {"channel",   1, 0, 'c'},
     {"page",      1, 0, 'b'}, {"physmem",   1, 0, 'u'}, {"offset",    1, 0, 'o'},
     {"pattern",   1, 0, 'p'}, {"events",    1, 0, 'e'}, {"generator", 1, 0, 'g'},
     {"no_scatter",0, 0, 'd'}, {"check",     1, 0, 'x'}, {"no_RDYRX",  0, 0, 'z'},
     {"init_word", 1, 0, 'i'}, {"init_count",1, 0, 'n'}, {"reset_lev", 1, 0, 'r'},
     {"length",    1, 0, 'l'}, {"rand_len",  1, 0, 'j'},
     {"stat_file", 1, 0, 's'}, {"sleep_time",1, 0, 't'},
     {"max_fifo",  1, 0, 'F'}, {"min_fifo",  1, 0, 'f'},
     {"resp_wait", 1, 0, 'w'}, {"DDL_header",0, 0, 'y'},
     {"byte_print",1, 0, 'Q'}, {"page_print",1, 0, 'q'},
     {"output_file",1, 0,'k'}, {"front_end", 1, 0, 'a'},
     {0, 0, 0, 0}
   };

  /* default parameters */

  p_minor = 0;
  channel = 0;
  user_init_count = 1;
  init_word = 0;
  max_len = MAX_LEN;
  random_seed = RORC_DG_NO_RANDOM_LEN;
  pattern = 'i';
  block_len = PAGE_LEN;
  mem_offset = 0;
  max_event = RORC_DG_INFINIT_EVENT;
  generator = 0;
  scatter = 1;
  reset_level = 3;
  loopback = 0;
  check_level = CHECK_LEVEL;
  ddl_header = 0;
  send_trig = 1;
  fee_address = 0;
  read_back = 0;
  ustimeout = DDL_RESPONSE_TIME;
  stat_file = (char*) NULL;
  output_file = (char*) NULL;
  pattern_file = (char*) NULL;
  max_fifo = MAX_FIFO;
  min_fifo = MIN_FIFO;
  sleep_time = SLEEP_TIME;
  resp_wait = DDL_RESPONSE_TIME;
  gb_print = GB_PRINT;
  page_print = 0;

  /* Check parameters */

  if (argc > 1)
  {
    while ((c = getopt_long(argc, argv, short_options,
                                        long_options, &option_index)) != EOF)
    {
      switch (c)
      {
        case 'v': break;
        case 'M':
        case 'm': p_minor = atoi(optarg); break;
        case 'C':
        case 'c': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf("Error: Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'B':
        case 'b': block_len = atoi(optarg);
          if (block_len < MIN_BLOCK_LEN)
          {
            printf("Error: minimum block length is %d bytes.\n", MIN_BLOCK_LEN);
            exit (EXIT_FAILURE);
          }
          block_len = (block_len >> 3) << 3;  // align to 8 bytes
          break;
        case 'U':
        case 'u': max_mapping = atoi(optarg)*1024*1024; break;
        case 'O':
        case 'o': mem_offset = atoi(optarg)*1024*1024; break;
        case 'P':
        case 'p': strcpy(pattern_file_name, optarg);  break;
        case 'E':
        case 'e': max_event = atoi(optarg); break;
        case 'G':
        case 'g': generator = 1; loopback = atoi(optarg); break;
        case 'D':
        case 'd': scatter = 0; break;
        case 'R':
        case 'r': reset_level = atoi(optarg); break;
        case 'X': 
        case 'x': check_level = atoi(optarg); break;
        case 'Y':
	case 'y': ddl_header = DDL_HEADER_SIZE_WORD; break;
        case 'Z':
        case 'z': send_trig = 0; break;
        case 'I':
        case 'i': init_word = strtoul(optarg, (char**)NULL, 16); break;
        case 'N':
        case 'n': user_init_count = atoi(optarg); break;
        case 'l': 
        case 'L': max_len = atoi(optarg);
          if (max_len < 1 || max_len + 1 > MAX_LEN)
          {
            printf("Error: data length must be positive and less than %d word.\n",
                    MAX_LEN);
            exit (EXIT_FAILURE);
          }
          break;
        case 'J':
        case 'j': random_seed = strtoul(optarg, (char**)NULL, 16); break;
        case 'S':
        case 's': strcpy(stat_file_name, optarg); stat_file = stat_file_name; break;
        case 'A':
        case 'a': fee_address = strtoul(optarg, (char**)NULL, 16); read_back = 1; break;
        case 'K':
        case 'k': strcpy(output_file_name, optarg); output_file = output_file_name; break;
        case 'F': max_fifo = atoi(optarg); 
          if (max_fifo < 1 || max_fifo > MAX_FIFO)
          {
            printf ("Error: max fifo must be positive and less or equal to %d.\n", 
                     MAX_FIFO);
            exit (EXIT_FAILURE);
          }
          break;
        case 'f': min_fifo = atoi(optarg); break;
        case 'T':
        case 't': sleep_time = atoi(optarg); break;
        case 'W':
        case 'w': resp_wait = atoi(optarg); break;
        case 'Q': gb_print = atoi(optarg); break;
        case 'q': page_print = atoi(optarg); break;
        default : usage (argv[0]); exit (EXIT_FAILURE); break;
      }
    }
  }

  init_count = user_init_count;

  block_len_word = block_len/4;

  if (strlen(pattern_file_name) > 1)
  {
    pattern_file = pattern_file_name;
    pattern = 'm'; 
  }
  else
    pattern = tolower(pattern_file_name[0]);

  switch (pattern)
  {
    case 'm': pattern_no = 0; break;
    case 'c': pattern_no = 1; break;
    case 'a': pattern_no = 2; break;
    case '0': pattern_no = 3; init_word = -2; break;
    case '1': pattern_no = 4; init_word = 1 ; break;
    case 'i': pattern_no = 5; break;
    case 'd': pattern_no = 6; break;
    default : pattern = 'i' ; pattern_no = 5; break;
  }

  if (output_file != NULL)
    check_level = 0;

  switch (check_level)
  {
    case 0: counter_check = 0; data_check = 0; break;
    case 1: counter_check = 1; data_check = 0; break;
    case 2: counter_check = 0; data_check = 1; break;
    case 3: 
    default: counter_check = 1; data_check = 1; break;
  }

  if (generator)
  {
    send_trig = 0;
    if ((loopback < 0) || (loopback > 3))
      loopback = 3;

    if (pattern_file != NULL)
    {
      printf("Error: data generator can not use mif file.\n");
      printf("Options -G and -P <pattern file> are mutually exclusive.\n");
      exit (EXIT_FAILURE);
    }
  }
  
  if (read_back)
  {
    send_trig = 0;
    counter_check = 0;
  }
}

/********************************************************/
void fill_expected(int continuation)
{
  int i, starti;


  if (!continuation)
  {
    expected[0] = init_count;
    init_count++;
    expected[1] = init_word;
    starti = 2;
  }
  else
  {
    switch(pattern)
    {
      case 'c': expected[0] = expected[max_len-1]; break;
      case 'a': expected[0] = ~expected[max_len-1]; break;
      case '1':
        expected[0] = expected[max_len-1] << 1;
        if (expected[0] == 0)
          expected[0] = init_word;
        break;
      case '0':
        expected[0] = (expected[max_len-1] << 1) | 1;
        if (expected[0] == -1)
          expected[0] = init_word;
        break;
      case 'i': expected[0] = expected[max_len-1] + 1; break;
      case 'd': expected[0] = expected[max_len-1] - 1; break;
    }
    starti = 1;
  }

  for (i = starti; i < max_len; i++)
  {
    switch(pattern)
    {
      case 'c': expected[i] = expected[i-1]; break;
      case 'a': expected[i] = ~expected[i-1]; break;
      case '1':
        expected[i] = expected[i-1] << 1;
        if (expected[i] == 0)
          expected[i] = init_word;
        break;
      case '0':
        expected[i] = (expected[i-1] << 1) | 1;
        if (expected[i] == -1)
          expected[i] = init_word;
        break;
      case 'i': expected[i] = expected[i-1] + 1; break;
      case 'd': expected[i] = expected[i-1] - 1; break;
    }
  }

}

/********************************************************/
int main(int argc, char **argv)
{
	int eventcounter;
  int open_transaction;
  int data_len, rounded_len, stopped;
  int last_data_len, sum_data_len, dtsw_len; 
  int i, j, istart, ret, loop, event, comp_err, st, he;
  int step_indicator;
  int read_word;
  int continuation, n_ddl_block;
  rorcDescriptor_t prorc;
  rorcHwSerial_t hw_ser;
  unsigned long status, dtsw; 
  stword_t stw;
  __u32 fw;
  int rf_index, data_arrived, status_arrived;
  int next_index, n_read, fwvers_major, fwvers_minor;
  volatile unsigned long *buffer_address;
  unsigned long *addr_user_data;
  volatile unsigned long *addr_user_rf;
  unsigned long phys_addr_rf, phys_addr_data, phys_addr_end;
  unsigned long phys_size_physmem;
  int sum_byte=0, sum_Gbyte=0;
  struct timeval start_tv, now_tv;
  time_t start_time;
  int physmem_fd;
  int output_open;
  FILE *ofile = (FILE*) NULL;
  FILE *patf = (FILE*) NULL;
eventcounter=0;
  /* sw version */
  printf("Starting rorc_receive program.\n");
  for (i = 0; i < argc; i++)
    printf("%s ", argv[i]);
  printf("\nCompiled RORC sw version: %s.\n", RORC_DRIVER_VERSION);

  /* physmem */

  if (physmemOpen (&physmem_fd, &addr_user, &phys_addr,
                   &phys_size_physmem))
    exit (EXIT_FAILURE);

  printf(" Allocation of the physmem memory.\n");
  printf(" Physical address: 0x%08lx.\n", phys_addr);
  printf(" Size of the physmem memory: %ld (0x%08lx) bytes, %ld (0x%08lx) words.\n",
           phys_size_physmem, phys_size_physmem, 
           phys_size_physmem >> 2, phys_size_physmem >> 2);

  /* treat parameters */
  max_mapping = MAX_MAPPING;
  if (phys_size_physmem < MAX_MAPPING)
    max_mapping = phys_size_physmem;
  treat_param(argc, argv);
  if (!scatter)
    max_fifo = 1;
  if (min_fifo >= max_fifo)
    min_fifo = max_fifo - 1;

  if ((mem_offset + max_mapping) > phys_size_physmem)
  {
    printf("Error: useable memory offset (%ld) + size (%ld) must be "
           "less than the physmem area physiscal size (%ld).\n",
                    mem_offset, max_mapping, phys_size_physmem);
    exit (EXIT_FAILURE);
  }
 
  phys_addr += mem_offset;
  addr_user += mem_offset/4;

  printf(" Start address of useable memory: 0x%08lx (%p), offset: 0x%08lx.\n",
             phys_addr, addr_user, mem_offset);
  printf(" Size of the useable memory: %ld (0x%08lx) bytes, %ld (0x%08lx) words.\n",
           max_mapping, max_mapping, 
           max_mapping >> 2, max_mapping >> 2);

  max_buffer_size = max_mapping - MAX_MAPPING_RF - 4*MAX_LEN - 2*PAGE_LEN;
  he = max_buffer_size % PAGE_LEN;
  if (he)
    max_buffer_size += PAGE_LEN - he;
  max_block = max_buffer_size/block_len;

  if (max_fifo > max_block)
  {
    printf("Error: max_fifo (%d) times block_len (%d) must be "
           "less than reserved physmem area buffer size (%ld).\n", 
                    max_fifo, block_len, max_buffer_size);
    exit (EXIT_FAILURE);
  }

  phys_addr_rf = phys_addr + max_buffer_size;
  addr_user_rf = addr_user + (max_buffer_size >> 2);

  phys_addr_data = phys_addr_rf + MAX_MAPPING_RF;
  addr_user_data = (unsigned long *)addr_user_rf + (MAX_MAPPING_RF >> 2);
  he = phys_addr_data % PAGE_LEN;
  if (he)
  {
    phys_addr_data += PAGE_LEN - he;
    addr_user_data += (PAGE_LEN - he) >> 2;
  }

  if (check_level)
  {
    /* fill 'expected' buffer */

    expected = addr_user_data;
    if (data_check)
    {
      if (pattern_file != NULL)
      {
        patf = fopen(pattern_file, "r");
        if (patf == NULL)
        {
          printf("Error opening pattern file: %s. ", pattern_file_name);
          perror("");
          exit (EXIT_FAILURE);
        }
        for (i = 0; i < max_len; i++)
          expected[i] = (unsigned long)&expected[i];
        ret = read_mif(patf, expected, max_len, &read_word);
        if (ret)
        {
          printf("Error found in pattern file %s at line number %d.\n",
                   pattern_file_name, ret);
          exit (EXIT_FAILURE);
        }
        printf(" Number of words read from file %s: %d\n",
                 pattern_file_name, read_word);
      }
      else // if not mif file
      {
        fill_expected(0);
      }
    }
    else if (counter_check)
    {
      expected[0] = init_count;
      init_count++;
    }
  }

  if (output_file != NULL)
    full_event = addr_user_data;

  phys_addr_end = phys_addr_data + (MAX_LEN << 2);

  rf = (rorcReadyFifo_t *)addr_user_rf;

  /* open RORC and map mailboxes */
  if (rorcOpenChannel(&prorc, p_minor, channel))
    exit (EXIT_FAILURE);
  step_indicator = 100000 * prorc.loop_per_usec;
  timeout = ustimeout * prorc.loop_per_usec;
  resp_cycle = resp_wait * prorc.loop_per_usec;
  if (resp_cycle > DDL_MAX_WAIT_CYCLE )
  {
    printf("Error: time-out value %ld us is too large.\n", resp_wait);
    printf("Please choose a value less than or equal to %d us.\n", 
                                                      prorc.max_resp_time);
    exit (EXIT_FAILURE);
  }

  /* check hw */
  hw_ser = rorcSerial(&prorc);

  /* check max_fifo */
  fw = rorcReadFw(&prorc);
  if (rorcFFSize(fw) && (rorcFFSize(fw) < max_fifo))
  {
    printf("Error: max_fifo (%d) too big.\n", max_fifo);
    rorcInterpretFw(fw);
    rorcClose(&prorc);
    exit (EXIT_FAILURE);
  }

  /* check fw version */
  fwvers_major = rorcFWVersMajor(fw);
  fwvers_minor = rorcFWVersMinor(fw);

  kell_diu = (send_trig || read_back || (loopback < 3));
  kell_siu = (kell_diu &&  (loopback != 1));

  if (reset_level)
  {
    rorcArmDDL(&prorc, RORC_RESET_RORC);
    printf(" RORC reset\n");
  }

  if ((reset_level > 1) && kell_diu)
  {
    rorcArmDDL(&prorc, RORC_RESET_DIU);
    printf(" DIU reset\n");

    if ((reset_level > 2) && (loopback != 1))
    {    
      /* wait a little before SIU reset */
      usleep(100000);
      /* reset SIU */
      if (rorcArmDDL(&prorc, RORC_RESET_SIU) != RORC_STATUS_OK)
      {
        printf("Error: Could not reset SIU. I stop.\n");
        rorcClose(&prorc);
        exit (EXIT_FAILURE);
      }
      printf(" SIU reset\n");

      rorcArmDDL(&prorc, RORC_RESET_DIU);
      printf(" DIU reset\n");
    }

    rorcArmDDL(&prorc, RORC_RESET_RORC);
    printf(" RORC reset\n");
  }

  if (reset_level)
  {
    /* wait a little after reset */
    usleep(100000);
  }

  /* clear SIU/DIU status */
  if (kell_siu)
  {
    status = ddlReadSiu(&prorc, 0, timeout);
    if (status != -1)
      printf(" SIU status cleared. Status word: 0x%08lx\n", status);
  }
  if (kell_diu)
  {
    status = ddlReadDiu(&prorc, 0, timeout);
    if (status != -1)
      printf(" DIU status cleared. Status word: 0x%08lx\n", status);
  }

  if (sleep_time)
    srand(time(NULL));

  /* signal handling */
  interrupt_arrived = 0;
  set_sig_hand();

  /* start data cycles */
  printf(" Asking RORC %d/%d to recieve data of pattern '%c' and of %d word length, %d time(s).\n",
          p_minor, channel, pattern, max_len, max_event);
  if (random_seed != RORC_DG_NO_RANDOM_LEN)
  {
    rounded_len = roundPowerOf2(max_len);
    printf(" Event lengths will vary randomly between 1 and %d. Seed = 0x%x.\n",
                        rounded_len, random_seed);
  }
  printf(" Event buffer address: 0x%08lx (%8p), size: 0x%08lx (%ld) bytes\n",
           phys_addr, addr_user, max_buffer_size, max_buffer_size);
  printf(" Memory block (page) size: 0x%08x (%d) bytes, number of blocks: %d\n",
           block_len, block_len, max_block);
  printf(" Ready FIFO address: 0x%08lx (%8p), offset: 0x%08lx, size: %d bytes\n", 
           phys_addr_rf, addr_user_rf, phys_addr_rf - phys_addr+mem_offset, MAX_MAPPING_RF);
  printf(" Expected event address: 0x%08lx (%8p), offset: 0x%08lx,\n",
           phys_addr_data, addr_user_data, phys_addr_data - phys_addr+mem_offset);
  printf("          size: 0x%08lx (%ld) bytes (%ld words)\n",
           phys_addr_end - phys_addr_data,
           phys_addr_end - phys_addr_data,
           (phys_addr_end - phys_addr_data) >> 2);
  printf(" Start of free memory: 0x%08lx, offset: 0x%08lx\n",
           phys_addr_end, phys_addr_end - phys_addr + mem_offset);
  printf(" Free memory size: 0x%08lx (%ld) bytes (%ld words)\n",
           max_mapping - (phys_addr_end + 4 - phys_addr),
           max_mapping - (phys_addr_end + 4 - phys_addr),
           (max_mapping - (phys_addr_end + 4 - phys_addr)) >> 2);
  printf(" max fifo: %d, min fifo: %d, sleep time: %d ms.\n",
           max_fifo, min_fifo, sleep_time);
  printf(" Timeout values used: %d us (%d cycles),\n", ustimeout, timeout);
  printf(" resp_wait: %ld us (%ld cycles).\n", resp_wait, resp_cycle);
  if (output_file != NULL)
    printf(" Events will be written to file %s\n", output_file_name);
  if (read_back)
    printf(" Events will be read using STBRD command\n");
  if (counter_check)
  {
    printf(" Event counter will be checked.\n");
  }
  else
    printf(" Event counter will be not checked.\n");
  if (data_check)
  {
    printf(" Events will be checked.\n");
  }
  else
    printf(" Events will be not checked.\n");
  if (ddl_header)
    printf(" The first %d words are considered as DDL header.\n", ddl_header);
  rorcInterpretSerial(hw_ser);
  printf(" Firmware version: %d.%d.\n", fwvers_major, fwvers_minor);

  time(&start_time);
  gettimeofday(&start_tv, NULL);
  loop = 0;
  event = 0;
  continuation = 0;
  n_ddl_block = 0;
  open_transaction = 0;

  /* push free fifo for the first max_fifo events */
  rorcReset(&prorc, RORC_RESET_FF);
  if (rorcCheckFreeFifo(&prorc) == RORC_FF_EMPTY)
    printf(" Rx address FIFO emptied\n");
  else
  {
    printf("Error: I can not empty the Rx address FIFO\n");
    exit (EXIT_FAILURE);
  }
  free_block = max_block;
  next_free_index = 0;
  rf_free_index = 0;
  next_free_block_phys = phys_addr;
  next_free_block = addr_user;
  block[next_free_index].rf = rf_free_index;
  block[next_free_index].addr = next_free_block;;
  debug("First phys=0x%08lx, user=%8p, index=%d into block[%d]\n", 
      next_free_block_phys, next_free_block, rf_free_index, next_free_index);
  debug("PUSH %d entries\n", max_fifo);
  for (i = 0; i < max_fifo; i++)
    next_push(&prorc, data_check);

  /* send start data collecting command */
  rorcStartDataReceiver(&prorc, phys_addr_rf);
  printf(" Data Receiver started\n");

  if (generator)
  {
    /* start Data Generator */
    ret = rorcArmDataGenerator(&prorc, user_init_count, init_word, RORC_DG_INCR, 
                               max_len, random_seed, &rounded_len);
    if (ret != RORC_STATUS_OK)
      printf("Error: Data Generator returned error %d\n", ret);
    else
      printf(" Data Generator armed\n");

    if (loopback)
    {
      if (loopback == 1) // DIU loopback
      {
       ret = ddlDiuLoopBack(&prorc, timeout, &stw);
       if (ret != RORC_STATUS_OK)
         printf("Error: DIU loopback returned error %d\n", ret);
       else
         printf(" DIU set in loopback\n");
      }	
      else if (loopback == 2) // SIU loopback
      {
       ret = ddlSetSiuLoopBack(&prorc, timeout, &stw);
       if (ret != RORC_STATUS_OK)
         printf("Error: SIU loopback returned error %d\n", ret);
       else
         printf(" SIU set in loopback mode\n");
      }
      else // RORC loopback
      {      
        rorcParamOn(&prorc, PRORC_PARAM_LOOPB);
        printf(" Rorc channel loopback mode enabled\n");
      }
    }

    /* wait a little after loopback setting */
    usleep(100000);
    /* clear SIU/DIU status */
    if (kell_siu)
    {
      status = ddlReadSiu(&prorc, 0, timeout);
      if (status != -1)
        printf(" SIU status cleared. Status word: 0x%08lx\n", status);
    }
    if (kell_diu)
    {
      status = ddlReadDiu(&prorc, 0, timeout);
      if (status != -1)
        printf(" DIU status cleared. Status word: 0x%08lx\n", status);
    }
      
    rorcStartDataGenerator(&prorc, max_event);
    printf(" Data Generator started\n");
  }
  else if(send_trig)
  {
    /* clear SIU/DIU status */
    if (kell_siu)
    {
      status = ddlReadSiu(&prorc, 0, timeout);
      if (status != -1)
        printf(" SIU status cleared. Status word: 0x%08lx\n", status);
    }
    if (kell_diu)
    {
      status = ddlReadDiu(&prorc, 0, timeout);
      if (status != -1)
        printf(" DIU status cleared. Status word: 0x%08lx\n", status);
    }

    /* RDYRX command to FEE */
    ret = rorcStartTrigger(&prorc, resp_cycle, &stw);
    if (ret == RORC_LINK_NOT_ON)
    {
      printf("Error: LINK IS DOWN, RDYRX command can not be sent. I stop\n");
      goto vege;
    }
    if (ret == RORC_STATUS_ERROR)
    {
      printf("Error: RDYRX command can not be sent. I stop\n");
      goto vege;
    }
    if (ret == RORC_NOT_ACCEPTED)
      printf(" No reply arrived for RDYRX in timeout %ld usec\n", resp_wait); 
    else
      printf(" FEE accepted the RDYDX command. Its reply: 0x%08lx\n",
                  stw.stw);
    open_transaction = 1;
  }
  fflush (stdout);

  /* start looping */
  stopped = 0;
  status_arrived = 0;
  last_data_len = 0;
  sum_data_len = 0;
  comp_err = 0;
  interrupt_arrived = 0;
  n_read = 0;
  next_index = 0;
  rf_index = block[next_index].rf;
  buffer_address = block[next_index].addr;
  debug("For the first cycle rf_index: %d, buff_addr: %p from block[%d]\n", 
              rf_index, buffer_address, next_index);
  dtsw = 0; //to make the compiler happy
  output_open = 0;
  printf("\n\n");
  while (1)
  {
    if (read_back)
    {
      /* send STBRD command */
      ret = ddlStartBlockRead(&prorc, fee_address, resp_cycle, &stw);
      if (ret == RORC_LINK_NOT_ON)
      {
        printf("Error: LINK IS DOWN, STBRD command can be sent. I stop\n");
        goto vege;
      }
      if (ret == RORC_STATUS_ERROR)
      {
        printf("Error: STBRD command can not be sent. I stop\n");
        goto vege;
      }
      if (ret == RORC_NOT_ACCEPTED)
        printf(" No reply arrived for STBRD in timeout %ld usec\n\n", resp_wait);
      open_transaction = 1;
    }

    /* loop for DMA ending */
    i = 0;
    data_arrived = 1;
    while (!rorcHasData(rf, rf_index))
    { 
      if(stopped)
      {
        if (ddlCheckStatus(&prorc))
        {
          printf(" Status arrived\n");
          stw = ddlReadStatus(&prorc);
// check status word !!!!!!!!!!!!!!!!!!!
//         printf("\nUnexpected status: 0x%08lx\n", stw.stw);
          printf(" Status word after stop: 0x%08lx\n", stw.stw);
          loop++;
          goto vege;
        }
      }
      if (interrupt_arrived && !stopped)
      {
        data_arrived = 0;
        break;
      }
      i++;
      if ((i % step_indicator) == 0)
      {
        if (!page_print)
	{	
          printf("%s%c ==== Data received: at least %d Gbyte ====\n",LINE_UP, 
	                indicator[(i / step_indicator) % 4], sum_Gbyte);
	}
	else
	{
          printf("%s%c ==== Page counter: %d ====\n",LINE_UP, 
	                indicator[(i / step_indicator) % 4], loop);
	}
        
        if (loopback < 3)
        {
          /* check if link is on */
          if (rorcCheckLink(&prorc) != RORC_STATUS_OK)
          {
            printf("\nError: The DDL link is down.\n"); 
            goto vege;
          }
        }

        if(stopped)
        {
          printf("\nNo data arrived in timeout after the stop\n");
          goto vege;
        }
      }
    }

    if (data_arrived)
    {
      n_read++;
      data_len = rf[rf_index].length;
      dtsw = rf[rf_index].status;
//      if (!pRorc(&prorc)) dRorcReadReg(&prorc, 19); // read debug register
      if (dtsw)
      {
        n_ddl_block++;
        if (!continuation)
        {
          event++;
        }

        if (read_back)
        {
          /* send EOBTR command */
          ret = ddlStopDataTransfer(&prorc, resp_cycle, &stw);
          if (ret == RORC_STATUS_ERROR)
            printf("Error: EOBTR (after data read) command can not be sent.\n\n");

          if (ret == RORC_NOT_ACCEPTED)
            printf(" No reply for EOBTR (after data read) in timeout %ld usec\n\n",
                                                                     resp_wait);
          open_transaction = 0;
        }
      }
      dtsw_len = (dtsw & 0x7ffff000) >> 12;
      debug("DATA: event=%d cycle=%d rf_index=%d dtstw=0x%08lx buffer=%8p\n", 
                    event, loop+1, rf_index, dtsw, buffer_address);

      sum_byte += 4*data_len;
      if (sum_byte >= GIGA)
      {
        sum_Gbyte++;
        sum_byte -= GIGA;
        if (!page_print && ((sum_Gbyte % gb_print) == 0))
          printf("%s  ==== Data received: at least %d Gbyte ====\n", 
                                                      LINE_UP, sum_Gbyte);
      }

      /* check the length and DTSTW */
      comp_err = 0;
      sum_data_len += data_len;
      if (dtsw)
      {
        if (((dtsw & 0xff) != 0x82) || (sum_data_len != dtsw_len))
        {
          printf("\nDTSTW ERROR in cycle %d. DTSTW: 0x%08lx\n",
                         loop + 1, dtsw);
          printf(" data len: 0x%08x (%d), sum_data_len: 0x%08x (%d), dtsw_len: 0x%08x (%d)\n",
               data_len, data_len, sum_data_len, sum_data_len, dtsw_len, dtsw_len);
          comp_err = 1;
        }
        if (dtsw & 0x80000000)
        {
          printf("\nERROR bit set in DTSTW in cycle %d. DTSTW: 0x%08lx\n",
                          loop + 1, dtsw);
          comp_err = 1;
        }
      }
 
      /* check the data */
      if (check_level)
      {
        istart = 0;
        if (!continuation && ddl_header)
          istart = ddl_header; 
	      
        for (i = istart, j = last_data_len; i < data_len; i++, j++)
        {
          if (!counter_check && (j == 0) && (i == istart) && !read_back)
            continue;
	  if (!data_check && (continuation || (i > istart))) 
	    break;
          if (*(buffer_address + i) != expected[j])
          {
            comp_err = 1;
            printf("\nDATA COMPARE ERROR in cycle %d\n", loop + 1);
            printf("%d: expected = %08lx, received = %08lx\n",
                    j, expected[j], *(buffer_address + i));
            break;
          }
        }
      } // data check

      if (output_file != NULL)
      {
        for (i = 0, j = last_data_len; i < data_len; i++, j++)
          full_event[j] = *(buffer_address + i);
        if (dtsw)
        {
          //write out the event 
          if (!output_open)
          {
            output_open = 1;
			sprintf(filename,"%s%d",output_file_name,file_counter);
            ofile = fopen(filename, "wb");    
            if(ofile == NULL)
            {
              printf("Can not open output file %s - events not written\n",
                       output_file_name);
            }
            else
            {
              //fprintf(ofile, "# %s %s", output_file_name, ctime(&start_time));
            }
          }
          if (ofile != NULL)
          {
            if (!continuation)
              //fprintf(ofile, "\n# Event %d\n", event);
            //fprintf(ofile, "\n# Block %d\n", n_ddl_block);
            //fprintf(ofile, "0x%08x\n\n", sum_data_len);
            for (i = 0; i < sum_data_len; i++)
              //fprintf(ofile, "0x%08lx\n", full_event[i]);
			fwrite(&full_event[i],1,sizeof(unsigned long),ofile);
          }
        }         
      } //output file

      if (dtsw)
      {
        last_data_len = 0;
        sum_data_len = 0;

        /* prepare for the next DDL block */

        continuation = dtsw & CONTINUATION_BIT;
        if (continuation)
        {
          if (data_check)
            fill_expected(continuation);
        }
        else
        {
          if (data_check)
          {
            if (n_ddl_block > 1)
              fill_expected(continuation);
            else
            {
              expected[0] = init_count;
              init_count++;
            }
          }
	  else if (counter_check)
          {
            expected[0] = init_count;
            init_count++;
          }
          n_ddl_block = 0;
        }
      }  
      else
      {
        last_data_len += data_len;
      }

    } // data arrived

    if(!stopped)
    {
      if (comp_err || (interrupt_arrived && !continuation) || 
                     (max_event && (event >= max_event) && !continuation) )
      {
        if (generator)
        {
          rorcStopDataGenerator(&prorc);
          printf(" Data Generator stopped\n");
/*
	  if (loopback == 2)
          {
             ret = ddlClearSiuLoopBack(&prorc, timeout, &stw);
             if (ret != RORC_STATUS_OK)
               printf("Error: Stop SIU loopback returned error %d\n", ret);
             else
               printf(" SIU loopback cleared\n");
          }
*/
          loop++;
          break;
        }
        else if (send_trig || (read_back && open_transaction))
        {
          /* send EOBTR to FEE */
          ret = rorcStopTrigger(&prorc, timeout, &stw);
          if (ret == RORC_LINK_NOT_ON)
          {
            printf("Error: LINK IS DOWN, EOBTR command can not be sent. I stop\n");
            loop++;
            break;
          }
          if (ret == RORC_STATUS_ERROR)
          {
            printf("Error: EOBTR command can not be sent. I stop\n");
            loop++;
            break;
          }
          printf(" EOBTR command sent to the FEE\n");
          open_transaction = 0;
          if (ret != RORC_NOT_ACCEPTED)
          {
            printf(" FEE accepted the EOBTR command. Its reply: 0x%08lx\n", 
                      stw.stw);
            loop++;
            break;
          }
        }
        if (comp_err || (read_back && interrupt_arrived))
        {
          loop++;
          break;
        }
        stopped=1;
      }
    }

    if(data_arrived)
    {
      /* prepare for next block (page) */
      free_block++;
      next_index++;
      if (next_index >= max_block)
        next_index = 0;
      rf_index = block[next_index].rf;
      buffer_address = block[next_index].addr;
      debug("For the next cycle rf_index: %d, buff_addr: %p from block[%d]\n", 
                rf_index, buffer_address, next_index);

      if (dtsw)
      {
        if (sleep_time)
        {
          /* sleep random time, less than sleep_time, simulating LDC */        

//          st = (int)((float)(sleep_time + 1)*rand()/(RAND_MAX + 1.0));
          st = (float)sleep_time*rand()/RAND_MAX;
          debug("SLEEP: %d ms\n", st);
          usleep(1000 * st);
        }
        if ((max_fifo - n_read) <= min_fifo)
        {          /* push free fifo for the next n_read events */
          debug("PUSH %d entrie(s)\n", n_read);
          while (n_read)
          { 
            next_push(&prorc, data_check);
            n_read--;
          } /* while (n_read) */
        } /* if (max_fifo - n_read <= min_fifo) */
      }  /* if (dtstw) */
      else
      {
        if ((max_fifo - n_read) <= 1)
        {
          debug("PUSH 1 entry\n");
          next_push(&prorc, data_check);
          n_read--;
        }
      }

      /* end loop */
      loop++;
      if (page_print && ((loop % page_print) == 0))
        printf("%s  ==== Page counter: %d ====\n",LINE_UP, loop);
      if (max_event && (event >= max_event) && !continuation)
        break;
     } /* if data_arrived */
	eventcounter++;
	if(eventcounter==2){ 
		eventcounter=0;
	fclose(ofile);
	output_open=0;
	file_counter++;
	}
  } /* while(1) */

vege:
  /* send stop data collecting command */
  rorcStopDataReceiver(&prorc);

  printf("The program has read %d event(s).\n", event);
  if (output_file != NULL)
  {
    if (ofile != NULL)
    {
      //fprintf(ofile, "\n# End of events. %d event(s) in the file.\n", event);
      fclose(ofile);
      printf("Events written to file %s\n", output_file_name);
    } 
  }

  /* calculate and print the elapsed time */
  gettimeofday(&now_tv, NULL);
  print_time(loop, &now_tv, &start_tv, &start_time, sum_Gbyte, sum_byte, stat_file);

  /* unmap and close RORC */
  rorcClose(&prorc);

  /* unmap and close physmem memory */
  if (physmemClose (physmem_fd, (unsigned long *)(addr_user-mem_offset/4), phys_size_physmem))
    exit (EXIT_FAILURE);

  exit (EXIT_SUCCESS);
}
