/****************************************************
 * rorc_reg
 *
 *  Accessing pRORC or D-RORC  registers
 *
 * last updated: 22/02/2005
 * written by: Ervin Denes
 ****************************************************/
#define SOURCE_FILE "rorc_reg.c"
#define DESCRIPTION "Show/set RORC registers"

#include <ctype.h>
#include "rorc_ddl.h"

#define ADDR_MAX 0x7c

char *reg_name[] = {
	"RCSR", "RERR", "RFID", "RHID", "CCSR", "CERR", "CDCR", "CDSR",
	"PGP1", "PGP2", "PGP3", "PGP4", "PGSR", "RRBA", "RAFL", "RAFH",
	"TRBA", "TAFL", "TAFH", "RVED", "RDR1", "RDR2", "RDR3", "RDR4",
	"TDR1", "TDR2", "TDR3", "TDR4", "RXDC", "TXDC", "RXDA", "TXDA"
};

/* register attributes: {read, write, integer print} */
int reg_att[][3] = {
	{1, 1, 0}, {1, 0, 0}, {1, 0, 0}, {1, 1, 0},
        {1, 1, 0}, {1, 0, 0}, {0, 1, 0}, {0, 0, 0}, 
	{0, 1, 0}, {0, 1, 0}, {0, 1, 0}, {0, 1, 0},
	{1, 0, 0}, {0, 1, 0}, {0, 1, 0}, {0, 1, 0},
	{0, 1, 0}, {0, 1, 0}, {0, 1, 0}, {1, 0, 0},
	{1, 0, 1}, {1, 0, 1}, {1, 0, 1}, {1, 0, 1},
	{1, 0, 1}, {1, 0, 1}, {1, 0, 1}, {1, 0, 1},
	{1, 0, 1}, {1, 0, 1}, {1, 0, 0}, {1, 0, 0}
};

int shift = 12;
int len = 32 + 8;

/******************************************************/
void usage(char *name)
{
  printf("usage: %s [-{M|m} <minor>|0] [-{C|c} <channel|0>]\n", name);
  printf("                [-{A|a} <register_addres|0>] [-{V|v} <hex value>]\n");
  printf("       Maximum register address value is 0x%x\n", ADDR_MAX);
  printf("       If data value not given, the given (or all) register will be read out.\n");
  printf("       -a and -v are ignored if pRorc has revision number 1.\n");
}
/*******************************************************
* print table header                                   *
*******************************************************/
void print_head(int minor, int channel, int revision)
{
  int j, la;
  char ba[8];

  sprintf (ba, "Ch: %d %d", minor, channel);
  la = strlen(ba);
  for(j = shift+len; j > 0; j--) printf("-"); printf("\n");
  printf("%s", ba);
  for(j = shift - la; j > 0; j--) printf(" ");
  for(j = 31; j > 9; j--)
  {
    printf("%d", (j/10));
    if (!(j % 4)) printf (" ");
    if (j == 16) printf (" ");
  }
  printf("\n");
  sprintf (ba, "%s%d", "rev = ", revision);
  la = strlen(ba);
  printf("%s", ba);
  for(j = shift - la; j > 0; j--) printf(" ");
  for(j = 31; j >= 0; j--)
  {
    printf("%d",(j%10));
    if (!(j % 4)) printf (" ");
    if (j == 16) printf (" ");
  }
  printf("\n");
  for(j = len+shift; j > 0; j--) printf("-"); printf("\n");

}

/*******************************************************
* print a register                                     *
*******************************************************/
void print_reg(char *name, volatile unsigned long *addr, __u32 mask, int dec)
{
  int j;
  __u32 he;
    
  printf("%s (%02x) - ", name, (__u32)addr & 0xff);
  he = (__u32)*addr & mask;
  for(j = 31; j >=0; j--)
  {
    printf("%d",((he >> j) & 1));
    if (!(j % 4)) printf (" ");
    if (j == 16) printf (" ");
  }
  printf("- ");
  if (he)
    printf("%04x %04x", ((he >> 16) & 0xffff), (he & 0xffff));
  else
    printf("%x", he);
  if (dec && he) printf (" (%u)", he);
  printf("\n");
}
	
/******************************************************/
int main(int argc, char **argv)
{
  int c, i, minor, channel;
  int addr_min, addr_max, no_addr, no_val;
  __u32 mask = (__u32)-1;

  rorc_pci_dev_t prorc;
  int addr = 0;
  __u32 data = 0;

  no_addr =1;
  no_val = 1;
  addr_min = 0;
  addr_max = ADDR_MAX;

  /* Check parameters */
  minor = 0;
  channel = 0;
  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:A:a:V:v:")) != EOF)
    {
      switch(toupper(c))
      {
        case 'M': minor = atoi(optarg); if (minor < 0) minor = 0; break;
        case 'C': 
          channel = atoi(optarg); 
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                       MAX_CHANNEL);
            usage(argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'A':
          addr = strtoul(optarg, (char**)NULL, 16); 
          if (addr > addr_max)
          {
            usage(argv[0]);
            exit (EXIT_FAILURE); 
          }
	  no_addr = 0;
          break;
        case 'V':
          data  = strtoul(optarg, (char**)NULL, 16); no_val = 0; break;
        default : usage (argv[0]); exit (EXIT_FAILURE); break;
      }
    }
  }

  if (rorcMapChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);

  printf("\n");

  if (pRorc(&prorc))
  {
    print_head(minor, channel, 1);

    print_reg ("mbef", prorc.mbef, mask, 0);
    print_reg ("omb1", prorc.omb[1], mask, 1);
    print_reg ("omb2", prorc.omb[2], mask, 1);
    print_reg ("omb3", prorc.omb[3], mask, 1);
    print_reg ("omb4", prorc.omb[4], mask, 1);
    print_reg ("imb1", prorc.imb[1], mask, 1);
    print_reg ("imb2", prorc.imb[2], mask, 1);
    print_reg ("imb3", prorc.imb[3], mask, 1);
    print_reg ("imb4", prorc.imb[4], mask, 1);
    print_reg ("mwar", prorc.mwar, mask, 0);
    print_reg ("mwtc", prorc.mwtc, mask, 0);
    print_reg ("mrar", prorc.mrar, mask, 0);
    print_reg ("mrtc", prorc.mrtc, mask, 0);
    print_reg ("mbef", prorc.mbef, mask, 0);
    print_reg ("incs", prorc.intcsr, mask, 0);
    print_reg ("mcsr", prorc.mcsr, mask, 0);

    for(i = shift+len; i > 0; i--) printf("-"); printf("\n");
  }
  else
  {
    i = addr >> 2;
    if (!no_val)
    {
      if (!reg_att[i][1])
      {
        printf("%s (0x%02x) is not a writable register.\n", reg_name[i], addr);
        goto vege;
      }
      *prorc.reg[i] = data;
    }

    if (no_val && !no_addr)
    {
      if (!reg_att[i][0])
      {
        printf("%s (0x%02x) is not a readable register.\n", reg_name[i], addr);
//        goto vege;
  	reg_att[i][0] = 1;   // print not readable as well
      }
      addr_min = addr_max = addr >> 2;
    }
    else
      addr_max = ADDR_MAX >> 2;

    print_head(minor, channel, prorc.rorc_revision);

    for (i = addr_min; i <= addr_max; i++)
    {
      if (reg_att[i][0])
      {
        print_reg(reg_name[i], prorc.reg[i], mask, reg_att[i][2]);
      }
    }
    for(i = shift+len; i > 0; i--) printf("-"); printf("\n");
  }

vege:
  printf("\n");

  rorcClose(&prorc);

  return 0;
}

