/****************************************************
 * rorc_send.c
 *
 *  Testing RORC's DMA download capabilities
 *
 * last updated: 17/02/2006
 * written by: Peter Csato and Ervin Denes
 *
 * history:                                           
 *                         
 * v.4.4.5  17-Feb-2006: check time-out option
 * v.4.4.3  22-Jul-2005: DG counters not implemented for D-RORC
 *          10-Jul-2005: infinit wait time for data sending
 *          13-Jun-2005: physmem size should be less than physiscal size
 * v.4.3.5  07-Aug-2004: use usec timeouts    
 * v.4.3    16-Feb-2004: download using D-RORC
 * v.4.2    23-Aug-2003: add channel variable
 *
 ****************************************************/
#define SOURCE_FILE "rorc_send.c"
#define DESCRIPTION "Test of RORC's DMA download"

#define LINE_UP "[1A[0J"
//#define LINE_UP "[1A[80D[0J"
#define PAGE_LEN (1024*2) 

#include <ctype.h>
#include <signal.h>
#include <getopt.h>
#include "rorc_ddl.h"

#define MAX_MAPPING   (1024*1024*30)
#define MAX_EV ((MAX_MAPPING - DDL_MAX_BYTE - 16)/(DDL_MAX_BYTE+4))
// #define MAX_EV 1000
#define EV_LEN 2047
#define FEE_MEMLEN 512

#define GB_PRINT 1

static struct sigaction sigac;
// int interrupt_arrived = 0;

int check, r_loopback, d_loopback, x_loopback, send_command, inf_time;
unsigned long fee_address;
unsigned long max_mapping, mem_offset;
int generator, init_count, gb_print;
int pattern, pattern_no, data_length;
int read_back;
int ustimeout, timeout, p_minor, channel, max_loop;
unsigned long init_word;
char *input_file;
char input_file_name[80];

int rounded_len, random_seed, random_data;

/*******************************************************/

void usage(char *name)
{
  printf("usage: %s  [{-M|-m|--minor}      <RORC's minor number>|0]\n", name);
  printf("                  [{-C|-c|--channel}    <RORC's channel  number>|0]\n");
  printf("                  [{-L|-l|--length}     <data_length in words>|%d]\n", EV_LEN); 
  printf("                  [{-B|-b|--rand_len}   <seed>|%d]\n", RORC_DG_NO_RANDOM_LEN);
  printf("                  [-J|-j|--rand_data]\n"); 
  printf("                  [-F|-f|--front_end]\n");
  printf("                  [{-A|-a|--address}    <fee address>|0]\n");
  printf("                  [{-P|-p|--pattern}    {c|a|0|1|I|d}]\n");
  printf("                  [{-I|-i|--init_word}  <init word>|0]\n");
  printf("                  [{-T|-t|--time_out}   <time-out in usec>|%d]\n",
                                    DDL_RESPONSE_TIME);
  printf("                  [{-U|-u|--physmem}    <utilisable physmem memory in MBs>|%d]\n",
                                    MAX_MAPPING/1024/1024);
  printf("                  [{-O|-o|--offset}     <offset of utilisable memory in MBs>|0]\n");
  printf("                  [{-E|-e|--events}     <max event>|0]\n");
  printf("                  [{-N|-n|--init_count} <initial event count>|1]\n");
  printf("                  [{-K|-k|--file}       <input file name>]\n");
  printf("                  [{-Q|-q|--byte_print} <gigabytes>]\n");
  printf("                  [-G|-g|--generator]\n");
  printf("                  [-R|-r|--rorc_lback]\n");
  printf("                  [-D|-d|--diu_lback]\n");
  printf("                  [-Y|-y|--ext_lback]\n");
  printf("                  [-S|-s|--STBWR]\n");
  printf("                  [-W|-w|--inf_wait]\n");
  printf("                  [-X|-x|--nocheck]\n");
}

/********************************************************
*          signal handler                               *
********************************************************/
void sig_hand(int signum)
{
/* signum: 1(SIGHUP)  2(SIGINT)  3(SIGQUIT) */
  if(signum < 4)
  {
    interrupt_arrived = 1;
  }
}

/********************************************************
*          signal handling                              *
********************************************************/
void set_sig_hand(void)
{
  sigac.sa_flags = 0;
  sigac.sa_handler =  sig_hand;
  if (sigaction(SIGHUP, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGHUP ");
     exit(1);
  }
  if (sigaction(SIGINT, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGINT ");
     exit(1);
  }
  if (sigaction(SIGQUIT, &sigac, NULL) < 0 )
  {
     perror(" sigaction SIGQUIT ");
     exit(1);
  }
}

/********************************************************
*         treat program parameters                      *
********************************************************/
void treat_param(int argc, char **argv)
{
  int c;
  int option_index = 0;
  char *short_options = 
     "M:m:C:c:FfA:a:L:l:B:b:JjP:p:I:i:T:t:E:e:N:n:K:k:Q:q:GgRrDdYySsWwXxU:u:O:o:"; 
  const struct option long_options[] =
  {
    {"minor",     1, 0, 'm'}, {"channel",   1, 0, 'c'},
    {"front_end", 0, 0, 'f'}, {"address",   1, 0, 'a'},
    {"length",    1, 0, 'l'}, {"rand_len",  1, 0, 'b'}, {"rand_data", 0, 0, 'j'},
    {"pattern",   1, 0, 'p'}, {"init_word", 1, 0, 'i'}, {"time_out",  1, 0, 't'},
    {"events",    1, 0, 'e'}, {"init_count",1, 0, 'n'}, {"file",      1, 0, 'k'},
    {"byte_print",1, 0, 'q'}, {"generator", 0, 0, 'g'},
    {"rorc_lback",0, 0, 'r'}, {"diu_lback", 0, 0, 'd'}, {"ext_lback", 0, 0, 'y'},
    {"STBWR",     0, 0, 's'}, {"inf_wait",  0, 0, 'w'}, {"no_check",  0, 0, 'x'},
    {"physmem",   1, 0, 'u'}, {"offset",    1, 0, 'o'},
    {0, 0, 0, 0}
  };

  /* default parameters */
  p_minor = 0;
  channel = 0;
  data_length = EV_LEN;
  random_seed = RORC_DG_NO_RANDOM_LEN;
  random_data = 0;
  fee_address = 0;
  pattern = 'i';
  init_word = 0;
  max_loop = 0;
  init_count = 1;
  gb_print = GB_PRINT;
  check = 1;
  inf_time = 0;
  r_loopback = 0;
  d_loopback = 0;
  x_loopback = 0;
  send_command = 0;
  read_back = 0;
  ustimeout= DDL_RESPONSE_TIME;
  generator = 0;
  max_mapping = MAX_MAPPING;
  mem_offset = 0;
  input_file = (char *)NULL;

  /* Check parameters */

  if (argc > 1)
  {
    while ((c = getopt_long(argc, argv, short_options, 
                                        long_options, &option_index)) != EOF)
    {
      switch(toupper(c))
      {
        case 'M': p_minor = atoi(optarg); break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'U': max_mapping = atoi(optarg)*1024*1024; break;
        case 'O': mem_offset = atoi(optarg)*1024*1024; break;
        case 'L': data_length = atoi(optarg); break;
        case 'B': random_seed = strtoul(optarg, (char**)NULL, 16); break;
        case 'J': random_data = 1; break;
        case 'F': read_back = 1; break;
        case 'A': fee_address = strtoul(optarg, (char**)NULL, 16); break;
        case 'P': pattern = tolower(*optarg); break;
        case 'I': init_word = strtoul(optarg, (char**)NULL, 16); break;
        case 'T': ustimeout = atoi(optarg); break;
        case 'E': max_loop = atoi(optarg); break;
        case 'N': init_count = atoi(optarg); break;
        case 'K': strcpy(input_file_name, optarg); input_file = input_file_name; break;
        case 'Q': gb_print = atoi(optarg); break;
        case 'G': generator = 1; break;
        case 'R': r_loopback = 1; break;
        case 'D': d_loopback = 1; break;
        case 'Y': x_loopback = 1; break;
        case 'W': inf_time = 1; break;
        case 'X': check = 0; break;
        case 'S': send_command = 1; break;
        default : usage (argv[0]); exit (EXIT_FAILURE); break;
      }
    }
  }

  switch (pattern)
  {
    case 'c': pattern_no = 1; break;
    case 'a': pattern_no = 2; break;
    case '0': pattern_no = 3; init_word = -2; break;
    case '1': pattern_no = 4; init_word = 1 ; break;
    case 'i': pattern_no = 5; break;
    case 'd': pattern_no = 6; break;
    default : pattern = 'i' ; pattern_no = 5; break;
  }

  if (read_back)
  {
    send_command = 1;
    if (data_length > FEE_MEMLEN)
      data_length = FEE_MEMLEN;
  }

  if (random_data && generator)
  {
    printf(" -J and -K are mutually exclusive!\n");
     exit (EXIT_FAILURE);
  }

  if ((input_file != NULL) && generator)
  {
    printf(" -G and -K are mutually exclusive!\n");
     exit (EXIT_FAILURE);
  }

  if (input_file != NULL)
    check = 0;

}

/********************************************************/

int main(int argc, char **argv)
{
  int open_transaction;
  int receive_length, compare_length;
  int waitrdy;
  int dtsw_len, max_ev;
  int i, ret, he, comp_err; 
  int loop, d_loop, loopback;
  int old_cdgs, d_bytes;
  double old_bytes;
  int fwvers_major, fwvers_minor;
  rorc_pci_dev_t prorc;
  __u32 fw;
  rorcHwSerial_t hw_ser;
  unsigned long data, dtsw;
  stword_t stw;
  rorcStatus_t status;

  volatile unsigned long *addr_user;
  volatile unsigned long *trbar_address; 
  volatile unsigned long *buffer_address;
  volatile unsigned long *rrbar_address;
  volatile unsigned long *loopb_address;

  unsigned long phys_addr;
  unsigned long trbar_phys;
  unsigned long buffer_phys;
  unsigned long rrbar_phys;

  unsigned long loopb_phys;
  int physmem_fd;
  unsigned long phys_size_physmem;

  typedef struct
  {
    volatile unsigned long *user;
    unsigned long phys;
    int length;
  } event_addr_t;

  int event_index, n_val;
  event_addr_t next_event[MAX_EV]; 
  FILE *inpf = NULL;

  int dsec, dusec;
  double dtime;
  struct timeval start_tv, now_tv;
  rorcCounter_t byte_counter;

  int sum_byte=0, sum_Gbyte=0;

  /* treat parameters */
  treat_param(argc, argv);
  loopback = (input_file == NULL) && (r_loopback || d_loopback || x_loopback);

  if (input_file != NULL)
  {
    inpf = fopen(input_file, "r");
    if (inpf == NULL)
    {
      printf("Error opening input file: %s. ", input_file_name);
      fflush(stdout);
      perror("");
      exit (EXIT_FAILURE);
    }
  }

  if (physmemOpen (&physmem_fd, &addr_user, &phys_addr,
                   &phys_size_physmem))
    exit (EXIT_FAILURE);

  if (max_mapping < 1 || max_mapping > phys_size_physmem)
  {
    printf("The size of utilizable memory must be positive and less than %ld byte.\n",
            phys_size_physmem);
    max_mapping = phys_size_physmem;
    printf("Utilizable memory size is changed to %ld byte.\n", max_mapping); 
  }

  if ((mem_offset + max_mapping) > phys_size_physmem)
  {
    printf("Error: useable memory offset (%ld) + size (%ld) must be "
           "less than the physmem area physiscal size (%ld).\n",
                    mem_offset, max_mapping, phys_size_physmem);
    exit (EXIT_FAILURE);
  }

  if (data_length < 1 || data_length > DDL_MAX_WORD ||
      3 + (DDL_MAX_WORD + 1) + (data_length + 1) > max_mapping/4 ) //!!!!!
  {
    printf("Error: data length must be positive and less than %d and %ld words.\n",
            DDL_MAX_WORD, max_mapping/4 - DDL_MAX_WORD - 5);    //!!!!!_
    exit (EXIT_FAILURE);
  }

  max_ev = MAX_EV;

  phys_addr += mem_offset;
  addr_user += mem_offset/4;

  printf(" Start address of useable memory: 0x%08lx (%p), offset: 0x%08lx.\n",
             phys_addr, addr_user, mem_offset);
  printf(" Size of the useable memory: %ld (0x%08lx) bytes, %ld (0x%08lx) words.\n",
           max_mapping, max_mapping, max_mapping >> 2, max_mapping >> 2);

  /* load physmem area with the wanted data */

  rrbar_phys     = phys_addr;
  loopb_phys     = phys_addr + 8;
  trbar_phys     = phys_addr + 8 + DDL_MAX_BYTE + 4;
  he = trbar_phys % PAGE_LEN;
  if (he)
    trbar_phys += PAGE_LEN - he;
  buffer_phys    = trbar_phys + 8;

  rrbar_address  = addr_user;
  loopb_address  = addr_user + 2;
  trbar_address  = addr_user + 2 + DDL_MAX_WORD + 1;
  if (he)
    trbar_address += (PAGE_LEN - he) >> 2;
  buffer_address = trbar_address + 2;

  if (input_file != NULL)
  {
    next_event[0].phys = buffer_phys;
    next_event[0].user = (unsigned long*)buffer_address;
    for ( i = 1; i < max_ev; i++)
    {
      n_val = getVal((unsigned long *)&rounded_len, 1, inpf, "%x");
      if ((n_val != 1) || (rounded_len > DDL_MAX_WORD))
      {
        max_ev = i - 1;
        break;
      }
      next_event[i-1].length = rounded_len;
      n_val = getVal((unsigned long *)next_event[i-1].user, rounded_len,
                                         inpf, "%x");
      if (n_val != rounded_len)
      {
        max_ev = i - 1;
        break;
      }

      next_event[i].phys = next_event[i-1].phys + DDL_MAX_BYTE + 4; //!!!!!
      next_event[i].user = next_event[i-1].user + DDL_MAX_WORD + 1; //!!!!!
    }
    fclose(inpf);
  }
  else if (random_data)
  {
    next_event[0].phys = buffer_phys;
    next_event[0].user = (unsigned long*)buffer_address;
    *next_event[0].user = 1;
    for (i = 1; i < data_length; i++)
      *(next_event[0].user + i) = rand();
    next_event[0].length = data_length;

    next_event[1].phys = next_event[0].phys + DDL_MAX_BYTE + 4; //!!!!!
    next_event[1].user = next_event[0].user + DDL_MAX_WORD + 1; //!!!!!
    *next_event[1].user = 0;
    next_event[1].length = data_length;
  }
  else
  {
    *buffer_address = init_count - 1;
    data = init_word;
    for (i = 1; i < data_length; i++)
    {
      *(buffer_address + i) = data;
      switch (pattern)
      {
        case 'c': break;
        case 'a': data = ~data; break;
        case '0':
        data = (data <<1) | 1;
          if (data == -1)
          data = init_word;
          break;
        case '1': 
          data <<= 1;
          if (data == 0)
          data = init_word;
          break;
        case 'i': data++; break;
        case 'd': data--; break;
        default:  data++; break;
      }
    }  
    *(buffer_address + i) = 0xdeadbeef; //!!!!!
  }

  /* open RORC and map mailboxes */
  if (rorcOpenChannel(&prorc, p_minor, channel))
    exit (EXIT_FAILURE);
  timeout = ustimeout * prorc.loop_per_usec;
  if (timeout > DDL_MAX_WAIT_CYCLE )
  {
    printf("Error: time-out value %d us is too large.\n", ustimeout);
    printf("Please choose a value less than or equal to %d us.\n",
                                                      prorc.max_resp_time);
    exit (EXIT_FAILURE);
  }
  printf(" %d usec timeout corresponds to %d loops\n", ustimeout, timeout);

  /* initialize the DMA */

  printf(" RORC %d/%d tries to download data of pattern '%c'.\n", p_minor, channel, pattern);

  if (random_seed != RORC_DG_NO_RANDOM_LEN)
  {
    printf(" Event lengths will vary randomly between 1 and ");
    if (generator)
    {
      rounded_len = roundPowerOf2(data_length);
      printf("%d.", rounded_len);
    }
    else
    {
      srand(random_seed);
      printf("%d.", data_length);
    }
    printf(" Seed = 0x%x.\n", random_seed);
  }
  else
    printf(" Event length will be %d words.\n", data_length);

  printf(" The physical address of the source data and return address: 0x%08lx 0x%08lx.\n", 
           buffer_phys, trbar_phys);
  if (input_file != NULL)
  {
    printf(" Events will be read from file %s\n", input_file_name);
    printf(" The maximal number of different events are %d\n", max_ev);
  }
  if (loopback)
  {
    if (r_loopback)
      printf(" RORC loopback is selected.\n");
    else if (d_loopback)
      printf(" DIU loopback is selected.\n");
    else
      printf(" External loopback is supposed.\n");
    printf(" Loop-back data is expected at physical address 0x%08lx with Ready FIFO at 0x%08lx.\n",
           loopb_phys, rrbar_phys);
    if (check)
      printf(" Events will be checked.\n");
    else
      printf(" Events will be not checked.\n");
  }
  if (read_back)
    printf(" Downloaded data will read back to physical address 0x%08lx with Ready FIFO at 0x%08lx.\n",
           loopb_phys, rrbar_phys);


  /* check hw */
  hw_ser = rorcSerial(&prorc);
  rorcInterpretSerial(hw_ser);
  /* check fw version */
  fw = rorcReadFw(&prorc);
  fwvers_major = rorcFWVersMajor(fw);
  fwvers_minor = rorcFWVersMinor(fw);
  printf(" Firmware version: %d.%d.\n", fwvers_major, fwvers_minor);

  /* put RORC or DIU to loopback and start data receiving*/
  if (r_loopback)
  {
    rorcParamOn(&prorc, PRORC_PARAM_LOOPB);
    printf(" RORC set in loopback.\n");
  }

  if (d_loopback)
  {
    ret = ddlDiuLoopBack(&prorc, timeout, &stw);
    if (ret == RORC_STATUS_ERROR)
      printf(" DIU Loopback command can not be sent.\n");
    else if (ret == RORC_NOT_ACCEPTED)
      printf(" No reply arrived for DIU Loopback command in timeout %d usec\n", 
                 ustimeout);
    else
      printf(" DIU set in loopback. Its reply: 0x%08lx\n", stw.stw);
  }

  if (read_back)
  {
    ret = rorcArmDDL(&prorc, RORC_RESET_SIU);
    if (ret == RORC_STATUS_OK)
      printf(" SIU reset OK.\n");
  }

  if (loopback || read_back)
  {
//    rorcReset(&prorc, RORC_RESET_FF);          temporary 2004.02.13.
    rorcStartDataReceiver(&prorc, rrbar_phys);
    printf(" Data Receiver started\n");
  }

  /* signal handling */
  interrupt_arrived = 0;
  set_sig_hand();

  gettimeofday(&start_tv, NULL);
  rorcReset(&prorc, RORC_RESET_COUNTERS);

  if (generator)
  {
    /* start Data Generator */
    ret = rorcArmDataGenerator(&prorc, init_count, init_word, pattern_no,
                                data_length, random_seed, &rounded_len);
    if (ret != RORC_STATUS_OK)
      printf(" Data Generator returned error %d\n", ret);
    else
      printf(" Data Generator armed\n");

    rorcStartDataGenerator(&prorc, max_loop);
    printf(" Data Generator started\n");
  }

  /* start looping */
  event_index = 0;
  loop = 0;
  open_transaction = 0;
  old_cdgs = 0;
  old_bytes = 0.0;
  printf("\n\n");
  while (1)
  {
    if (interrupt_arrived)
      break;

    if (loopback || read_back)
    {
      /* set Ready FIFO and push Free FIFO */
      if (check)
      {
        /* fill receive buffer */
        for (i = 0; i < data_length+1; i++) // !!!!!
          *(loopb_address + i) = (unsigned long) (loopb_address + i);
      }
      *rrbar_address = -1;
      *(rrbar_address + 1) = -1;
      rorcPushFreeFifo(&prorc, loopb_phys, data_length+1, 0); // !!!!!
    }

    /* prepare data */
    if ((input_file == NULL) & !random_data)
     (*buffer_address)++;     // inrcement event count
    if (!generator)
    {
      if (input_file != NULL)
      {
        if (loop && (event_index == 0))
        {
          for (i = 0; i < max_ev; i++)
            *next_event[i].user += max_ev;  // increment event count
        }
        buffer_phys = next_event[event_index].phys;
        buffer_address = next_event[event_index].user;
        data_length = next_event[event_index].length;
        event_index++;
        if (event_index >= max_ev)
          event_index = 0;
      }

      if (random_data)
      {
        buffer_phys = next_event[event_index].phys;
        buffer_address = next_event[event_index].user;
        data_length = next_event[event_index].length;
        event_index ^= 1;
      }

      /* calculate random length */
      rounded_len = data_length;
      if ((random_seed != RORC_DG_NO_RANDOM_LEN) && (input_file == NULL))
      {
//        rounded_len = (int)((float)data_length * rand()/(1.0 + RAND_MAX)) + 1;
rounded_len = loop % 512 + 1;  //debug
// for (i = 3; i < rounded_len; i ++)
//  *(buffer_address + i) = rounded_len;  // debug

if ((rounded_len < 1) || (rounded_len > 512)) 
{
 printf("loop: %d, length: %d\n", loop, rounded_len); fflush(stdout);
 rorcClose(&prorc);
 exit (-1);
}
      }

      if (send_command)
      {
        /* STBWR command to FEE */
        ret = ddlStartBlockWrite(&prorc, fee_address, timeout, &stw);
        if (ret == RORC_LINK_NOT_ON)
        {
          printf("\nLINK IS DOWN. Download is not possible!\n");
          break;
        }
        if (ret == RORC_NOT_ACCEPTED)
        {
          printf(" No reply arrived for STBWR in timeout %d usec\n\n",
			                                     ustimeout);
          break;
        }
        if (ret == RORC_STATUS_ERROR)
        {
          printf(" STBWR command can not be sent.\n\n");
          break;
        }
        open_transaction = 1;
      }

      /* send_data */
      waitrdy = 1;
      while(1)
      {
        ret = ddlSendData(&prorc, buffer_phys, rounded_len, 
                                  trbar_phys, trbar_address, 
				  timeout, waitrdy); 
        if (!inf_time || interrupt_arrived || 
                             ((ret != RORC_NOT_ABLE) && (ret != RORC_TIMEOUT)))
          break;
      }
      if (ret == RORC_LINK_NOT_ON)
      {
        printf("\nLINK IS DOWN. Download is not possible!\n");
        break;
      }
      if (ret == RORC_NOT_ABLE)
      {
        printf("\nPREVIOUS DOWNLOAD NOT FINISHED. New download is not possible!\n");
        break;
      }
      if (ret == RORC_TIMEOUT)
      {
        printf("\nTransfer %d not done in timeout %d usec.\n", 
			               loop+1, ustimeout);
        break;
      }

      if (send_command)
      {
        /* send EOBTR to FEE */
        ret = rorcStopTrigger(&prorc, timeout, &stw);
        if (ret == RORC_LINK_NOT_ON)
        {
          printf("\nLINK IS DOWN, EOBTR command can not be sent\n");
          break;
        }
        if (ret == RORC_STATUS_ERROR)
        {
          printf(" EOBTR command can not be sent.\n\n");
          break;
        }
        if (ret == RORC_NOT_ACCEPTED)
        {
          printf(" No reply arrived for EOBTR in timeout %d usec\n\n", 
			                                         ustimeout);
          break;
        }
        open_transaction = 0;
      }

      if (random_data)
      {
        /* prepare next event */
        *next_event[event_index].user += 2;
        for (i = 1; i < data_length; i++)
          *(next_event[event_index].user + i) = rand();
      }
    } // !generator

    if (read_back)
    {
      /* send STBRD command */
      ret = ddlStartBlockRead(&prorc, fee_address, timeout, &stw);
      if (ret == RORC_LINK_NOT_ON)
      {
        printf(" LINK IS DOWN, STBRD command can not be sent.\n\n");
        loop++;
        break;
      }

      if (ret == RORC_STATUS_ERROR)
      {
        printf(" STBRD command can not be sent.\n\n");
        loop++;
        break;
      }

      if (ret == RORC_NOT_ACCEPTED)
        printf(" No reply arrived for STBRD timeout %d usec \n\n", ustimeout);
      open_transaction = 1;
    }

    if (loopback || read_back)
    {
      /* wait for arriving data */
      i = 0;
      while (*(rrbar_address + 1) == -1)
      {
        if (interrupt_arrived)
          break;
        i++;
        if ((i % 100000000) == 0)
        {
          printf ("!");
          fflush (stdout);
        }
      }
      if (interrupt_arrived)
        break;

      if (read_back)
      {
        /* send EOBTR command */
        ret = rorcStopTrigger(&prorc, timeout, &stw);
        if (ret == RORC_LINK_NOT_ON)
        {
          printf(" LINK IS DOWN. EOBTR (after data read) can not be sent.\n\n");
          loop++;
          break;
        }

        if (ret == RORC_STATUS_ERROR)
        {
          printf(" EOBTR (after data read) can not be sent.\n\n");
          loop++;
          break;
        }

        if (ret == RORC_NOT_ACCEPTED)
          printf(" No reply for EOBTR (after data read) in timeout %d usec\n\n",
                                                                     ustimeout);
        open_transaction = 0;
      }

      if (check)
      {
        /* check data */
        comp_err = 0;
        receive_length = *rrbar_address;
        dtsw = *(rrbar_address + 1);
        dtsw_len = (dtsw & 0x7ffff000) >> 12;
        if (((dtsw & 0xff) != 0x82) || 
            (!read_back && !generator && (dtsw_len != rounded_len)) ||
            (!read_back &&  generator && (dtsw_len  > rounded_len)) ||
            ( read_back               && (dtsw_len != FEE_MEMLEN ))
           )
        {
          printf("\nDTSTW ERROR in cycle %d. DTSTW: 0x%08lx\n",
                         loop + 1, dtsw);
          printf(" data len: 0x%08x (%d), dtsw_len: 0x%08x (%d)\n",
                         rounded_len, rounded_len, dtsw_len, dtsw_len);
          comp_err = 1;
        }
        if (receive_length != dtsw_len)
        {
          printf("\nLENGTH ERROR in cycle %d. DTSTW: 0x%08lx\n",
                         loop + 1, dtsw);
          printf(" dtstw len: 0x%08x (%d), received_len: 0x%08x (%d)\n",
                       dtsw_len, dtsw_len, receive_length, receive_length);
          comp_err = 1;
        }
        compare_length = receive_length;
        if (read_back && (random_seed != RORC_DG_NO_RANDOM_LEN))
          compare_length = rounded_len;
        for (i = 0; i < compare_length; i++)
        {
          if (*(buffer_address + i) != (*(loopb_address + i)))
          {
            comp_err = 1;
            printf("\nDATA COMPARE ERROR in cycle %d\n", loop + 1);
            printf("%d: expected = %08lx, received = %08lx\n",
                    i, *(buffer_address + i), *(loopb_address + i));
            break;
          }
        }
        if (comp_err)
        {
          loop++;
          break;
        }
      }
    }

    if (generator && pRorc(&prorc))
      rorcReadCounters(&prorc, &byte_counter, 0);
    if (loopback || read_back || !generator)
    {
      sum_byte += 4 * rounded_len;
      loop++;
    }
    else // (!loopback && !read_back && generator)
    {
      rorcReadRorcStatus(&prorc, &status); 
      d_loop = status.cdgs - old_cdgs; // Csaba should imlement these counters
                                       // 2004.12.14
      old_cdgs = status.cdgs;
      if (d_loop < 0)
        d_loop += 0x1000000;
      loop += d_loop;

      if (pRorc(&prorc))
      {
        d_bytes = (int)(byte_counter.bytes - old_bytes);
        old_bytes = byte_counter.bytes;
        sum_byte += d_bytes;
      } 
    }
    if (sum_byte >= GIGA)
    {
      sum_Gbyte++;
      sum_byte -= GIGA;
      if ((sum_Gbyte % gb_print) == 0)
      {
        gettimeofday(&now_tv, NULL);
	elapsed(&now_tv, &start_tv, &dsec, &dusec);
        dtime = (double)dsec + (double)dusec/MEGA;
        if (dtime)
        {
          printf("%s==== Data sent: ", LINE_UP);
          if (sum_Gbyte > 0)
            printf("%d.%09d Gbyte at ", sum_Gbyte, sum_byte);
          else
            printf("%d byte at ", sum_byte);
          if (sum_Gbyte > 0)
            printf("%.4f G", ((double)sum_Gbyte+(double)sum_byte/GIGA)/dtime);
          else
            printf("%s", order((double)sum_byte/dtime, 3));
          if (generator)
          {
            if (pRorc(&prorc))
              printf("B/s (%sB/s) speed. ====\n", order(byte_counter.lspeed, 3));
          }
          else
            printf("byte/sec speed. ====\n");
        }
      }
    }

    if (max_loop && (loop >= max_loop))
      break;
  }

  if (generator)
  {
    rorcStopDataGenerator(&prorc);
    printf(" Data Generator stopped\n");
    if (!loopback && !read_back)
    {
      rorcReadRorcStatus(&prorc, &status);
      d_loop = status.cdgs - old_cdgs;
      if (d_loop < 0)
        d_loop += 0x1000000;
      loop += d_loop;

      if (pRorc(&prorc))
      {
        rorcReadCounters(&prorc, &byte_counter, 0);
        d_bytes = (int)(byte_counter.bytes - old_bytes);
        sum_byte += d_bytes; 

        if (sum_byte >= GIGA)
        {
          sum_Gbyte++;
          sum_byte -= GIGA;
        }
      }
    }
  }
  if (read_back && open_transaction)
  {
    /* send EOBTR to FEE */
    ret = rorcStopTrigger(&prorc, timeout, &stw);
    printf(" EOBTR command sent to the FEE\n");
    open_transaction = 0;
    if (ret == RORC_LINK_NOT_ON)
      printf(" LINK IS DOWN, EOBTR command can not be sent.\n");
    else if (ret == RORC_STATUS_ERROR)
      printf(" EOBTR command can not be sent.\n");
    else if (ret != RORC_NOT_ACCEPTED)
      printf(" FEE accepted the EOBTR command. Its reply: 0x%08lx\n", stw.stw);
  }

  /* stop data collecting */
  if (loopback || read_back)
    rorcStopDataReceiver(&prorc);

  /* calculate and print the elapsed time */
  if (generator && pRorc(&prorc))
    rorcReadCounters(&prorc, &byte_counter, 0);
  gettimeofday(&now_tv, NULL);
  printf("The program has sent %d event(s).\n", loop);
  if (generator && pRorc(&prorc))
    printf("Running time, according to RORC: %f seconds.\n", byte_counter.time);
  print_time(loop, &now_tv, &start_tv, 0, sum_Gbyte, sum_byte, (char *)NULL);
  
  /* unmap and close RORC */
  rorcClose(&prorc);

  /* unmap and close physmem memory */
  if (physmemClose (physmem_fd, (unsigned long *)(addr_user-mem_offset/4), phys_size_physmem))
    exit (EXIT_FAILURE);

  exit (EXIT_SUCCESS);
}
