/***********************************************************
 * rorc_send_command.c
 *
 *  Send DDL command via pRORC         
 *
 * last updated: 17/02/2006
 * written by: Peter Csato and Ervin Denes
 *
 * history:                                           
 *                                                   
 * v.4.4.5 23-Aug-2003: check time-out option
 * v.4.2   23-Aug-2003: add channel variable
 *         26-Aug-2003: add JTAG commands
 *
 ***********************************************************/
#define SOURCE_FILE "rorc_send_command.c"
#define DESCRIPTION "Send DDL commands via pRORC"

#include <ctype.h>
#include "rorc_ddl.h"

#define MAX_ADDRESS 0x20

/* variables for command recognition */
char *command_tokens[] = {                   /* RORC commands */
                          "PRSTAT",          /* Get pRORC status */
                          "PRID",            /* Get pRORC IDs */
                          "LBON",            /* RORC loop-back ON */
                          "LBOFF",           /* RORC loop-back OFF */
                          "HLT_FLCON",       /* HLT flow-control on */
                          "HLT_FLCOFF",      /* HLT flow-control off */
                          "HLTON",           /* HLT splitter ON */
                          "HLTOFF",          /* HLT splitter OFF */
                          "STEON",           /* pRORC stop-on-error ON */
                          "STEOFF",          /* pRORC stop-on-error OFF */
                          "STARTWDMA",       /* Start W_DMA */
                          "STOPWDMA",        /* Stop W_DMA */
                          "STARTDG",         /* Start Data Generator */
                          "STOPDG",          /* Stop Data Generator */
                          "STARTDLOAD",      /* Start R_DMA */
                          "STOPDLOAD",       /* Stop R_DMA */
                          "STARTJTAG",       /* Start JTAG sending */
                          "STOPJTAG",        /* Stop JTAG sending*/
                                             /* DDL commands */
                          "RDYRX",           /* Ready to receive */
                          "EOBTR",           /* End of block transfer */
                          "DTCC",            /* Data tr. check command */
                          "JSTART",          /* Start JTAG block */
                          "EOJTR",           /* End of JATG transfer */
                          "JTCC",            /* JTAG tr. check command */
                          "STBWR",           /* Start block write */
                          "STBRD",           /* Start block read */
                          "FECTRL",          /* FEE control */
                          "FESTRD",          /* FEE status read */
                          "SRST",            /* SIU reset */
                          "SUSPEND",         /* Go to POFF state */
                          "WAKEUP",          /* Wakeup from POFF state */
                          "TXLOOP",          /* DIU loop-back */
                          "RXLOOP",          /* SIU loop-back on/off */
                          "SIU_TSTM",        /* Set PRBS test mode in SIU */
                          "DIU_TSTM",        /* Set PRBS test mode in DIU */
                          "TSTM_OFF",        /* Stop PRBS test mode in DIU */
                          "CL_TST_CNT",      /* Clear PRBS test counter */
                          "RD_TST_CNT",      /* Read PRBS test counter, */
                          "DIUHW",           /* DIU hardware version read */
                          "SIUHW",           /* SIU hardware version read */
                          "DIUFW",           /* DIU firmeware version read */
                          "SIUFW",           /* SIU firmeware version read */
                          "DIUST",           /* DIU status read */
                          "SIUST",           /* SIU status read */
                          "DIUPM",           /* DIU power monitor read */
                          "SIUPM",           /* SIU power monitor read */
                          ""
                         };
enum command_index {CD_PRSTAT=1, CD_PRID, CD_LBON, CD_LBOFF,
                    CD_HLT_FLCON, CD_HLT_FLCOFF, CD_HLTON, CD_HLTOFF,
		    CD_STEON, CD_STEOFF,
                    CD_STARTWDMA, CD_STOPWDMA, CD_STARTDG, CD_STOPDG,
                    CD_STARTDLOAD, CD_STOPDLOAD, CD_STARTJTAG, CD_STOPJTAG, 
                    CD_RDYRX, CD_EOBTR, CD_DTCC, CD_JSTART, CD_EOJTR, CD_JTCC,
                    CD_STBWR, CD_STBRD,
                    CD_FECTRL, CD_FESTRD, CD_SRST, CD_SUSPEND, CD_WAKEUP,
                    CD_TXLOOP, CD_RXLOOP, CD_SIU_TSTM, CD_DIU_TSTM,
                    CD_TSTM_OFF, CD_CL_TST_CNT, CD_RD_TST_CNT,
                    CD_DIUHW, CD_SIUHW, CD_DIUFW, CD_SIUFW, CD_DIUST, CD_SIUST,
                    CD_DIUPM, CD_SIUPM};

/********************************************************
*          print symbolic command names                 *
********************************************************/
void symb_comm(void)
{
  int i;

  printf("Symbolic command names are:\n");
  i = 0;
  while (1)
  {
    printf("%s ", command_tokens[i]);
    if (strcmp(command_tokens[i], "") == 0)
      break;
    i++;
    if ((i % 10) == 0)
      printf("\n");
  }
  printf("\n");
  printf("===============================================================================\n");
  exit(0);
}

/*******************************************************/

void usage(char *name)
{
  printf("Send a DDL command via RORC.\n");
  printf("usage: %s [-{M|m} <minor>] [-{C|c} <channel>] -{W|w} <command>\n", name);
  printf("                [-{R|r} <reg_num>] [-{T|t} <time-out>] [-{V|v} <diu_version>]\n");
  printf("   or\n");
  printf("       %s [-{M|m} <minor>] [-{C|c} <channel>] -{W|w} <cmd_code>\n", name);
  printf("                -{D|d} <destination> [-{I|i} <tr_id>] [-{P|p} <parameter>]\n");
  printf("                [-{R|r} <reg_num>] [-{T|t} <time-out>] [-{V|v} <diu_version>]\n");
  printf(" <minor>	RORC's card number. Default = 0.\n");
  printf(" <channel>	RORC's channel number. Default = 0.\n");
  printf(" <command>	the command to send (32 bit hex number or symbolic command name).\n");
  printf(" <reg_num> 	register number (from 0 to 27) in case or D-RORC.\n");
  printf("           	Default: %d (DDL command register) in case of DDL commands,\n", C_DCR);
  printf("           	and      %d (Control/status reg.) in case of RORC commands.\n", C_CSR);
  printf("           	-R is ignored in case of pRORC or if using symbolic command.\n");
  printf(" <cmd_code>   	the command code (4 bit hex number).\n"); 
  printf(" <destination>	the command destination (4 bit):\n");
  printf("              	0: pRORC, 1: DIU, 2: SIU, 4: FEE, 8: JTAG.\n");
  printf(" <tr_id>	the transaction id (0-15). Default = 0.\n");
  printf(" <parameter>	the command parameter (19 bit hex number). Default = 0.\n");
  printf(" <time-out>	time-out for command sending and status receiving.\n"); 
  printf("		Default = %d usec.\n", DDL_RESPONSE_TIME);
  printf(" <diu_version>	%d: no DIU, %d: old DIU, %d: new DIU. Default =  new DIU.\n",
            NO_DIU, OLD, NEW);

  symb_comm();
}

/********************************************************/

int main(int argc, char **argv)
{
  int c, minor, channel, egyben, reg_num, dest, transid, diu_version;
  int time, ustime, ret, n, wait_repl;
  rorc_pci_dev_t prorc;
  __u32 command, param, mb2, mb3, mb4;
  stword_t stw;
  int c_index, rorc_command, hex_code;
  char command_code[16];
  unsigned long fee_addr;
  int hw_addr, hw_char;

  /* Check parameters */

  if (argc < 2)
  {
    printf(" The command parameter is missing!\n");
    usage (argv[0]);
    exit (EXIT_FAILURE);
  }

  minor = 0;
  channel = 0;
  reg_num = -1;
  dest = -1;
  command_code[0] = '\0';
  transid = 0;
  fee_addr = 0;
  param = 0;
  ustime = DDL_RESPONSE_TIME;
  diu_version = NEW;
  egyben = 1;
  if (argc > 1)

  while ((c = getopt(argc, argv, "M:m:C:c:W:w:R:r:D:d:I:i:P:p:T:t:V:v:")) != EOF)
  {
    switch(toupper(c))
    {
      case 'M': minor = atoi(optarg); if (minor < 0) minor = 0; break;
      case 'C': channel = atoi(optarg);
        if ((channel < 0) || (channel >= MAX_CHANNEL))
        {
          printf(" Channel number should be positive and less than %d\n",
                              MAX_CHANNEL);
          usage (argv[0]);
          exit (EXIT_FAILURE);
        }
        break;
      case 'W': sprintf(command_code, "%s", optarg); break;
      case 'R': reg_num = atoi(optarg); break;
      case 'D': dest = atoi(optarg) & 0xf; egyben = 0; break;
      case 'I': transid = atoi(optarg); egyben = 0; break;
      case 'P': param = strtoul(optarg, (char**)NULL, 16) & 0x7ffff; egyben = 0; break;
      case 'T': ustime = atoi(optarg); break;
      case 'V': 
        diu_version = atoi(optarg); 
        if ((diu_version != NO_DIU) && (diu_version != OLD) && (diu_version != NEW))
        {
          printf("Version number should be %d, %d or %d\n", NO_DIU, OLD, NEW);
          exit (EXIT_FAILURE);
        }
        break;
      default : usage (argv[0]); exit (EXIT_FAILURE); break;
    }
  }

  if (strlen(command_code) == 0)
  {
    printf(" The command parameter is missing!\n");
    usage(argv[0]);
    exit (EXIT_FAILURE);
  }

  if (!egyben && (dest == -1))
  {
    printf(" Command destination is missing!\n");
    usage(argv[0]);
    exit (EXIT_FAILURE);
  }

  /* open RORC and map mailboxes */
  if (rorcMapChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);
  time = ustime * prorc.loop_per_usec;
  if (time > DDL_MAX_WAIT_CYCLE )
  {
    printf("Error: time-out value %d us is too large.\n", ustime);
    printf("Please choose a value less than or equal to %d us.\n",
                                                      prorc.max_resp_time);
    exit (EXIT_FAILURE);
  }
  prorc.diu_version = diu_version;
  printf(" RORC revision = %d\n", prorc.rorc_revision);
  printf(" Supposed DIU version = %s\n", prorc.diu_version == NO_DIU ? "no DIU" :
                                         prorc.diu_version == OLD ? "old" : "new");

  /* set command */

  c_index = n_key_word(command_tokens, command_code);
  command = 0;
  rorc_command = 0;
  hex_code = 0;
  wait_repl = 1;
  if (pRorc(&prorc))
  {
    switch (c_index)
    {
      case CD_PRSTAT:     command = PRORC_CMD_GET_STAT;     break;
      case CD_PRID:       command = PRORC_CMD_GET_ID;       break;
      case CD_LBON:       command = PRORC_CMD_LOOPB_ON;     break;
      case CD_LBOFF:      command = PRORC_CMD_PARAM_RESET;  break;
      case CD_HLT_FLCON:  command = -1;                     break;
      case CD_HLT_FLCOFF: command = -1;                     break;
      case CD_HLTON:      command = -1;                     break;
      case CD_HLTOFF:     command = -1;                     break;
      case CD_STEON:      command = PRORC_CMD_STOP_ERR_ON;  break;
      case CD_STEOFF:     command = PRORC_CMD_PARAM_RESET;  break;
      case CD_STARTWDMA:  command = PRORC_CMD_START_W_DMA;  break;
      case CD_STOPWDMA:   command = PRORC_CMD_STOP_W_DMA;   break;
      case CD_STARTDG:    command = PRORC_CMD_START_DG;     break;
      case CD_STOPDG:     command = PRORC_CMD_STOP_DG;      break;
      case CD_STARTDLOAD: command = PRORC_CMD_DOWNL_DATA;   break;
      case CD_STOPDLOAD:  command = -1;                     break;
      case CD_STARTJTAG:  command = PRORC_CMD_DOWNL_JTAG;   break;
      case CD_STOPJTAG:   command = -1;                     break;
    }
    if (command == -1)
    {
      printf("Command %s not allowed for pRORC.\n", command_code);
      rorcClose(&prorc);
      exit (EXIT_FAILURE);
    }
  }
  else
  {
    switch (c_index)
    {
      case CD_PRSTAT:     command = -1; break;
      case CD_PRID:       command = -1; break;
      case CD_LBON:
        if (dRorcCheckLoopBack(&prorc))
        {
          printf("RORC is in loop-back, no command sent\n");
          rorcClose(&prorc);
          exit (EXIT_SUCCESS);
        }
        command = DRORC_CMD_LOOPB_ON_OFF;
        break;
      case CD_LBOFF:
        if (!dRorcCheckLoopBack(&prorc))
        {
          printf("RORC is not in loop-back, no command sent\n");
          rorcClose(&prorc);
          exit (EXIT_SUCCESS);
        }
        command = DRORC_CMD_LOOPB_ON_OFF;
        break;
      case CD_HLT_FLCON:
        if (prorc.rorc_revision < INTEG)
          command = -1;
        else
        {
          if (dRorcCheckHltFlctl(&prorc))
          {
            printf("HLT flow-control is on, no command sent\n");
            rorcClose(&prorc);
            exit (EXIT_SUCCESS);
          }
          command = DRORC_CMD_HLT_FLC_ON_OFF;
        }
        break;
      case CD_HLT_FLCOFF:
        if (prorc.rorc_revision < INTEG)
          command = -1;
        else
        {
          if (!dRorcCheckHltFlctl(&prorc))
          {
            printf("HLT flow-control is off, no command sent\n");
            rorcClose(&prorc);
            exit (EXIT_SUCCESS);
          }
          command = DRORC_CMD_HLT_FLC_ON_OFF;
        }
        break;
      case CD_HLTON:
        if (prorc.rorc_revision < INTEG)
          command = -1;
        else
        {
          if (dRorcCheckHltSplit(&prorc))
          {
            printf("HLT splitter is on, no command sent\n");
            rorcClose(&prorc);
            exit (EXIT_SUCCESS);
          }
          command = DRORC_CMD_HLT_SPL_ON_OFF;
        }
        break;
      case CD_HLTOFF:
        if (prorc.rorc_revision < INTEG)
          command = -1;
        else
        {
          if (!dRorcCheckHltSplit(&prorc))
          {
            printf("HLT splitter is off, no command sent\n");
            rorcClose(&prorc);
            exit (EXIT_SUCCESS);
          }
          command = DRORC_CMD_HLT_SPL_ON_OFF;
        }
        break;
      case CD_STEON:      command = -1; break;
      case CD_STEOFF:     command = -1; break;
      case CD_STARTWDMA:
        if (dRorcReadReg(&prorc, C_CSR) & DRORC_CMD_DATA_RX_ON_OFF)
        {
          printf("Data Receiver already started, no command sent\n");
          rorcClose(&prorc);
          exit (EXIT_SUCCESS);
        }
        command = DRORC_CMD_DATA_RX_ON_OFF;
        break;
      case CD_STOPWDMA:
        if (!(dRorcReadReg(&prorc, C_CSR) & DRORC_CMD_DATA_RX_ON_OFF))
        {
          printf("Data Receiver already stopped, no command sent\n");
          rorcClose(&prorc);
          exit (EXIT_SUCCESS);
        }
        command = DRORC_CMD_DATA_RX_ON_OFF;
        break;
      case CD_STARTDG:    command = DRORC_CMD_START_DG; break;
      case CD_STOPDG:     command = DRORC_CMD_STOP_DG;  break;
      case CD_STARTDLOAD:
        if (dRorcReadReg(&prorc, C_CSR) & DRORC_CMD_DATA_TX_ON_OFF)
        {
          printf("Data Transmitter already started, no command sent\n");
          rorcClose(&prorc);
          exit (EXIT_SUCCESS);
        }
        command = DRORC_CMD_DATA_TX_ON_OFF;
        break;
      case CD_STOPDLOAD:
        if (!(dRorcReadReg(&prorc, C_CSR) & DRORC_CMD_DATA_TX_ON_OFF))
        {
          printf("Data transmitter already stopped, no command sent\n");
          rorcClose(&prorc);
          exit (EXIT_SUCCESS);
        }
        command = DRORC_CMD_DATA_TX_ON_OFF;
        break;
      case CD_STARTJTAG:  command = DRORC_CMD_JTAG_DOWN_ON;  break;
      case CD_STOPJTAG:   command = DRORC_CMD_JTAG_DOWN_OFF; break;
    }
    if (command == -1)
    {
      printf("Command %s not allowed for D-RORC revision %d.\n",
              command_code, prorc.rorc_revision);
      rorcClose(&prorc);
      exit (EXIT_FAILURE);
    }
  }
  if (command)
  {
    rorc_command = 1;
    wait_repl = 0;
  }
  else
  {
    switch (c_index)
    {
      case CD_RDYRX:      command = RDYRX;     dest = FEE;  break;
      case CD_EOBTR:      command = EOBTR;     dest = FEE;  break;
      case CD_DTCC:       command = DTCC;      dest = SIU;  break;
      case CD_JSTART:     command = JSTART;    dest = JTAG; break;
      case CD_EOJTR:      command = EOJTR;     dest = JTAG; break;
      case CD_JTCC:       command = JTSTW;     dest = JTAG; break;
      case CD_STBWR:      command = STBWR;     dest = FEE;  break;
      case CD_STBRD:      command = STBRD;     dest = FEE;  break;
      case CD_FECTRL:     command = FECTRL;    dest = FEE;  break;
      case CD_FESTRD:     command = FESTRD;    dest = FEE;  break;
      case CD_SRST:       command = SRST;      dest = DIU;  break;
      case CD_SUSPEND:    command = SUSPEND;   dest = DIU;  break;
      case CD_WAKEUP:     command = WAKEUP;    dest = DIU;  break;
      case CD_TXLOOP:     command = IFLOOP;    dest = DIU;  break;
      case CD_RXLOOP:     command = IFLOOP;    dest = SIU;  break;
      case CD_SIU_TSTM:   command = TSTMODE;   dest = SIU;  fee_addr = 0;       break;
      case CD_DIU_TSTM:   command = TSTMODE;   dest = DIU;  fee_addr = STRPRBS; break;
      case CD_TSTM_OFF:   command = TSTMODE;   dest = DIU;  fee_addr = STPPRBS; break;
      case CD_CL_TST_CNT: command = TSTMODE;   dest = DIU;  fee_addr = CLRPRBS; break;
      case CD_RD_TST_CNT: command = TSTMODE;   dest = DIU;  fee_addr = RDPRBS;  break;
      case CD_DIUHW:      command = RHWVER;    dest = DIU;  break;
      case CD_SIUHW:      command = RHWVER;    dest = SIU;  break;
      case CD_DIUFW:      command = RFWVER;    dest = DIU;  break;
      case CD_SIUFW:      command = RFWVER;    dest = SIU;  break;
      case CD_DIUST:      command = RandCIFST; dest = DIU;  break;
      case CD_SIUST:      command = RandCIFST; dest = SIU;  break;
      case CD_DIUPM:      command = RPMVAL;    dest = DIU;  break;
      case CD_SIUPM:      command = RPMVAL;    dest = SIU;  break;
      default:
        if (command_code[0] == '0' && command_code[1] == 'x')
        {
          command = strtoul(command_code, (char**)NULL, 16 );
          hex_code = 1;
          if (!pRorc(&prorc) && (reg_num == -1))
          {
            printf("Please give a register number as well.\n");
            rorcClose(&prorc);
            exit (EXIT_FAILURE);
          }
        }
        else
        {
          printf ("Invalid command code %s\n", command_code);
          rorcClose(&prorc);
          usage(argv[0]);
          exit (EXIT_FAILURE);
        }
    }
  }

  if (param == 0)
    param = fee_addr;

  if (dest != -1)
    egyben = 0;

  if (dest == 0)
    rorc_command = 1;

  if (pRorc(&prorc) || (!rorc_command && !hex_code))
  {
    if (reg_num != -1)
      printf(" Register number ignored.\n");
    /* prepare and send DDL command */
    if (!egyben)
      command = command & 0xf;
    ddlSendCommand(&prorc, dest, command, transid,  param, time);
    if (pRorc(&prorc))
      printf (" Command sent: ");
    else
      printf (" Data written into register %d: ", C_DCR);
    if (egyben)
      printf ("0x%08x\n", command);
    else
      printf ("0x%05x%1x%1x%1x\n", param, transid, command, dest);
  }
  else
  {
    if (!hex_code || (reg_num == -1))
    {
      if (reg_num != -1)
        printf(" Register number ignored.\n");
      reg_num = C_CSR;
    }
    dRorcWriteReg(&prorc, reg_num, command);
    printf (" Data written into register %d: 0x%08x\n", reg_num, command);
  }

  /* check the reply */

  if (wait_repl)
  {
    n = 0;
    while (1) 
    {
      ret = ddlWaitStatus(&prorc, time);
      if (ret >= time)
        break;
      n++;
      stw = ddlReadStatus(&prorc);
      if (n > 1)
        printf ("%d. ", n);
      printf("Replied DDL status: 0x%08lx: ", stw.stw);
      ddlInterpretStw(stw);
      printf ("\n");
      if (((stw.part.dest == DIU) || (stw.part.dest == SIU))) 
      {
        if (stw.part.code == IFSTW)
          ddlInterpretIFSTW(&prorc, stw.stw);
        else if (stw.part.code == HWSTW)
        {
          hw_addr = (stw.stw >> 12) & 0xff;
          hw_char = (stw.stw >> 20) & 0xff;
          printf("%s HW id's word %d = %c (0x%x)\n",
             stw.part.dest == DIU ? "DIU" : "SIU", hw_addr+1, hw_char, hw_char);
        }
        else if (stw.part.code == FWSTW)
          ddlInterpretFW(stw.stw);
      }

      if (pRorc(&prorc))
      {
        /* if something in MB2, MB3  and MB4 */
        if (pRorcCheckMb(&prorc, PRORC_NE_IMB2))
        {
          mb2 = pRorcReadMb (&prorc, 2);
          printf("      in MB2: 0x%08x (%d)\n", mb2, mb2);
        }
        if (pRorcCheckMb(&prorc, PRORC_NE_IMB3))
        {
          mb3 = pRorcReadMb (&prorc, 3);
          printf("      in MB3: 0x%08x (%d)\n", mb3, mb3);
        }
        if (pRorcCheckMb(&prorc, PRORC_NE_IMB4))
        {
          mb4 = pRorcReadMb (&prorc, 4);
          printf("      in MB4: 0x%08x (%d)\n", mb4, (mb4 & 0x7ffffff));
        }
      }
    }

    if (n)  
      printf("No more");
    else
      printf("No");
    printf(" reply arrived in time-out %d us.\n", ustime);
  }
    
  /* unmap and close RORC */
  rorcClose(&prorc);

  exit (0);
}
