/***********************************************************
 * rorc_send_jatag.c  
 *
 *  Send JTAG data via pRORC         
 *
 * last updated: 17/02/2006
 * written by: Peter Csato and Ervin Denes
 *                                                           
 * history:                                           
 *                                                   
 * v.4.1    07-May-2003: original version                      
 * v.4.1    04-Jul-2003: read input from file                  
 * v.4.2    24-Jul-2003: use physmem offset
 * v.4.2    22-Aug-2003: add channel variable
 * v.4.2    24-Aug-2003: display return code's text
 * v.4.3.5  07-Aug-2004: use us in timeout
 * v.4.4.5  17-Feb-2006: check time-out option
 * 
 ***********************************************************/
#define SOURCE_FILE "rorc_send_jtag.c"
#define DESCRIPTION "Send JTAG data via RORC (obsolate)"

#include <ctype.h>
#include "rorc_ddl.h"

/*******************************************************/
void usage(char *name)
{
  printf("Send JTAG data via RORC.\n");
  printf("usage: %s [-{M|m} <minor>|0] [-{C|c} <channel>|0]\n", name);
  printf("                      -{I|i} <input_fn> -{O|o} <output_fn>\n");
  printf("                      [-{P|p} <pysmem_offset>|0] [-{A|a} <j_addr>|0>]\n");
  printf("                      [-{L|l} <length>|%d] [-{T|t} <time-out>|%d]\n",
           DDL_MAX_JTAG_LEN, DDL_RESPONSE_TIME);
  printf(" <minor>          minor number of the RORC device\n");
  printf(" <channel>        RORC's channel number\n");
  printf(" <j_addr>         value to write as parameter in the JSTART command\n");
  printf(" <physmem_offset> offset in bytes of the useable part of physmem memory\n");
  printf(" <input_fn>       name of the input file (obligatory parameter)\n");
  printf(" <output_fn>      name of the output file (obligatory parameter)\n");
  printf(" <length>         length in words of the JTAG frame;\n");
  printf("                  <length> will be truncated to %d words.\n",
            DDL_MAX_JTAG_LEN);
  printf(" <time-out>       time-out in usecs for command sending and status receiving.\n");
}

/********************************************************/
int main(int argc, char **argv)
{
  int i, c, minor, channel, transid, ustimeout, timeout, ret, n_reply, step;
  int length, n_word, sum_word, pm_offset;
  rorc_pci_dev_t prorc;
  stword_t stw[DDL_MAX_REPLY];
  unsigned long param;
  int physmem_fd;
  volatile unsigned long *addr_user, *data_addr, *return_addr;
  unsigned long phys_size_physmem, phys_addr, data_phys, return_phys;
  unsigned long offset, r_offset, phys_addr_rf;

  int i_file, o_file;
  char i_fn[80], o_fn[80];
  FILE *i_fp, *o_fp;

  /* Check parameters */

  if (argc < 2)
  {
    printf(" Some parameter is missing!\n");
    usage (argv[0]);
    exit (EXIT_FAILURE);
  }

  minor = 0;
  channel = 0;
  length = DDL_MAX_JTAG_LEN;
  transid = 0;
  param = 0;
  i_file = 0;
  o_file = 0;
  ustimeout = DDL_RESPONSE_TIME;
  pm_offset = 0;

  if (argc > 1)

  while ((c = getopt(argc, argv, "M:m:C:c:P:p:L:l:A:a:I:i:O:o:T:t:")) != EOF)
  {
    switch(toupper(c))
    {
      case 'M': minor = atoi(optarg); if (minor < 0) minor = 0; break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
      case 'L': 
        length = atoi(optarg); 
        if (length > DDL_MAX_JTAG_LEN) 
          length = DDL_MAX_JTAG_LEN; 
        break;
      case 'A': param = strtoul(optarg, (char**)NULL, 16); break;
      case 'I': strcpy(i_fn, optarg); i_file = 1; break;
      case 'O': strcpy(o_fn, optarg); o_file = 1; break;
      case 'T': ustimeout = atoi(optarg); break;
      case 'P': pm_offset = strtoul(optarg, (char**)NULL, 16); break;
      default : usage (argv[0]); exit (EXIT_FAILURE); break;
    }
  }

  if (i_file == 0)
  {
    printf(" No input defined, input file is missing!\n");
    usage(argv[0]);
    exit (EXIT_FAILURE);
  }

  if (o_file == 0)
  {
    printf(" No output defined, output file is missing!\n");
    usage(argv[0]);
    exit (EXIT_FAILURE);
  }

  i_fp = fopen(i_fn, "r");
  if (i_fp == NULL)
  {
    printf("Error opening input file: %s. ", i_fn);
    fflush(stdout);
    perror("");
    exit (EXIT_FAILURE);
  }

  o_fp = fopen(o_fn, "w");
  if (o_fp == NULL)
  {
    printf("Error opening output file: %s. ", o_fn);
    fflush(stdout);
    perror("");
    exit (EXIT_FAILURE);
  }

  /* physmem */

  if (physmemOpen (&physmem_fd, &addr_user, &phys_addr,
                   &phys_size_physmem))
    exit (EXIT_FAILURE);

  printf(" Allocation of the physmem memory, physical address: 0x%08lx.\n", 
                             phys_addr);
  printf(" Size of the physmem memory: %ld (0x%08lx) bytes,\n",
           phys_size_physmem, phys_size_physmem);
  printf("                             %ld (0x%08lx) words.\n",
           phys_size_physmem >> 2, phys_size_physmem >> 2);

  phys_addr += pm_offset;
  addr_user += (pm_offset >> 2);
 
  printf(" Usable part of physmem memory starts at physical address: 0x%08lx.\n",
		             phys_addr);
  
  r_offset = 0;
  offset = length/4 + 2;
  data_phys = phys_addr + (offset << 2);
  data_addr = addr_user + offset;
  return_phys = phys_addr + (r_offset << 2);
  return_addr = addr_user + r_offset;

  printf(" JTAG data physical address: 0x%08lx,\n", data_phys  );
  printf(" JTAG response physical addresses; 0x%08lx\n", return_phys);

  /* open RORC and map registers */
  if (rorcOpenChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);
  timeout = ustimeout * prorc.loop_per_usec;
  if (timeout > DDL_MAX_WAIT_CYCLE )
  {
    printf("Error: time-out value %d us is too large.\n", ustimeout);
    printf("Please choose a value less than or equal to %d us.\n",
                                                      prorc.max_resp_time);
    exit (EXIT_FAILURE);
  }
  printf(" RORC revision = %d\n", prorc.rorc_revision);
  printf(" DIU version = %s\n", prorc.diu_version == NO_DIU ? "no DIU" :
                                prorc.diu_version == OLD ? "old" : "new");

 /* send start data receiver */
  phys_addr_rf = return_phys;
  rorcStartDataReceiver(&prorc, phys_addr_rf);
  printf(" Data Receiver started\n");

  sum_word = 0;
  while (1)
  {
    /* read JTAG data from file */
    n_word = JTAG2Memory(i_fp, data_addr, 8*length);
    if (n_word == 0)
      break;

    /* send JTAG */
    ret = ddlWriteJTAGBlock(&prorc, data_phys, n_word, 
                            return_phys, return_addr, param, 
                            timeout, stw, &n_reply, &step);

    if (ret)
    {
      printf(" ddlWriteJTAGBlock return code: %s, step = %d.\n", 
		       rorcInterpretReturnCode(ret), step);
      break;
    }

printf("%d words sent\n", n_word);

    for (i = 0; i < n_reply; i++)
      printf(" received status %d: 0x%08lx\n", i+1, stw[i].stw);

    /* write result to file */
    Memory2JTAG(o_fp, data_addr, return_addr, n_word);

    /* loop */
    sum_word += n_word;
    data_phys += (n_word << 2);
    data_addr += n_word;
  }
 
  /* close files */
  fclose(i_fp);
  fclose(o_fp);
  printf(" %d JTAG values sent and read back.\n", 8*sum_word);
 
  /* send stop data receiver */
  rorcStopDataReceiver(&prorc);
  printf(" Data Receiver stopped\n");
  
  /* unmap and close RORC */
  rorcClose(&prorc);

  /* unmap and close physmem memory */
  addr_user -= (pm_offset >> 2);
  if (physmemClose (physmem_fd, (unsigned long *)addr_user, phys_size_physmem))
    exit (EXIT_FAILURE);

  exit (0);
}
