/***************************************************************
* prorc_set_femu.c                                             *
* set (and read) FEE Emulator Card (FEMU) parameters           *
* Written by E. Denes                                          *
* Last modified on 26/09/2005                                  *
****************************************************************/
#define SOURCE_FILE "rorc_set_femu.c"
#define DESCRIPTION "Set FEIC registers"

#include "rorc_ddl.h"

unsigned long address, value;
int minor, channel, volt_cim;
int TimeOut, usTimeOut = DDL_RESPONSE_TIME;

/* variables for FEMU addresses */
char *address_tokens[] = {"PATGEN",          /* pattern generator mode */ 
                          "EVLEN",           /* generated event length */
                          "TRIG",            /* trigger mode */
                          "FLCTRL",          /* flow control mode */
                          "TRDIS",           /* transfer disable mode */
                          "SEED",            /* seed value */
                          ""
                         };
enum address_index {AD_PATGEN=0, AD_EVLEN, AD_TRIG, AD_FLCTRL, AD_TRDIS, AD_SEED};

/*************************************************************/
void print_usage(char *s)
{
   printf("\n-----------------------------------------------------------\n");
   printf("Usage: %s [-{M|m} <minor>] [-{C|c} <channel>\n", s);
   printf("                      [-{A|a} <FEMU_addr>] [-{D|d} <value>]\n");
   printf("where\n");
   printf("-M <minor>    : number of the RORC card (default: 0)\n");
   printf("-C <channel>  : number of the RORC channel (default: 0)\n");
   printf("-A <FEMU_addr>: FEMU address, where the value should be written\n");
   printf("                Possible values are:\n");
   printf("                     0x0 or 'PATGEN' : set pattern generator\n");
   printf("                     0x1 or 'EVLEN'  : set event length\n");
   printf("                     0x2 or 'TRIG'   : set trigger mode\n");
   printf("                     0x3 or 'FLCTRL' : set flow control mode\n");
   printf("                     0x4 or 'TRDIS'  : set transfer disable mode\n");
   printf("                     0x5 or 'SEED'   : set seed value\n");
   printf("   If no address given the program displays the values at all possible FEMU addresses\n");
   printf("-D <value>    : FEMU parameter corresponding to the given address.\n");
   printf("              0 (the default) means read and display the previously set value.\n"); 
   printf("-----------------------------------------------------------\n\n");

   exit (-1);
}

/****************************************************************************/


/********************************************************
*          treat parameters                             *
********************************************************/
void treat_param(int argc, char **argv)
{
  int c, c_index;
  char addr_code[16];


/* default values of the parameters */
  strcpy(addr_code, "PATGEN");
  volt_cim = 0;
  value = 0;

  minor = 0;
  channel = 0;

  if(argc >1)
    while((c = getopt(argc, argv, "M:m:C:c:A:a:D:d:")) != EOF)
      switch(toupper(c))
      {
        case 'M': minor = atoi(optarg); break;
        case 'C': channel = atoi(optarg);
          if ((channel < 0) || (channel >= MAX_CHANNEL))
          {
            printf(" Channel number should be positive and less than %d\n",
                                MAX_CHANNEL);
            print_usage (argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        case 'A': sprintf(addr_code, "%s", optarg); volt_cim = 1; break;
        case 'D': value = strtoul(optarg, (char**)NULL, 16); break;
        default : print_usage(argv[0]); break;
      }

  /* set FEMU address */

  if (volt_cim)
  {
    c_index = n_key_word(address_tokens, addr_code);
    if (c_index > 0) 
       address = c_index - 1;
    else if (addr_code[0] == '0' && addr_code[1] == 'x')
       address = strtoul(addr_code, (char**)NULL, 16); 
    else
       address = -1;
    if (address < 0)
    {
       printf ("Invalid address code %s\n", addr_code);
       print_usage(argv[0]);
    }
  }

}


/********************************************************
*    wait and read a CTSTW                              *
*    returns 0: OK                                      *
*           -1: not CTSTW arrived                       *
*           -2: no status in time out                   *
*           -3: error bit set in CTSTW                  *
********************************************************/
int readCTSTW(rorc_pci_dev_t *prorc, int transid, int destination, int time)
{
  int i, retval;
  stword_t stw;

  retval = 0;
  for (i = 0; i < time; i++)
  {
    if (rorcCheckStatus(prorc))
      break;
  }
  if (i >= time)
  {
    printf("readCTSTW: no status arrived in time-out %d usec\n", 
		                               time/prorc->loop_per_usec);
    retval = -2;
    return(retval);
  }

/* Read status */
  
  stw = ddlReadStatus(prorc);
  if (stw.part.code != CTSTW ||
      stw.part.dest != destination || 
      stw.part.trid != transid)
  {
    printf("NOT the expected CTSTW! Expected: 0x%x%x%x, received: 0x%x%x%x\n",
                             CTSTW, transid, destination,
                             stw.part.code, stw.part.trid, stw.part.dest);
    retval = -1;
  }
  else if(stw.part.error)
  {
    printf("Error bit set in CTSTW!\n");
    retval = -3;
  }
  return(retval);

}



/****************************************************************************/
int main (int argc, char **argv)
{

  rorc_pci_dev_t prorc;
  unsigned long command, fee_addr;
  unsigned long address1, address2;
  int retval, retval1, transid;
  stword_t stw;

  /* treat parameters */
  treat_param(argc, argv);

  /* open RORC and map mailboxes */
  if (rorcOpenChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);
  TimeOut = usTimeOut * prorc.loop_per_usec;

  transid = 0;
  if (value && volt_cim)
  {
    /* Send FE command */
    command = FECTRL;
    fee_addr = address * 0x100 + value;
    transid = incr15(transid);
    retval = ddlSendCommand(&prorc, FEE, command, transid, fee_addr, TimeOut);
    if (retval == RORC_LINK_NOT_ON)
    {
      printf("Link is not on, FEE command can not be sent\n");
      rorcClose(&prorc);
      exit (EXIT_FAILURE);
    }
    if (retval == RORC_TIMEOUT)
    {
      printf("FEE command can not be sent in timeout %d usec\n", usTimeOut);
      rorcClose(&prorc);
      exit (EXIT_FAILURE);
    }
    retval1 = readCTSTW(&prorc, transid, FEE, TimeOut);
  }

  if (volt_cim)
  {
    address1 = address;
    address2 = address;
  }
  else
  {
    address1 = 0;
    address2 = 0x5;
  }
  for (address = address1; address <= address2; address++)
  {
    /* Send FE status read */
    fee_addr = address * 0x100;
    transid = incr15(transid);
    stw = ddlReadFESTW(&prorc, fee_addr, transid, TimeOut);
    if (stw.part.param == 0x7ffff)
    {
      printf("FEE status can not be read\n");
      break;
    }
    else
      interpretFEMUSTW(address, stw);
  } 

  /* unmap and close RORC */
  rorcClose(&prorc);

  exit (0);
}
