/***********************************************************
 * rorc_status.c
 *
 *  Display the RORC's status and error registers         
 *
 * last updated: 25/08/2006
 * written by: Peter Csato and Ervin Denes
 ***********************************************************/
#define SOURCE_FILE "rorc_status.c"
#define DESCRIPTION "Show RORC channel's status and error registers"

#include <ctype.h>
#include "rorc_ddl.h"

/*******************************************************/

void usage(char *name)
{
  printf("Display the RORC channel's status and error registers.\n");
  printf("usage: %s [-{M|m} <RORC minor>|0] [-{C|c} <channel>|0]\n", name);
}

/********************************************************/

int main(int argc, char **argv)
{
  int c, minor, channel;
  int n1, n2, n3, n4;
  int ff, ff1, ff2;
  rorc_pci_dev_t prorc;
  rorcHwSerial_t hw;
  unsigned long fw;
  rorcStatus_t status;

  /* Check parameters */
  minor = 0;
  channel = 0;
  if (argc > 1)
  {
    while ((c = getopt(argc, argv, "M:m:C:c:")) != EOF)
    {
      switch(toupper(c))
      {
        case 'M': minor = atoi(optarg); if (minor < 0) minor = 0; break;
        case 'C': channel = atoi(optarg); 
          if ((channel < 0) || (channel >= MAX_CHANNEL)) 
          {
            printf(" Channel number should be positive and less than %d\n",
                                       MAX_CHANNEL);
            usage(argv[0]);
            exit (EXIT_FAILURE);
          }
          break;
        default : usage (argv[0]); exit (EXIT_FAILURE); break;
      }
    }
  }

  /* sw version */
  printf("Compiled RORC sw version: %s\n", RORC_DRIVER_VERSION);

  /* open RORC and map mailboxes */
  if (rorcMapChannel(&prorc, minor, channel))
    exit (EXIT_FAILURE);

  /* driver version */
  printf("RORC driver version: %d.%d.%d\n",
        prorc.driver_major, prorc.driver_minor, prorc.driver_release);
  printf("RORC minor = %d, channel = %d\n", minor, channel);
  if (prorc.rorc_revision > 0)
    printf("RORC revision = %d\n\n", prorc.rorc_revision);

  /* read non volatile ram or hw id from I2C bus */
  hw = rorcSerial(&prorc);
  rorcInterpretSerial(hw);

  if (pRorc(&prorc))
  {
    /* empty incoming mb1, 2, 3 and 4 if necessary */
    n1 = pRorcEmptyMb(&prorc, 1, 0);
    n2 = pRorcEmptyMb(&prorc, 2, 0);
    n3 = pRorcEmptyMb(&prorc, 3, 0);
    n4 = pRorcEmptyMb(&prorc, 4, 0);
    if (n1 + n2 + n3 + n4)
      printf ("Before sending pRORC command\n");
    if (n1)
      printf ("     IMB1 contained %d status.\n", n1);
    if (n2)
      printf ("     IMB2 contained %d status.\n", n2);
    if (n3)
      printf ("     IMB3 contained %d status.\n", n3);
    if (n4)
      printf ("     IMB4 contained %d status.\n", n4);
  }

  /* read fw id */
  fw = rorcReadFw (&prorc);
  rorcInterpretFw(fw);

  printf("RORC fw version is ");
  if (rorcCheckVersion(&prorc))
    printf("NOT ");
  printf("compatible with the present sw.\n");

  rorcReadRorcStatus(&prorc, &status);

  printf("\nrorc staus:\n");

  if (pRorc(&prorc))
    printf("OCSR: 0x%08x\n", status.ccsr);
  else
    printf("C_CSR: 0x%08x\n", status.ccsr);
  rorcInterpretStatus(&prorc, status.ccsr);

  if (pRorc(&prorc))
  {
    printf("RSER: 0x%08x\n", status.cerr);
    rorcInterpretErrors(status.cerr);
    rorcReset (&prorc, PRORC_CMD_CLEAR_ERROR);
  }
  else
  {
    printf("C_ERR: 0x%08x\n", status.cerr);
    rorcReset (&prorc, DRORC_CMD_CLEAR_ERROR);
  }

  if (pRorc(&prorc))
    printf("ECNT: 0x%06x (%d)\n", status.cdgs, status.cdgs);
  else
    printf("C_DGS: 0x%06x (%d)\n", status.cdgs, status.cdgs);

  if (pRorc(&prorc))
  {
    ff = rorcCheckFreeFifo(&prorc);
    printf("Free FIFO status: ");
    if (ff == RORC_FF_FULL)
      printf("full\n");
    else if (ff == RORC_FF_EMPTY)
      printf("empty\n");
    else
    {
      ff2 = (ff + 1) * 8;
      ff1 = ff2 - 7;
      printf("number of entries loaded is between %d and %d\n", ff1, ff2);
    }
  }

  printf("\n");

  /* unmap and close RORC */
  rorcClose(&prorc);

  exit (0);
}
